; Flip 3 . 2  FORTH-LIKE INTERPRETER PROGRAM
; Assembly Language Version
; by W. S. Walker, Professor of Digital Electronics and Robotics
; West Virginia State University CTC
; Assembled with Masm 5.1

; This implementation of FLIP will run on any machine from 8086 to pentium.
; It is written entirely in 8086 assembly language and requires the 8087
; coprocessor.

; version 3.2 changes over 3.1:
;  added < system w@ w! reparse error-on error-off clear-mem > words
;  corrected error with i, j, k words
;  corrected error where e formats weren't initialized on cold start
;  fixed an error with sin and cosine function
; version 3.1 changes over 3.0:
;  added text capabilities in graphics format
;  created font
;  added < g. xorgr-on xorgr-off l. eplaces estring > font words
; version 3.0 was the first assembly language version, previous versions
; where written in a basic compiler

.model   compact

.data

;------------------------------------------------------------------------------
; equates
;------------------------------------------------------------------------------

TRUE     equ     -1                    ; logical true
FALSE    equ     0                     ; logical false
BIOS     equ     10h                   ; bios video interrupt 
DOS      equ     21h                   ; DOS interrupt
VIDEO    equ     0B800h                ; color screen text segment
EGA      equ     0A000h                ; ega graphics segment
NULL     equ     0                     ; null character
CR       equ     0Dh                   ; enter key
LF       equ     0Ah                   ; line feed
BS       equ     08                    ; back space
SPC      equ     20h                   ; space
TAB      equ     09                    ; tab key
QUOTE    equ     22h                   ; quote - "
TIC      equ     27h                   ; tic mark or single quote mark
FWARW    equ     10h                   ; forward arrow
BKARW    equ     11h                   ; back arrow
UPARW    equ     18h                   ; up arrow
ALT_Q    equ     10h                   ; abort to DOS code
ESCP     equ     1Bh                   ; abort to mainloop code
TILDE    equ     7Eh                   ; tilde
varMax   equ     100                   ; maximum number of variables
fvarMax  equ     200                   ; maximum number of fvariables
conMax   equ     25                    ; maximum number of constants
fconMax  equ     50                    ; maximum number of fconstants
lblMax   equ     100                   ; maximum number of labels
wrdMax   equ     100                   ; maximum number of user words
rStkMax  equ     10                    ; run stack max
doStkMax equ     3                     ; do stack max
padStkMx equ     3                     ; pad stack max
pStkMax  equ     3                     ; push stack max
fpStkMax equ     3                     ; fpush stack max
mem      equ     32766                 ; bytes of storage
;8087
rndzbit  equ     0C00h                 ; 8087 control word round zero bits
rndzmsk  equ     0F3BFh                ; 8087 control word round zero mask

;------------------------------------------------------------------------------
; ega colors
;------------------------------------------------------------------------------

BLACK    equ     0
DARKBLUE equ     1
DARKGREN equ     2
DARKCYAN equ     3
DARKRED  equ     4
DARKMAGE equ     5
BROWN    equ     6
GREY     equ     7
DARKGREY equ     8
BLUE     equ     9
GREEN    equ     10
CYAN     equ     11
RED      equ     12
MAGENTA  equ     13
YELLOW   equ     14
WHITE    equ     15

;-----------------------------------------------------------------------------
; general storage
;-----------------------------------------------------------------------------

lbuf     db      0                     ; length of buffer
lpad     db      0                     ; length of pad
bufPtr   dw      0                     ; buffer pointer
token    db      80 dup (' '),NULL     ; current token
tname    db      '        ',NULL       ; temporary name storage
forClr   db      YELLOW                ; default foreground color
bakClr   db      DARKBLUE              ; default background color
inpClr   db      WHITE                 ; default input color
tmpClr   db      0                     ; hold color
tabWid   db      3                     ; convert tab to space count
ifmt     db      6                     ; integer field width
preIio   db      9 dup (' ')           ; pre integer i/o area
intIO    db      '      ',NULL         ; integer i/o
prmptch  db      FWARW                 ; prompt character
intStack dw      10 dup (0)            ; integer stack
iStkX    db      0                     ; integer stack index
mulDX    dw      0                     ; hold DX on double multiply
nDigs    db      0                     ; number of digits
cmdLin   db      6 dup (' '),'Command line: '
fname    db      12 dup (' '),NULL     ; filename
defFn    db      'test.4',NULL         ; default filename
outFn    db      'output.asc',NULL     ; output filename
iofname  db      12 dup(' '),NULL      ; i/o file name
iohand   dw      0                     ; i/o file handle
fhand    dw      0                     ; file handle
prnhand  dw      0                     ; print output file handle
prnTee   db      0                     ; print out flag, 1=tee ouput, 0=norm
fsize    dw      0                     ; file size
prgEnd   dw      0                     ; points to null at end of prgm space
nxtAdr   dw      0                     ; next scan adr after current token
strAdr   dw      0                     ; start address
holdEnd  dw      0                     ; hold end of program space
offVal   dw      0                     ; offset value
endAdr   dw      0                     ; end address
vadr     db      0                     ; current variable address
fvadr    db      0                     ; current fvariable address
temp     dw      0                     ; temporary storages
temp2    dw      0
         dw      0
ChrTmp   db      0                     ; character temporary
gflag    db      0                     ; graphics mode flag
xorGrfx  db      80h                   ; xor graphics bit
nxtNum   dw      0                     ; next number in buffer
pflag    db      0                     ; float print to pad flag
trflag   db      0
ranNum   dw      13579                 ; random number
gcurx    dw      0                     ; graphics cursor x
gcury    dw      0                     ; graphics cursor y
fmtwid   db      12                    ; fp total digits int + dec 
fmtdec   db      6                     ; fp decimal places
expdec   db      9                     ; exponential form # mantissa decimals
expdig   db      2                     ; exponential form # exponent digits
errflg   db      0                     ; error on flag
errjmp   db      8 dup (' '), NULL     ; error on jump tag

;-----------------------------------------------------------------------------
; The root dictionary
;-----------------------------------------------------------------------------

nDct     db      231                   ; number of dictionary entries
Dict     db      '.S',NULL             ; (001) display integer stack               
         db      '.',NULL              ; (002) print integer
         db      '+',NULL              ; (003) add integers
         db      '-',NULL              ; (004) sub integers
         db      '*',NULL              ; (005) mul integers
         db      '/',NULL              ; (006) div integers
         db      'D*',NULL             ; (007) mul integers double
         db      'D/',NULL             ; (008) div integers double
         db      'CR',NULL             ; (009) new line
         db      '.F',NULL             ; (010) display float stack
         db      'F.',NULL             ; (011) print float
         db      'F+',NULL             ; (012) add floats
         db      'F-',NULL             ; (013) sub floats
         db      'F*',NULL             ; (014) mul floats
         db      'F/',NULL             ; (015) div floats
         db      'DUP',NULL            ; (016) int dupe
         db      'FDUP',NULL           ; (017) flt dupe
         db      'DROP',NULL           ; (018) int drop
         db      'FDROP',NULL          ; (019) flt drop
         db      'SWAP',NULL           ; (020) int swap
         db      'FSWAP',NULL          ; (021) flt swap
         db      QUOTE,NULL            ; (022) quote
         db      'TYPE',NULL           ; (023) type
         db      'LEN',NULL            ; (024) pad length
         db      'CLS',NULL            ; (025) clear screen
         db      '\',NULL              ; (026) back slash - comment
         db      '(',NULL              ; (027) left paren - comment
         db      'FVARIABLE',NULL      ; (028) fvariable
         db      'VARIABLE',NULL       ; (029) variable
         db      'F!',NULL             ; (030) float store
         db      'F@',NULL             ; (031) float fetch
         db      '!',NULL              ; (032) integer store
         db      '@',NULL              ; (033) integer fetch
         db      'F?',NULL             ; (034) float fetch and print
         db      '?',NULL              ; (035) int fetch and print
         db      'IF',NULL             ; (036) if
         db      'JUMP',NULL           ; (037) jump
         db      '0<',NULL             ; (038) int test less than zero
         db      '0=',NULL             ; (039) int test equal to zero
         db      '0>',NULL             ; (040) int test g.t. zero
         db      'F0<',NULL            ; (041) flt test less than zero
         db      'F0=',NULL            ; (042) flt test equal to zero
         db      'F0>',NULL            ; (043) flt test g.t. zero
         db      '<',NULL              ; (044) int stack next < stack top
         db      '=',NULL              ; (045) int stack next = stack top
         db      '>',NULL              ; (046) int stack next > stack top
         db      'F<',NULL             ; (047) flt stack next < stack top
         db      'F=',NULL             ; (048) flt stack next = stack top
         db      'F>',NULL             ; (049) flt stack next > stack top
         db      'F0~',NULL            ; (050) flt approx equal to
         db      'NOT',NULL            ; (051) not
         db      'AND',NULL            ; (052) and
         db      'OR',NULL             ; (053) or
         db      'XOR',NULL            ; (054) xor
         db      'ROT',NULL            ; (055) int rotate
         db      'FROT',NULL           ; (056) float rot
         db      'PICK',NULL           ; (057) int pick
         db      'FPICK',NULL          ; (058) float pick
         db      ':',NULL              ; (059) colon (define word)
         db      ';',NULL              ; (060) semi-colon (end word def)
         db      'STOP',NULL           ; (061) return to prompt mode
         db      'SKIP',NULL           ; (062) skip
         db      'DO',NULL             ; (063) do
         db      'LOOP',NULL           ; (064) loop
         db      '+LOOP',NULL          ; (065) +loop
         db      'I',NULL              ; (066) i
         db      'J',NULL              ; (067) j
         db      'K',NULL              ; (068) k
         db      'UNDO',NULL           ; (069) undo
         db      'BEGIN',NULL          ; (070) begin
         db      'AGAIN',NULL          ; (071) again
         db      'UNBEGIN',NULL        ; (072) unbegin
         db      'ACCEPT',NULL         ; (073) accept
         db      'EXECUTE',NULL        ; (074) execute
         db      'EMIT',NULL           ; (075) emit
         db      'SPACE',NULL          ; (076) space
         db      'SPACES',NULL         ; (077) spaces
         db      '1+',NULL             ; (078) 1+ increment
         db      '1-',NULL             ; (079) 1- decrement
         db      'INC',NULL            ; (080) increment
         db      'DEC',NULL            ; (081) decrement
         db      'FINT',NULL           ; (082) float to integer (round)
         db      'F>N',NULL            ; (083) float to integer (no rounding)
         db      'N>F',NULL            ; (084) integer to float 
         db      'KEY',NULL            ; (085) get key 
         db      'FSQRT',NULL          ; (086) float sqrt
         db      'F10^X',NULL          ; (087) float 10^x
         db      'FE^X',NULL           ; (088) float e^x
         db      'FLOG10',NULL         ; (089) float log10
         db      'FLN',NULL            ; (090) float loge
         db      'FTAN',NULL           ; (091) float tan
         db      'PI',NULL             ; (092) pi
         db      'FD2R',NULL           ; (093) degress to radians
         db      'FR2D',NULL           ; (094) radians to degrees
         db      'FATAN',NULL          ; (095) float arctan
         db      'FSIN',NULL           ; (096) float sin
         db      'FCOS',NULL           ; (097) float cos
         db      'ABS',NULL            ; (098) abs
         db      'FABS',NULL           ; (099) fabs
         db      'NEG',NULL            ; (100) neg
         db      'FNEG',NULL           ; (101) fneg
         db      'FEMPTY',NULL         ; (102) fempty
         db      'EMPTY',NULL          ; (103) empty
         db      'CLEAN',NULL          ; (104) clean
         db      'DEPTH',NULL          ; (105) depth
         db      'FDEPTH',NULL         ; (106) fdepth
         db      'PLACES',NULL         ; (107) places
         db      'FPLACES',NULL        ; (108) fplaces
         db      'FGCOLOR',NULL        ; (109) foreground color
         db      'BGCOLOR',NULL        ; (110) background color
         db      'INCOLOR',NULL        ; (111) input routine text color
         db      'COLD',NULL           ; (112) cold start reset
         db      'LIST',NULL           ; (113) list
         db      'LOAD',NULL           ; (114) load
         db      '+LOAD',NULL          ; (115) +load
         db      'WORDS',NULL          ; (116) words
         db      'H.',NULL             ; (117) hex print
         db      'AT-XY',NULL          ; (118) set text screen location
         db      'TRUE',NULL           ; (119) true
         db      'FALSE',NULL          ; (120) false
         db      'BYE',NULL            ; (121) bye - exits to OS
         db      'MEM',NULL            ; (122) mem - compute memory
         db      'ALLOT',NULL          ; (123) allot variable space
         db      'FALLOT',NULL         ; (124) allot fvariable space
         db      'PUSH',NULL           ; (125) push
         db      'POP',NULL            ; (126) pop
         db      'FPUSH',NULL          ; (127) fpush
         db      'FPOP',NULL           ; (128) fpop
         db      'EPSILON!',NULL       ; (129) epsilon store
         db      'EPSILON@',NULL       ; (130) epsilon fetch
         db      'FCONSTANT',NULL      ; (131) fconstant
         db      'CONSTANT',NULL       ; (132) constant
         db      'VLIST',NULL          ; (133) list variables and constants
         db      ',',NULL              ; (134) comma - array var
         db      'F,',NULL             ; (135) fcomma - array fvar
         db      'TEE-ON',NULL         ; (136) tee on
         db      'TEE-OFF',NULL        ; (137) tee off
         db      'GRAPHICS',NULL       ; (138) graphic mode 640x350x16 (VGA)
         db      'PLOT',NULL           ; (139) plot pixel
         db      'TEXT',NULL           ; (140) text
         db      'FORGET',NULL         ; (141) forget last word
         db      'FATAN2',NULL         ; (142) float full arctangent
         db      'MOD',NULL            ; (143) mod
         db      'FACOS',NULL          ; (144) float arccos
         db      'FASIN',NULL          ; (145) float arcsin
         db      'FPWR',NULL           ; (146) float x^y
         db      'FCUBRT',NULL         ; (147) float cube root
         db      'FSINH',NULL          ; (148) float hyperbolic sin
         db      'FCOSH',NULL          ; (149) float hyperbolic cos
         db      'FTANH',NULL          ; (150) float hyperbolic tan 
         db      'FATANH',NULL         ; (151) float hyperbolic arc tan
         db      'LINE',NULL           ; (152) draw line
         db      'EXIT',NULL           ; (153) exit
         db      'NUMBER',NULL         ; (154) pad to int number
         db      'FNUMBER',NULL        ; (155) pad to flt number
         db      'STRING',NULL         ; (156) append int to pad
         db      'FSTRING',NULL        ; (157) append flt to pad
         db      'CIRCLE',NULL         ; (158) craw circle
         db      'PAD!',NULL           ; (159) pad store
         db      'PAD@',NULL           ; (160) pad fetch
         db      'FREE',NULL           ; (161) address of free memory 
         db      'PAD',NULL            ; (162) address of pad
         db      'C!',NULL             ; (163) store character
         db      'C@',NULL             ; (164) fetch character
         db      TIC,NULL              ; (165) tic
         db      'DROP2',NULL          ; (166) int drop two
         db      'FDROP2',NULL         ; (167) flt drop two
         db      '+SPACES',NULL        ; (168) add spaces to pad
         db      'SLICE',NULL          ; (169) slice pad
         db      'CONCAT',NULL         ; (170) concatenate
         db      'UPPER',NULL          ; (171) upper case
         db      'LOWER',NULL          ; (172) lower case
         db      'BLACK',NULL          ; (173) black 0
         db      'DARKBLUE',NULL       ; (174) dark blue 1
         db      'DARKGREEN',NULL      ; (175) dark green 2
         db      'DARKCYAN',NULL       ; (176) dark cyan 3
         db      'DARKRED',NULL        ; (177) dark red 4
         db      'DARKMAGENTA',NULL    ; (178) dark magenta 5
         db      'BROWN',NULL          ; (179) brown 6
         db      'GREY',NULL           ; (180) grey 7
         db      'DARKGREY',NULL       ; (181) dark grey 8
         db      'BLUE',NULL           ; (182) blue 9
         db      'GREEN',NULL          ; (183) green 10
         db      'CYAN',NULL           ; (184) cyan 11
         db      'RED',NULL            ; (185) cyan 12
         db      'MAGENTA',NULL        ; (186) magenta 13
         db      'YELLOW',NULL         ; (187) yellow 14
         db      'WHITE',NULL          ; (188) white 15
         db      'DATE',NULL           ; (189) date
         db      'TIME',NULL           ; (190) time
         db      'RANDOM',NULL         ; (191) random
         db      'ASPECT',NULL         ; (192) aspect
         db      '=>',NULL             ; (193) copy TOIS to pstack
         db      '<=',NULL             ; (194) copy pstack to TOIS
         db      'F=>',NULL            ; (195) copy TOFS to fpstack
         db      'F<=',NULL            ; (196) copy fpstack to TOFS
         db      '!=',NULL             ; (197) int not equal
         db      'F!=',NULL            ; (198) float not equal
         db      'PAD=',NULL           ; (199) pad =
         db      'SEARCH',NULL         ; (200) search
         db      'CREATE-FILE',NULL    ; (201) create file
         db      'OPEN-FILE',NULL      ; (202) open file
         db      'CLOSE-FILE',NULL     ; (203) close file
         db      'WRITE-FILE',NULL     ; (204) write file
         db      'READ-FILE',NULL      ; (205) read file
         db      'TAB',NULL            ; (206) tab width
         db      'LOCATE',NULL         ; (207) locate text cursor
         db      'READ',NULL           ; (208) read constant
         db      'WRITE',NULL          ; (209) write constant
         db      'APPEND',NULL         ; (210) append constant
         db      'TRACE-ON',NULL       ; (211) trace on
         db      'TRACE-OFF',NULL      ; (212) trace off
         db      'E.',NULL             ; (213) print float as exponential
         db      'EPLACES',NULL        ; (214) float exponential form format
         db      'ESTRING',NULL        ; (215) float exp form to pad
         db      '.P',NULL             ; (216) display push stacks
         db      'G.',NULL             ; (217) print graphics char
         db      'L.',NULL             ; (218) logical print
         db      'FONT',NULL           ; (219) font
         db      'XORGR-ON',NULL       ; (220) xor graphics on
         db      'XORGR-OFF',NULL      ; (221) xor graphics off
         db      'SYSTEM',NULL         ; (222) system
         db      'W@',NULL             ; (223) word fetch
         db      'W!',NULL             ; (224) word store
         db      'REPARSE',NULL        ; (225) reparse
         db      '>>',NULL             ; (226) shift right
         db      '<<',NULL             ; (227) shift left
         db      'ERROR-ON',NULL       ; (228) error on
         db      'ERROR-OFF',NULL      ; (229) error off
         db      'CLEAR-MEM',NULL      ; (230) clear memory
         db      'FREE',NULL           ; (231) address first unused cpu byte
  
;-----------------------------------------------------------------------------       
; Error messages
;-----------------------------------------------------------------------------

codMsg   db      GREEN,'Code:  ',NULL         
wrdMsg   db      GREEN,'Word: ',NULL
iStkOEr  db      RED,'integer stack overflow',NULL
fStkOEr  db      RED,'float stack overflow',NULL
iStkEEr  db      RED,'insufficient integer stack',NULL
fStkEEr  db      RED,'insufficient float stack',NULL
wrdErr   db      RED,'not a word or number',NULL
qotErr   db      RED,'no ending quote',NULL
qtpErr   db      RED,'pad size exceeded',NULL
invmdEr  db      RED,'invalid mode',NULL
opnErr   db      RED,'file open error',NULL
closErr  db      RED,'file close error',NULL
creErr   db      RED,'file create error',NULL
wrtErr   db      RED,'file write error',NULL
opmErr   db      RED,'out of program memory',NULL
rdfErr   db      RED,'file read error',NULL
cmtErr   db      RED,'no ending paren',NULL
vrnErr   db      RED,'missing or invalid tag name',NULL
vdupErr  db      RED,'variable or fvariable name in use',NULL
cdupErr  db      RED,'constant or fconstant name in use',NULL
wdupErr  db      RED,'word name in use',NULL
ovsErr   db      RED,'out of integer variable space',NULL
ofvsErr  db      RED,'out of float variable space',NULL
ocnsErr  db      RED,'out of constant space',NULL
ofcnsEr  db      RED,'out of fconstant space',NULL
cnmErr   db      RED,'missing or invalid integer constant',NULL
fcnmErr  db      RED,'missing or invalid float constant',NULL
lbnErr   db      RED,'missing or invalid label name',NULL
lnfErr   db      RED,'label not found',NULL
owdsErr  db      RED,'out of word name space',NULL
wdnErr   db      RED,'missing or invalid word name',NULL
nscErr   db      RED,'no ; terminator',NULL
wddWrn   db      YELLOW,'word already defined',NULL
rStkOEr  db      RED,'run stack overflow',NULL
rStkEmt  db      RED,'run stack empty',NULL
lnxErr   db      RED,'skip out of range',NULL
doStkOEr db      RED,'do stack overflow',NULL
nodoErr  db      RED,'no do',NULL
idxaErr  db      RED,'index not available',NULL
nobaErr  db      RED,'no begin',NULL
baStkOEr db      RED,'begin...again stack overflow',NULL
mathErr  db      RED,'math error',NULL
rutwMsg  db      GREEN,'Root words: ',CR,LF,NULL
usrwMsg  db      GREEN,'User words: ',CR,LF,NULL
pStkOEr  db      RED,'push stack overflow',NULL
pStkEEr  db      RED,'insufficient push stack',NULL
fpStkOEr db      RED,'fpush stack overflow',NULL
fpStkEEr db      RED,'insufficient fpush stack',NULL
flnErr   db      RED,'missing or invalid file name',NULL
vrlMsg   db      GREEN,'variables:',CR,LF,NULL
fvrlMsg  db      GREEN,'fvariables:',CR,LF,NULL
cnlMsg   db      GREEN,'constants:',CR,LF,NULL
fcnlMsg  db      GREEN,'fconstants:',CR,LF,NULL
cmaErr   db      RED,'no array exist',NULL
pltErr   db      RED,'graphics mode only',NULL
frgtEr   db      RED,'nothing to forget',NULL
intWrn   db      YELLOW,'not an integer',NULL
fltWrn   db      YELLOW,'not a float',NULL
bpdsErr  db      RED,'pad storage error',NULL
ticErr   db      RED,'no character constant',NULL

;------------------------------------------------------------------------------ 
; messages
;------------------------------------------------------------------------------

msgLogo  db      '+ + + F L I P  version 3.2 + + + by W. S. Walker',CR,LF,NULL
imtMsg   db      '-- integer stack empty --',NULL
fmtMsg   db      '-- float stack empty --',NULL
pstMsg   db      '-- push stacks empty --',NULL
ldfMsg   db      ' bytes loaded',CR,LF,NULL
nunMsg   db      '-- none --',CR,LF,NULL
trwmMsg  db      'word: ',NULL
trisMsg  db      'int stk: ',NULL
trfsMsg  db      'flt stk: ',NULL
truMsg   db      'True ',NULL
FlsMsg   db      'False ',NULL

;------------------------------------------------------------------------------ 
; 8087 coprocessor data
;------------------------------------------------------------------------------
    
precbit  dw      57                    ; set for 57 bits of precision
npustat  dw      ?                     ; 8087 status
npucntl  dw      ?                     ; 8087 control
fltzer   dt      0.0                   ; fp constant zero
preFio   db      20 dup (' ')          ; blank area
fltIO    db      40 dup ('?')          ; fp buffer
fltint   dt      0                     ; fp integer bcd significand
fltexp   dw      0                     ; fp exponent
fpbufptr dw      0                     ; fp buffer pointer
fltStack dt      10 dup (0.0)          ; float stack
fStkX    db      0                     ; float stack index
fptemp   dt      0.0                   ; float temporary
epsilon  dt      0.0000000001          ; float test for apprx zero parm
trgsgn   dw      0                     ; trig function sign
; Misc floating constants
fpid2    dt      1.570796326794896619  ; pi/2
f3pid2   dt      4.712388980384689858  ; 3pi/2
f2pi     dt      6.283185307179586477  ; 2pi
fp5      dt      0.5                   ; 1/2
f1d3     dt      0.3333333333333333333 ; 1/3
fd2r     dt      0.01745329251994329577 ; pi/180
fr2d     dt      57.29577951308232088  ; 180/pi
log10e   dt      0.4342944819032518276 ; log base 10 of e
loge10   dt      2.302585092994045684  ; log base e of 10
fminus1  dt      -1.0                  ; minus 1
tst1lo   dt      0.999999999999999995  ; 1 neighborhood test low
tst1hi   dt      1.00000000000000005   ; 1 neighborhood test high
fnhi     dt      32768.                ; largest integer + 1
fnlo     dt      -32769.               ; smallest integer - 1
fpexp    dw      0                     ; floating point exponent
sinsgn   dw      0                     ; sine function sign

;------------------------------------------------------------------------------
; variable, constant, label, run, do and begin stack storage area
;------------------------------------------------------------------------------

; variables
varX     db      0
vnams    db      varMax dup ('        ')
vars     dw      varMax dup (0)
         dw      18 dup (0)
X1       dw      0                     ; last x pixel
Y1       dw      0                     ; last y pixel

; fvariables
fvarX    db      0
fvnams   db      fvarMax dup ('        ')
fvars    dt      fvarMax dup (0.0)
         dt      19 dup (0.0)
aspect   dt      1.2                   ; aspect ratio

; constants
conX     db      0
connams  db      conMax dup ('        ')
cons     dw      conMax dup (0)

; fconstants
fconX    db      0
fconnams db      fconMax dup ('        ')
fcons    dt      fconMax dup (0.0)

; labels
lblX     db      0
lblnams  db      lblMax dup ('        ')
lblAdr   dw      lblMax dup (0)

; words
wrdX     db      0
wrdnams  db      wrdMax dup ('        ')
wrdAdr   dw      wrdMax dup (0)
rStack   dw      rStkMax dup (0)
rStkX    db      0

; do...loop
dostkX   db      0
doStack  dw      doStkMax dup (0) 
doIJKStk dw      doStkMax dup (0)
doLimStk dw      doStkMax dup (0)

; begin...again
baStkX   db      0
baStack  dw      doStkMax dup(0)

;push stacks
pStkX    db      0
pStack   dw      pStkMax dup(0)
fpStkX   db      0
fpStack  dt      fpStkMax dup (0.0)

;pad stack
padStack db      padStkMx dup( NULL, 79 dup (' ') )

maxPsz   dw      mem                    ; max program 
maxPadr  dw      mem+offset program     ; max program address

;------------------------------------------------------------------------------
; subroutine area
; these are FLIP executables
; any FLIP words can be used except load, +load, clear-mem
; make sure CR/LF are supplied if skip is used
; calling sequence: save bp, set bp to address of sub, call parse, pop bp
;------------------------------------------------------------------------------

presub   db      NULL,CR,LF
         db      9 dup (' '),'( fatan2 )'
subr     db     ' '

; full 360 degree arctan ( X Y -- a ) -2*pi <= a <= 2*pi
fatan2   db      '2 fpick f0= if 1 skip',CR,LF,'fdup fpush fswap fdup fpush f/ '
         db      'fatan fpop fdup frot frot f0< n>f pi f* f- fswap f0> n>f fpop '
         db      'f0< n>f f* pi 2. f* f* f+ 4 skip',CR,LF,'fdup f0< if 2 skip',CR,LF
         db      'f0= if fdrop 0. 2 skip',CR,LF
         db      'fdrop pi 0.5 f* 1 skip',CR,LF,'fdrop2 pi 1.5 f*',CR,LF
         db      '~$$$fatn',CR,LF,NULL

; hyperbolic sin
         db      9 dup (' ')
         db      ' ( fsinh ) '
fsinh    db      'fdup fe^x fswap -1. f* fe^x f- 0.5 f*',CR,LF,NULL

; hyperbolic cos
         db      9 dup (' ')
         db      ' ( fcosh ) '
fcosh    db      'fdup fe^x fswap -1. f* fe^x f+ 0.5 f*',CR,LF,NULL

; hyperbolic tan
         db      9 dup (' ')
         db     ' ( ftanh ) '
ftanh    db      'fdup fsinh fswap fcosh f/',CR,LF,NULL

; hperbolic arc tan
         db      8 dup (' ')
         db      ' ( fatanh ) '
fatanh   db      'fdup 1. f+ fswap 1. fswap f- f/ fln 0.5 f*',CR,LF,NULL

         db      11 dup (' ')
         db      ' ( line ) '
line     db      '119 @ 120 @ 3 pick 2 pick - dup => 0= if jump $$$ld1',CR,LF
         db      '5 pick 4 pick - dup => 0= if jump $$$ld2',CR,LF
         db      '/ abs 1 > if jump $$$ld3',CR,LF
         db      '<= 0< 2 * 1+ pop n>f pop n>f fswap f/',CR,LF
         db      '3 pick begin',CR,LF
         db      ' " fdup dup n>f 4 pick n>f f- f* 3 pick n>f f+ .5 f+ dup f>n plot" execute',CR,LF
         db      ' 2 pick + dup 7 pick = if 1 skip',CR,LF
         db      'again',CR,LF
         db      'unbegin execute jump $$$ld4',CR,LF
         db      '~$$$ld3 pop pop dup 0< 2 * 1+ rot rot n>f n>f fswap f/',CR,LF
         db      '2 pick begin',CR,LF
         db      ' " fdup dup n>f 3 pick n>f f- f* 4 pick n>f f+ .5 f+ dup f>n swap plot" execute',CR,LF
         db      ' 2 pick + dup 6 pick = if 1 skip',CR,LF
         db      'again',CR,LF
         db      'unbegin execute',CR,LF
         db      '~$$$ld4 fdrop jump $$$ld5',CR,LF
         db      '~$$$ld1 5 pick 4 pick - dup rot = if jump $$$ld7',CR,LF
         db      '0< 2 * 1+',CR,LF
         db      '3 pick begin',CR,LF
         db      ' dup 4 pick plot',CR,LF
         db      ' 2 pick + dup 7 pick = if 1 skip',CR,LF
         db      'again',CR,LF
         db      'unbegin dup 4 pick plot jump $$$ld6',CR,LF
         db      '~$$$ld2 drop 0< 2 * 1+',CR,LF
         db      '2 pick begin',CR,LF
         db      ' dup 5 pick swap plot',CR,LF
         db      ' 2 pick + dup 6 pick = if 1 skip',CR,LF
         db      'again',CR,LF
         db      'unbegin dup 5 pick swap plot',CR,LF
         db      'pop drop',CR,LF
         db      '~$$$ld6 pop drop',CR,LF
         db      '~$$$ld5 drop',CR,LF
         db      '~$$$ld7 drop drop2 drop2',CR,LF,NULL

; draw circle
         db      8 dup (' ')
         db      ' ( circle ) '
circle   db      '0.',CR,LF
         db      'fdup fdup fsin dup n>f f* 2 pick n>f f+ .5 f+',CR,LF
         db      'fswap fcos dup n>f f* 220 f@ f* 3 pick n>f f+ .5 f+',CR,LF
         db      'f>n f>n plot 1. f+ fdup 360. f< if -3 skip',CR,LF
         db      'drop2 drop fdrop',CR,LF,NULL

;------------------------------------------------------------------------------
; program storage area
;------------------------------------------------------------------------------

prepad   db      NULL,CR,LF, 20 dup (' ')
pad      db      80 dup (' '),NULL      ; pad

prebuf   db      NULL,CR,LF, 20 dup (' ')
buffer   db      80 dup (' '),NULL      ; buffer

; program

preprgm  db 	NULL,CR,LF, 20 dup (' ')
program  db 	mem dup (' ')
EndPM    db     NULL

;------------------------------------------------------------------------------
; graphic character font table
;------------------------------------------------------------------------------

gfont   db      16 dup ( 8 dup(0) )
        db      00h,00h,30h,3Ch,3Eh,3Ch,30h,00h           ; right marker
        db      00h,00h,0Ch,3Ch,7Ch,3Ch,0Ch,00h           ; left marker
        db      6 dup ( 8 dup(0) )
        db      00h,18h,3Ch,7Eh,18h,18h,18h,00h           ; up arrow
        db      00h,18h,18h,18h,7Eh,3Ch,18h,00h           ; down arrow
        db      00h,00h,08h,0Ch,7Eh,0Ch,08h,00h           ; right arrow
        db      00h,00h,10h,30h,7Eh,30h,10h,00h           ; left arrow
        db      4 dup ( 8 dup(0) )
        db      8 dup(0)                                  ; blank
        db      18h,18h,18h,18h,00h,18h,18h,00h           ; !
        db      6Ch,6Ch,6Ch,6Ch,00h,00h,00h,00h           ; "
        db      6Ch,6Ch,0FEh,6Ch,0FEh,6Ch,6Ch,00h         ; #
        db      18h,3Ch,60h,30h,18h,0Ch,78h,30h           ; $
        db      00h,0C6h,0CCh,18h,30h,66h,0C6h,00h        ; %
        db      70h,0D8h,0D8h,70h,0DAh,0DCh,76h,00h       ; &
        db      00h,30h,30h,10h,00h,00h,00h,00h           ; '
        db      0Ch,18h,30h,30h,30h,30h,18h,0Ch           ; (
        db      30h,18h,0Ch,0Ch,0Ch,0Ch,18h,30h           ; )
        db      00h,0DBh,7Eh,3Ch,7Eh,0DBh,00h,00h         ; *
        db      00h,00h,18h,18h,7Eh,18h,18h,00h           ; +
        db      00h,00h,00h,00h,00h,30h,30h,10h           ; ,
        db      00h,00h,00h,00h,7Eh,00h,00h,00h           ; -
        db      00h,00h,00h,00h,00h,30h,30h,00h           ; .
        db      03h,06h,0Ch,18h,30h,60h,0C0h,00h          ; /
        db      38h,6Ch,0C6h,0D6h,0C6h,6Ch,38h,00h        ; 0
        db      18h,38h,18h,18h,18h,18h,3Ch,00h           ; 1
        db      7Ch,0C6h,06h,3Ch,60h,0C0h,0FEh,00h        ; 2
        db      7Ch,0C6h,06h,1Ch,06h,0C6h,7Ch,00h         ; 3
        db      1Ch,3Ch,6Ch,0CCh,0FEh,0Ch,0Ch,00h         ; 4
        db      0FCh,0C0h,0F8h,0Ch,06h,0C6h,7Ch,00h       ; 5
        db      0Ch,18h,30h,3Ch,66h,66h,3Ch,00h           ; 6
        db      0FEh,0C6h,0Ch,18h,30h,60h,60h,00h         ; 7
        db      3Ch,66h,66h,3Ch,66h,66h,3Ch,00h           ; 8
        db      3Ch,66h,66h,3Eh,0Ch,18h,30h,00h           ; 9
        db      00h,30h,30h,00h,00h,30h,30h,00h           ; :
        db      00h,00h,30h,30h,00h,30h,30h,10h           ; ;
        db      0Ch,18h,30h,60h,30h,18h,0Ch,00h           ; <
        db      00h,00h,7Ch,00h,7Ch,00h,00h,00h           ; =
        db      60h,30h,18h,0Ch,18h,30h,60h,00h           ; >
        db      7Ch,0C6h,0CCh,18h,18H,00h,18h,00h         ; ?
        db      3Eh,63h,0CFh,0DBh,0CEh,60h,3Ch,00h        ; @
	db      38h,6Ch,0C6h,0C6h,0FEh,0C6h,0C6h,00h      ; A
        db      0FCh,66h,66h,07Ch,066h,066h,0FCh,00h      ; B 
        db      3Ch,66h,0C0h,0C0h,0C0h,66h,3Ch,00h        ; C 
        db      0FCh,6Eh,66h,66h,66h,6Eh,0FCh,00h         ; D 
        db      0FEh,62h,68h,78h,68h,62h,0FEh,00h         ; E 
        db      0FEh,62h,68h,78h,68h,60h,0F0h,00h         ; F 
        db      3Ch,66h,0C0h,0C0h,0CEh,66h,3Eh,00h        ; G 
        db      0C6h,0C6h,0C6h,0FEh,0C6h,0C6h,0C6h,00h    ; H 
        db      3Ch,18h,18h,18h,18h,18h,3Ch,00h           ; I 
        db      3Ch,18h,18h,18h,0D8h,0D8h,070h,00h        ; J 
        db      0E6h,06Ch,78h,70h,78h,6Ch,0E6h,00h        ; K 
        db      0F0h,60h,60h,60h,60h,66h,0FEh,00h         ; L 
        db      0C3h,0E7h,0FFh,0DBh,0C3h,0C3h,0C3h,00h    ; M 
        db      0C6h,0E6h,0F6h,0DEh,0CEh,0C6h,0C6h,00h    ; N 
        db      3Ch,66h,0C3h,0C3h,0C3h,66h,3Ch,00h        ; O 
        db      0FCh,66h,66h,7Ch,60h,60h,0F0h,00h         ; P
        db      3Ch,66h,0C3h,0C3h,0C3h,6Eh,3Ch,06h        ; Q
        db      0FCh,66h,66h,7Ch,6Ch,66h,0E3h,00h         ; R
        db      7Ch,0C6h,60h,38h,0Ch,0C6h,7Ch,00h         ; S
        db      0FFh,99h,18h,18h,18h,18h,3Ch,00h          ; T
        db      0C6h,0C6h,0C6h,0C6h,0C6h,6Ch,7Ch,00h      ; U
        db      0C6h,0C6h,0C6h,6Ch,6Ch,38h,10h,00h        ; V
        db      0C3h,0C3h,0C3h,0Dbh,0FFh,0E7h,0C3h,00h    ; W
        db      0C6h,6Ch,38h,10h,38h,6Ch,0C6h,00h         ; X
        db      0C6h,0C6h,0C6h,7Ch,18h,18h,18h,00h        ; Y
        db      0FEh,0C6h,8Ch,18h,32h,66h,0FEh,00h        ; Z
        db      78h,60h,60h,60h,60h,60h,78h,00h           ; [
        db      0C0h,60h,30h,18h,0Ch,06h,03h,00h          ; \
        db      78h,18h,18h,18h,18h,18h,78h,00h           ; ]
        db      10h,38h,6Ch,0C6h,00h,00h,00h,00h          ; ^
        db      00h,00h,00h,00h,00h,00h,00h,7Eh           ; _
        db      00h,0Ch,0Ch,08h,00h,00h,00h,00h           ; tic
        db      00h,00h,70h,18h,78h,0DCh,76h,00h          ; a
        db      0E0h,60h,60h,7Ch,66h,66h,0FCh,00h         ; b
        db      00h,00h,3Ch,66h,60h,66h,3Ch,00h           ; c
        db      1Ch,0Ch,0Ch,7Ch,0CCh,0CCh,7Eh,00h         ; d
        db      00h,00h,3Ch,66h,7Eh,60h,3Ch,00h           ; e
        db      1Ch,36h,30h,7Ch,30h,30h,78h,00h           ; f
        db      00h,00h,3Eh,66h,66h,3Eh,06h,3Ch           ; g
        db      0E0h,60H,60H,7Ch,66h,66h,0E6h,00h         ; h
        db      00h,18h,00h,38h,18h,18h,3Ch,00h           ; i
        db      00h,18h,00h,38h,18h,18h,0D8h,70h          ; j
        db      0E0h,60h,66h,6Ch,78h,6Ch,0E6h,00h         ; k
        db      38h,18h,18h,18h,18h,18h,3Ch,00h           ; l
        db      00h,00h,0C6h,0EEh,0D6h,0C6h,0C6h,00h      ; m
        db      00h,00h,7Ch,66h,66h,66h,66h,00h           ; n
        db      00h,00h,3Ch,66h,66h,66h,3Ch,00h           ; o
        db      00h,00h,7Ch,36h,36h,3Ch,30h,78h           ; p
        db      00h,00h,3Bh,6Eh,6Ch,3Ch,0Ch,0Eh           ; q
        db      00h,00h,7Ch,36h,30h,30h,78h,00h           ; r
        db      00h,00h,3Ch,62h,3Ch,46h,3Ch,00h           ; s
        db      00h,18h,7Eh,18h,18h,1Ah,0Ch,00h           ; t
        db      00h,00h,0CCh,0CCh,0CCh,0CEh,7Bh,00h       ; u
        db      00h,00h,0C6h,0C6h,0C6h,6Ch,38h,00h        ; v
        db      00h,00h,0C3h,0DBh,0FFh,0E7h,0C3h,00h      ; w
        db      00h,00h,0C6h,6Ch,38h,6Ch,0C6h,00h         ; x
        db      00h,00h,0CCh,0CCh,7Ch,0Ch,0Ch,38h         ; y
        db      00h,00h,0FCh,98h,30h,64h,0FCh,00h         ; z
        db      0Ch,18h,18h,70h,18h,18h,0Ch,00h           ; {
        db      18h,18h,18h,00h,18h,18h,18h,00h           ; |
        db      30h,18h,18h,0Eh,18h,18h,30h,00h           ; }
        db      00h,76h,0DCh,00h,00h,00h,00h,00h          ; ~ 
        db      00h,10h,38h,6Ch,0C6h,0C6h,0FEh,00h        ; delta
        db      77 dup ( 8 dup(0) )
        db      00h,00h,00h,0FFh,00h,0FFh,00h,00h         ; double ul- 205
        db      21 dup ( 8 dup(0) )
        db      00h,00h,0FEh,6Ch,6Ch,6Ch,6Eh,00h          ; pi
        db      20 dup ( 8 dup(0) )
        db      00h,1Ch,36h,1Ch,00h,00h,00h,00h           ; degree
        db      8 dup(0)
        db      00h,00h,00h,18h,18h,00h,00h,00h           ; dot
        db      2 dup ( 8 dup(0) )

.code

;------------------------------------------------------------------------------
; clear screen
; entry: bakClr and forClr are background color and foreground colors
; cursor location updated
; uses direct screen write for color, BIOS call for text
;------------------------------------------------------------------------------
 
cls:     push ax                       ; save regs
         push cx                        
         push di
         push es
         mov  ax,VIDEO                 ; put screen address in es
         mov  es,ax  
         mov  di,0                     ; set starting address upper left
         mov  al,SPC
         mov  ah,bakClr                ; set colors
         mov  cl,4
         sal  ah,cl
         add  ah,forClr
         mov  cx,25*80                 ; word count of whole screen
         rep  stosw                    ; clear the screen
         pop  es                       ; restore regs
         pop  di                       
         pop  cx
         pop  ax
         ret                           ; done with clear screen

;-----------------------------------------------------------------------------
; graphics clear screen
;-----------------------------------------------------------------------------

gcls:     mov     ax,0010h             ; graphics - set video mode 
          int     BIOS 
          jmp     ghome                ; go set cursor home
         
;-----------------------------------------------------------------------------
; text home cursor
; sets row and column to 0,0
; uses set cursor
;-----------------------------------------------------------------------------

homCur:  push    dx                    ; save reg
         xor     dx,dx                 ; row, col = 0,0
         call    setCur                ; set cursor
         pop     dx                    ; restore dx
         ret                           ; done with home cursor

;-----------------------------------------------------------------------------
; graphics home cursor
; sets gcurx, gcury to 0,0
;-----------------------------------------------------------------------------

ghome:   mov     gcurx,0               ; set graphics cursor to 0,0
         mov     gcury,0
         ret                           ; done with ghome

;-----------------------------------------------------------------------------
; set cursor
; entry: dh = row, dl = column
; uses BIOS call
;-----------------------------------------------------------------------------

setCur:  push    ax                    ; save regs
         push    bx
         push    cx
         xor     bh,bh                 ; page 0
         mov     ah,2                  ; BIOS set cursor
         int     BIOS                  ; restore regs
         pop     cx
         pop     bx
         pop     ax
         ret                           ; done with set cursor 

;-----------------------------------------------------------------------------
; get cursor
; exit: dh = row, dl = column
; uses BIOS call
;-----------------------------------------------------------------------------

getCur:  push    ax                    ; save regs
         push    bx
         push    cx
         xor     bh,bh                 ; page 0
         mov     ah,3                  ; BIOS set cursor
         int     BIOS                  ; restore regs
         pop     cx
         pop     bx
         pop     ax
         ret                           ; done with get cursor 

;-----------------------------------------------------------------------------
; print space
; uses prnChr
; destroys ax
;-----------------------------------------------------------------------------

prnSpc:  mov     al,' '                ; space char
         jmp     short prnChr          ; go print it

;-----------------------------------------------------------------------------
; print back space
; uses prnChr
; destroys ax
;-----------------------------------------------------------------------------

prnBS:   mov     al,BS                 ; back space char
         jmp     short prnChr          ; go print it

;-----------------------------------------------------------------------------
; print cr/lf
; uses prnChr
; destroys ax
;-----------------------------------------------------------------------------

prnCRLF: mov     al,LF                 ; line feed
         call    prnChr                ; print it
         mov     al,CR                 ; carriage return, fall into print char

;-----------------------------------------------------------------------------
; print character
; entry: al = character, foreground color attribute stored in forClr
; destroys ah
; since BIOS has no color printing routine, this routine does a direct
;  screen write to set color, then calls BIOS to put the character on
;  the screen (so the cursor position will be properly updated)
; if prnTee flag is set, character is sent to the screen and to the file 
;  output.asc
; if gflag is set, calls graphics text printing
;-----------------------------------------------------------------------------

prnChr:  push    es                    ; save regs
         push    dx                    ; (BIOS call uses ax,bx,cx,dx)
         push    cx
         push    bx
         push    ax
         and     gflag,1               ; test for graphics mode
         je      prnChr2               ; go if off
         call    gprnChr               ; call graphics print char if on
         and     prnTee,1              ; test output tee
         je      prnChr3               ; skip write if off
         call    writTee               ; print char to file if on
prnChr3: pop     ax                    ; restore regs
         pop     bx
         pop     cx
         pop     dx
         pop     es
         ret                           ; done
prnChr2: and     prnTee,1              ; test output tee
         je      prnChr1               ; skip write if off
         call    writTee               ; print char to file if on
prnChr1: mov     ax,VIDEO              ; point es to screen
         mov     es,ax
         xor     bh,bh                 ; default page
         mov     ah,3                  ; get cursor location
         int     BIOS
         mov     al,dh                 ; get cursor row
         mov     bl,80                 ; 80 chars per row
         mul     bl                    ; 80*row
         xor     dh,dh                 ; make column a word
         add     ax,dx                 ; 80*row+col
         shl     ax,1                  ; 2*(80*row+col)
         inc     ax                    ; point to attribute
         mov     bx,ax                 ; use bx as index to screen coords
         mov     ah,bakClr             ; get color attribute
         mov     cl,4
         sal     ah,cl
         or      ah,forClr
         mov     byte ptr ES:[bx],ah   ; put it into screen attribute
         pop     ax                    ; get character back        
         mov     ah,14                 ; BIOS screen write
         int     BIOS
         pop     bx                    ; restore regs
         pop     cx
         pop     dx
         pop     es
         ret                           ; done with print char 

;-----------------------------------------------------------------------------
; print message
; entry: di points to message null terminated
;        first char in message is foreground color
; exit:  di points to byte after string (after the null terminator)
; destroys ax
; leaves foreground color as set
; uses prnStr
;-----------------------------------------------------------------------------

prnMsg:  mov     ah,[di]
         mov     forClr,ah
         inc     di

;-----------------------------------------------------------------------------
; print string
; entry: di points to string, null terminated
; exit:  di points to byte after string (after the null terminator)
; destroys ax
; uses prnChr
;-----------------------------------------------------------------------------

prnStr:  mov     al,[di]               ; get character
         inc     di                    ; bump to next one
         or      al,al                 ; work the flags
         je      prs01                 ; exit if null
         call    prnChr                ; print char
         jmp     short prnStr          ; loop back for next char
prs01:   ret                           ; done with print screen

;-----------------------------------------------------------------------------
; input character
; get an ascii character from keyboard, character NOT echoed
; ALT-Q causes exit to DOS with error code 01
; ESC - if in graphics mode, change to text mode; if in text mode, go to prompt
; uses BIOS call
;-----------------------------------------------------------------------------

inpChr:  xor  ah,ah                    ; BIOS call to get character
         int  16h                      ; char in al, scan code in ah
         cmp  al,ESCP                  ; is it escape?
         je   ipc03                    ; go if so
         or   al,al                    ; work the flags
         jne  ipc02                    ; jump if no extended code
ipc01:   cmp  ah,ALT_Q                 ; is it ALT-Q?
         jne  ipc02                    ; jump if not
         
         mov  ax,4C01h                 ; exit to DOS on ALT_Q
         int  DOS

ipc02:   ret                           ; done 

ipc03:   test gflag,1                  ; here if ESC presssed, test grafix flag
         jne  ipc04                    ; go set text mode
         mov  forClr,YELLOW            ; else set fg color to yellow
         call prnCRLF                  ; newline
         jmp  mloop                    ; go to prompt
ipc04:   call xeq140                   ; set text mode 
         mov  al,CR                    ; set char to carriage return
         ret                           ; done with input character
            
;-----------------------------------------------------------------------------
; input string
; characters are echoed as typed
; enter key terminates input
; back space erases last character
; tab inserts tabWid spaces (up to 31), won't tab past limit
; allow for a limited string length, up to 127 characters max accepted
; entry:  di points to where string will be stored
;         dl = character count limit; 0 translates to max
; exit:   dh destroyed
;         dl = actual count of chars not counting the null terminator
;         di points to next string
; uses:   inpChr, prnChr, prnSpc
;-----------------------------------------------------------------------------

inpStr:  push    bx                    ; save bx for temp storage
         mov     bx,di                 ; save copy of string pointer
         or      dl,dl                 ; test for no character count limit
         jnz     ips01                 ; skip if limit not zero
         dec     dl                    ; set limit to max
ips01:   and     dl,7Fh                ; and mask limit <= 127
ips02:   call    inpChr                ; get a character from keyboard
         cmp     al,CR                 ; is it an enter key?
         je      ips04                 ; jump if so
         cmp     al,BS                 ; is it a back space?
         je      ips05                 ; jump if so
         cmp     al,TAB                ; is it a tab?
         je      ips06                 ; jump if so
         mov     [di],al               ; store character to string
         inc     di                    ; bump pointer
         call    prnChr                ; echo character to screen
         dec     dl                    ; decrement count
         jne     ips02                 ; if not at limit, go input more
ips04:   mov     byte ptr [di],NULL    ; here at string end - store terminator
         sub     bx,di                 ; compute # chars input
         neg     bx                    ;  = -(original ptr - current ptr)
         mov     dl,bl                 ; put char count in dl
         inc     di                    ; ready for subsequent string
         pop     bx                    ; restore reg
         ret                           ; done
ips05:   cmp     di,bx                 ; here when BS hit, backed up to start?
         jbe     ips02                 ; if so, don't accept BS and go back
         call    prnBS                 ; back space
         call    prnSpc                ; erase previous char
         call    prnBS                 ; back space again
         dec     di                    ; also back up one character in string
         inc     dl                    ; and increase the count limit by 1
         jmp     short ips02           ; go back for more input
ips06:   mov     dh,tabWid             ; get tab width
         or      dh,dh                 ; is it zero?
         je      ips02                 ; go back if so
         and     dh,1Fh                ; 31 max tab
         sub     dl,dh                 ; can we do it? (count - tabWid plus?)
         jns     ips07                 ; go if so
         add     dl,dh                 ; restore count
         jmp     short ips02           ; and go back
ips07:   call    prnSpc                ; print space
         mov     byte ptr [di],al      ; store space in string
         inc     di                    ; bump pointer
         dec     dh                    ; dec space count
         jne     ips07                 ; loop back till tabWid spaces put
         or      dl,dl                 ; test count
         je      ips04                 ; finish up if zero
         jmp     short ips02           ; go back for next char if not

;-----------------------------------------------------------------------------
; input to buffer
; 80 characters max 
; sets color to inClr during input then restores color on exit
;-----------------------------------------------------------------------------

inpBuf:  push    di                    ; save regs
         push    bx
         push    dx
         mov     bl,forClr             ; save current color
         mov     bh,InpClr             ; set input color
         mov     forClr,bh
         mov     di,offset buffer      ; point to buffer
         mov     dl,80                 ; 80 characters max
         call    inpStr                ; get string from keyboard
         mov     lbuf,dl               ; set length of buffer parm
         mov     forClr,bl             ; restore color
         pop     dx                    ; restore regs
         pop     bx
         pop     di
         ret                           ; done with input buffer

;-----------------------------------------------------------------------------
;get prompt line 
;display prompt char and get asciiz string 80 chars max from user into buffer
;sets parse pointer (bp) to start address of buffer
;uses: inpBuf
;-----------------------------------------------------------------------------

getPLn:  mov     cx,100                 ; buffer+prebuf size
         mov     di,offset prebuf+3     ; point to pre-buffer  
gtpln1:  mov     byte ptr [di],SPC      ; fill with space character
         inc     di                     ; next buffer space
         loop    gtpln1                 ; loop till done
         mov     al,prmptch             ; print prompt character
         call    prnchr
         call    inpBuf                 ; get a line from user
         call    prnCRLF                ; issue new line
         mov     bp,offset buffer       ; set pointer
         ret

;-----------------------------------------------------------------------------
; integer to ascii
; convert a 16 bit signed integer into an asciiz string
; negative numbers have a '-' sign prepended
; entry:  ax = integer
; exit:   asciiz stored in intIO string, right justified
;         di points to string
;-----------------------------------------------------------------------------

int2asc:  push   dx                     ; save regs
          push   cx
          push   bx
          mov    di,offset intIO+5      ; store integer right just. in buffer
          push   di                     ; save lsd of integer
          mov    bl,' '                 ; blank to clear buffer & assume + sign
          mov    cx,6                   ; count of bytes to clear
i2a01:    mov    [di],bl                ; clear buffer to blanks
          dec    di  
          loop   i2a01
          pop    di                     ; get lsd back
          push   ax                     ; save integer
          or     ax,ax                  ; work flags
          jns    i2a02                  ; go if plus
          neg    ax                     ; make positive if not
          mov    bl,'-'                 ; and set sign flag to minus
i2a02:    mov    cx,10                  ; base ten dividend
i2a03:    sub    dx,dx                  ; zero remainder
          div    cx                     ; integer/10
          add    dl,'0'                 ; remainder converted to ascii
          mov    [di],dl                ; store digit
          dec    di                     ; drop pointer
          or     ax,ax                  ; check for int zero (all converted)
          jne    i2a03                  ; continue conversion if not
          mov    [di],bl                ; store sign
          pop    ax                     ; integer back
          pop    bx                     ; restore registers
          pop    cx
          pop    dx
          ret                           ; done with int2asc

;-----------------------------------------------------------------------------
; print integer formatted
; ifmt holds field width
; uses: asc2int, intStr, prnStr 
;-----------------------------------------------------------------------------

prnInt:  push    ax                     ; save regs
         push    bx
         push    cx
         push    di
         call    intStr                 ; convert integer to i/o string
         call    prnStr                 ; go print the field
         pop     di                     ; restore regs
         pop     cx
         pop     bx
         pop     ax
         ret                            ; done with print integer                  

; convert integer to I/O string

intStr:  call    int2asc                ; convert integer to ascii
         mov     di,offset preIio       ; blank area before intIO area
         push    di                     ; save pointer to preIio area                
         mov     cx,9                   ; nine bytes to clear
pri01:   mov     byte ptr [di],' '      ; store a blank
         inc     di                     ; bump up
         loop    pri01                  ; loop back till preIio blanked
         mov     bl,ifmt                ; get integer field width
         and     bl,0Fh                 ; limit it to 15 spaces
         xor     bh,bh                  ; make it a word
         neg     bx                     ; make it minus
         add     bx,15                  ; 15 - width
         pop     di                     ; get address of pre area
         add     di,bx                  ; address+width-15 = new pointer
         ret                            ; done with integer to i/o string

;-----------------------------------------------------------------------------
; put formatted integer to pad
; ifmt holds field width, lpad is current pad size
; uses asc2int, intStr, movStr
; sets cy if pad exceeded
;-----------------------------------------------------------------------------

int2pad: push    bx
         push    cx
         push    di
         call    intStr                 ; convert integer to i/o string
         mov     al,lpad                ; length of pad
         xor     ah,ah                  ; make it a word
         push    ax                     ; save it
         mov     ah,ifmt                ; get int format
         and     ah,0Fh                 ; must be <= 15
         add     al,ah                  ; new length
         cmp     al,80                  ; too big?
         ja      pint02                 ; go if so
         mov     lpad,al                ; new pad length
         pop     si                     ; length back
         add     si,offset pad          ; make it an address
         call    movStr                 ; move it to pad
         clc                            ; clear cy flag                            
pint01:  pop     di                     ; restore regs
         pop     cx
         pop     bx
         pop     ax
         ret                            ; done with put integer to pad             
pint02:  pop     ax                     ; clear stack
         stc                            ; set cy for error
         jmp     short pint01

;-----------------------------------------------------------------------------
; print integer formatted to 4 places
; entry: integer in ax
; uses normal foreground color
;-----------------------------------------------------------------------------

prnI4:   push    bx                     ; save reg
         mov     bl,iFmt                ; get current int format
         push    bx                     ; save it on the stack
         mov     iFmt,4                 ; set int format to 4 places
         call    prnInt                 ; print the integers
         pop     bx                     ; restore the format and color
         mov     iFmt,bl
         pop     bx                     ; restore reg
         ret                            ; done with print inetger 4 spaces
 
;-----------------------------------------------------------------------------
; is it a digit?
; entry:  char in bl
; exit:   carry clear if bl a decimal digit, carry set if not
;-----------------------------------------------------------------------------

isDig:   cmp     bl,'0'                 ; is it less than 0?
         jb      isd02                  ; go if so
         cmp     bl,'9'                 ; is it greater than 9?
         ja      isd01                  ; go if so
         and     bl,0Fh                 ; clear carry
         ret                            ; done
isd01:   stc                            ; not a number, set carry
isd02:   ret                            ; done with isDig

;-----------------------------------------------------------------------------
; ascii to integer
; convert an asciiz string into a 16 bit signed integer
; entry:  asc2int - string to be converted in token
;         asc2ina - string to be converted pointed to by si
; exit:   ax contains the integer
;         numbers are mod 16 bits
;         di is -1 if NAN (not-a-number, e.g., string contains a non-numeric)
;         otherwise, if the integer is good, di is set to zero
; patch:  added code to convert a hex string to integer. E.g., 7fffh = 32767
;         token must end in an h oh H to be a hex number
;-----------------------------------------------------------------------------

asc2int: mov     si,offset token        ; set si to point to token

; call here if user sets si to point to token or string

asc2ina: push    bx                     ; save registers
         push    cx
         push    dx
         xor     ax,ax                  ; start off with integer zero
         mov     bx,si                  ; set bx to token address
         add     bx,cx                  ; point to last char of token
         dec     bx
         mov     dl,byte ptr [bx]       ; get last char of token
         cmp     dl,'H'                 ; is token possibly a hex #?
         je      a2i09                  ; go if so
         mov     di,ax                  ; di holds sign, assume plus
         mov     cx,10                  ; base 10 multiplier
a2i01:   mov     bl,[si]                ; get a character from buffer
         inc     si                     ; bump up to next char
         or      bl,bl                  ; work flags
         je      a2i05                  ; go if terminator
         cmp     bl,' '                 ; is it a leading blank?
         je      a2i01                  ; if so, get next char
         cmp     bl,'+'                 ; is it a plus sign?
         je      a2i01                  ; if so, ignore (assumed plus)
         cmp     bl,'-'                 ; is it a minus?
         jne     a2i03                  ; go if not
         dec     di                     ; set sign flag to minus
         jmp     short a2i01            ; and loop back for more 
a2i02:   mov     bl,[si]                ; get another char from string
         inc     si                     ; bump up pointer
         or      bl,bl                  ; work flags
         je      a2i06                  ; go if terminator
a2i03:   call    isDig                  ; check if number decimal
         jc      a2i05                  ; go set NAN and exit if not
         mul     cx                     ; integer*10, shift left decimally
         xor     bh,bh                  ; make number a word
         add     ax,bx                  ; add it to shifted integer
         jmp     a2i02   
a2i05:   mov     di,-1                  ; here if NAN, set flag
         xor     ax,ax                  ; zero integer, carry flag
         jmp     short a2i08            ; go finish up
a2i06:   or      di,di                  ; proper exit here - test sign flag
         jns     a2i07                  ; go if plus
         neg     ax                     ; negate if minus
a2i07:   xor     di,di                  ; clear NAN flag, and carry
a2i08:   pop     dx                     ; restore registers
         pop     cx
         pop     bx
         ret                            ; done with asc2int

a2i09:   mov     cx,16                  ; base 16 multiplier
a2i10:   mov     dl,byte ptr [si]       ; get digit
         cmp     dl,'H'                 ; all processed?
         je      a2i07                  ; go to good number exit if so
         sub     dl,30h                 ; make it non-ascii
         js      a2i05                  ; go to bad number exit if < 0
         cmp     dl,9                   ; is it a dec range digit?
         jle     a2i11                  ; go if so
         sub     dl,7                   ; next good digit is hex A
         cmp     dl,10                  ; maybe between 9 and A?
         jl      a2i05                  ; go to bad number exit if so
         cmp     dl,16                  ; ok up thru F
         jge     a2i05                  ; go to bad number if not
a2i11:   push    dx
         mul     cx                     ; got a good digit, multiply int by 16
         pop     dx
         xor     dh,dh                  ; make digit a word
         add     ax,dx                  ; add it in
         inc     si                     ; bump pointer
         jmp     short a2i10            ; go back for next digit

;----------------------------------------------------------------------------- 
; prnFlt 
; print a floating point number
; entry:  fp number in ST(0)
; uses flt2dec, putFlt, prnFmt, exp10
; enter at flt2pad to append float string to pad
;-----------------------------------------------------------------------------

PrnFlt:  mov     pflag,0                ; set print to screen flag
         jmp     short prnfn            ; go print float
flt2pad: mov     pflag,1                ; set put to pad flag

prnfn:   push    cx                     ; save regs
         push    dx
; comment out from here if non-rounding on float print is wanted
         FTST                           ; test fp
         FSTSW   temp2                 
         FWAIT
         mov     al,fmtDec              ; get # decimals, say 6
         xor     ah,ah                  ; make it a word
         neg     ax                     ; make it negative, -6
         mov     temp,ax                ; put it on float stack
         FILD    temp
         call    exp10                  ; 10^-6,X
         FLD     fp5                    ; 0.5,10^-6,X
         FMUL                           ; .5E-6,X
         and     byte ptr temp2+1,01h   ; is fp number minus?
         jne     pnf01                  ; go if so
         FADD                           ; X rounded if +
         jmp     short pnf02
pnf01:   FSUB                           ; X rounded if -
; to here
pnf02:   call    flt2dec                ; make BCD
         mov     cx,20                  ; 20 chars max = 18 digits + sign + dp              
         call    putFlt                 ; put number in buffer
         call    prnfmt                 ; print buffer formatted
         pop     dx                     ; restore regs
         pop     cx
         ret                            ; done with prnFlt
      
;-----------------------------------------------------------------------------
;prnFmt 
; print float formatted
; uses the fmtwid (format width) and fmtdec (format decimals) bytes
; entry:  fp string in i/o buffer
; exit:   formatted string in the buffer
;         dx points to the string
; uses:   prnStr
;----------------------------------------------------------------------------- 

prnFmt:  push    ax                     ; save all regs
         push    bx
         push    cx
         push    dx
         push    si
         push    di
         mov     ax,'  '                ; blank guard bytes
         mov     word ptr preFio+18,ax
         mov     bl,fmtdec              ; set bx=# decimals
         xor     bh,bh 
         mov     si,offset fltIO        ; si points to i/o string
         mov     cx,40                  ; up to 40 chars
prnf01:  mov     al,[si]                ; get a char from i/o string
         cmp     al,'*'                 ; is it a star?
         jne     prnf15                 ; go if not...
         mov     dx,offset fltIO        ; ...else set up dx as pointer
         mov     bx,dx                  ; compute end of stars
         mov     al,fmtwid              ;  =offset iostr+fmtwid...
         xor     ah,ah
         add     bx,ax
         inc     bx                     ; ...plus 2
         inc     bx
         mov     byte ptr [bx],0        ; put the terminator
         jmp     prnf06                 ; print i/o string
prnf15:  cmp     al,'.'                 ; if not a star, maybe a dec. pt.?
         je      prnf02                 ; go if dec. pt. found
         inc     si                     ; else bump si...
         loop    prnf01                 ; ...and look some more
prnf02:  inc     si                     ; found the decimal point - bump ptr
         mov     dx,si                  ; and save in dx
         add     dx,bx                  ; bump dx to end of fraction field
         mov     di,dx                  ; ...and put it in di
         mov     byte ptr [di],0        ; put terminator at end of fraction
prnf03:  mov     al,[si]                ; now begin a search for the putflt term.
         or      al,al                  ; ...scanning from dec pt+1 to term loc
         je      prnf04                 ; when term found, jump
         inc     si                     ; else inc si...
         jmp     prnf03                 ; ...and scan some more
prnf04:  cmp     si,dx                  ; is this the terminator we put?
         jae     prnf55                 ; go if yes
         mov     al,'0'                 ;   otherwise its putflt's, so 0 fill
prnf05:  mov     [si],al                ; from here to our term
         inc     si                     ; bump si
         cmp     si,dx                  ;   and test for fill complete
         jne     prnf05                 ; keep filling if not
prnf55:  mov     cl,fmtwid              ; get format width
         xor     ch,ch                  ; make it a word
         sub     dx,cx                  ; and back up dx so it points to...
         dec     dx                     ; ...beginning of new field - bump back
         dec     dx                     ; ...two extra to allow for sign & '.'
prnf06:  mov     di,dx                  ; set di for prnStr call
         test    pflag,1                ; test append to pad flag
         jne     prnf08                 ; go if append to pad
         call    prnStr                 ; print the number 
prnf07:  pop     di                     ; restore regs
         pop     si
         pop     dx
         pop     cx
         pop     bx
         pop     ax
         ret                            ; done with prnFmt

prnf08:  mov     al,lpad                ; get length of pad
         xor     ah,ah                  ; make it a word
         push    ax                     ; save it
         mov     ah,fmtwid              ; format width
         and     ah,0Fh                 ; width <= 15
         add     al,ah                  ; add to pad length
         add     al,2                   ; allow for sign and dp
         cmp     al,80                  ; pad exceeded?
         ja      prnf09                 ; go if so
         mov     lpad,al                ; update pad length
         pop     si                     ; old pad length
         add     si,offset pad          ; make it an address
         call    movStr                 ; mov float to pad
         clc                            ; clear exceed flad
         jmp     short prnf07           ; clean up and go home
prnf09:  pop     ax                     ; clear stack
         stc                            ; exceed exceed flag
         jmp     short prnf07           ; clean up and go home

;-----------------------------------------------------------------------------
; putFlt
; put a floating point number into i/o buffer
; entry:  ST(0) = integer significand
;         dx = exponent
;         cx = total digits to output, 2<=cx<=20 
; exit:   puts either fp number as ascii in buffer or stars if out of range
; uses:   putchr, putzer, putper, 8087 coprocessor
;-----------------------------------------------------------------------------

putflt:  push    ax                     ; save all regs
         push    bx
         push    cx
         push    dx
         push    bp
         push    si
         push    di
         mov     fpbufptr,offset fltio  ; set up string pointer
         neg     dx
         mov     si,cx                  ; si = digit limit
         mov     di,-2                  ; set width+2 (room 4 sign & dec. pt.)
         FTST                           ; test ST(0)
         FSTSW   npustat                ; store stats
         FWAIT
         FBSTP   fltint                 ; store bcd significand
         FWAIT
         and     byte ptr npustat+1,40h ; ST(0)=0?
         je      ptf01                  ; go if not
         mov     al,' '                 ; ST(0) is all zeros, blank sign
         jmp     short ptf04            ;   and skip format fit check
ptf01:   mov     bp,offset fltint+9     ; bp points to mantissa
         mov     al,ds:[bp]             ; get sign byte
         and     al,7Fh                 ; mask away sign bit
         jne     ptf02                  ; and put stars if any other bits set
         mov     al,fmtwid              ; find allowable # integers
         sub     al,fmtdec              ;   = width - #fraction digits
         xor     ah,ah                  ; make it a word
         add     ax,dx                  ; add (dx-18)
         sub     ax,18                 
         inc     ax                     ; *** patch                    
         jns     ptf03                  ; OK, go convert
ptf02:   mov     al,'*'                 ;   else put stars
         inc     di                     ; bump up star count
         call    putchr                 ; put a star in buffer
         cmp     si,di                  ; buffer filled?
         jne     ptf02                  ;   no, loop back
         jmp     ptf11                  ;   yes, go finish up                 
ptf03:   mov     al,' '                 ; assume a space for positive numbers
         mov     cl,ds:[bp]             ; get sign byte again
         or      cl,cl                  ; work the flags
         je      ptf04                  ; go if plus
         mov     al,'-'                 ; assumption wrong, it's minus
ptf04:   call    putchr                 ; put sign
         xor     di,di                  ; di will count digits as output, set=0
         mov     cx,di                  ; cx will point to nyble=0
         mov     ah,ch                  ; ah is leading zero flag=0
         mov     bx,dx                  ; bx will hold initial exponent
         dec     bp                     ; point to first significand byte
         and     byte ptr npustat+1,40h ; ST(0)=0?
         jne     ptf05                  ;   go if mantissa 0
         cmp     dx,17                  ; check for -1 exp.
         je      ptf45                  ;   go if it is
         cmp     dx,18                  ; check range 
         ja      ptf05                  ;   go put .000---
         jb      ptf07                  ;   go put nnn---.
         call    putzer                 ;   else put 0.nnn---
         inc     di                     ; count digit
         call    putper                 ; put dec. pt.
         jmp     short ptf07            ; go load nyble
ptf45:   mov     al,ds:[bp]             ; exp=-1, check for zero just before dec. pt.
         and     al,0F0h                ; mask left nyble
         jne     ptf07                  ; not this case
         inc     ah                     ; don't blank it
         jmp     short ptf07            ; go load nyble
ptf05:   call    putzer                 ; exp=0 (or mant. 0), put a zero
         inc     di                     ; count digit
         call    putper                 ; now put the dec. pt.
ptf06:   call    putzer                 ; put zeros till exp. aligned with 18
         inc     di                     ; count digit
         cmp     si,di                  ; digit limit met?
         je      ptf11                  ;   done if so
         dec     dx                     ; dec exponent
         cmp     dx,18                  ; at 18 yet?
         jne     ptf06                  ;   nope, keep on putting zeroes
ptf07:   mov     al,ds:[bp]             ; load a nyble, part of a bcd pair
         and     cx,1                   ; if cx=0, left nyble
         jne     ptf08                  ; if cx=1, right nyble
         shr     al,1                   ; shift one digit to the right
         shr     al,1                   ;   using 8086 shifts (1 bit only)
         shr     al,1
         shr     al,1
ptf08:   and     al,0Fh                 ; mask the desired digit
         or      ah,al                  ; OR with leading zero flag - stay zero
         je      ptf09                  ;  ... if still lead zero, don't put it
         add     al,'0'                 ; not a leading 0 , make ascii
         call    putchr                 ; ... and put it
ptf09:   sub     bp,cx                  ; dec string ptr if cx=1 (right nyble)
         inc     cx                     ; next nyble
         inc     bx                     ; count this digit
         dec     dx                     ; dec exponent
         cmp     bx,18                  ; are we to dec. pt. yet?
         jne     ptf10                  ;   go if not
         call    putper                 ; else put the period
ptf10:   inc     di                     ; count digit
         cmp     si,di                  ; digit limit met?
         je      ptf11                  ;   go if so
         cmp     bp,offset fltint-1     ; have all digits been put?
         jne     ptf07                  ;   loop for more if not
ptf11:   xor     al,al                  ; put string terminator at buffer's end
         call    putchr
         pop     di                     ; restore all regs
         pop     si
         pop     bp
         pop     dx
         pop     cx
         pop     bx
         pop     ax
         ret                            ; done with putflt

putper:  mov     al,'.'                 ; put a period
         call    putchr
         inc     ah                     ; don't need to blank zeros anymore
         ret                            ; done with putper

putzer:  mov     al,'0'                 ; put a zero

putchr:  push    bx                     ; put a char to i/o str - save bx 1st
         mov     bx,fpbufptr            ; set bx to pointer in buffer
         mov     [bx],al                ; put the char
         inc     bx                     ; increment pointer
         mov     fpbufptr,bx            ; save pointer back
         pop     bx                     ; restore bx
         ret                            ; done with putchr/putzer

;-----------------------------------------------------------------------------
; putExp
; put a floating point number in exponential format to i/o buffer
; entry: ST(0) = integer significand
;        dx = exponent (negated)
;        expdec = # mantissa decimals
;        expdig = # exponent digits
; exit: puts an exponential form float +/-D.dddd...E+/-nnn 
; uses: putchr, putper, putzer, 8087 coprocessor
;-----------------------------------------------------------------------------

putExp:  push    ax                      ; save regs
         push    bx
         push    cx
         push    dx
         push    si
         push    di
         mov     cl,expdec              ; set # mantissa decimals
         xor     ch,ch
         mov     fpbufptr,offset fltio  ; set up string pointer
         FTST                           ; test ST(0)
         FSTSW   npustat                ; store status
         FWAIT
         FBSTP   fltint                 ; store bcd significand
         FWAIT
         and     npustat+1,40h          ; ST(0)=0?
         je      pte50                  ; go if not
         mov     al,' '                 ; put blank for sign
         call    putchr
         call    putzer                 ; put '0.'
         call    putper
pte45:   call    putzer                 ; put cx zeros
         loop    pte45
         mov     dx,-17                 ; set exponent
         jmp     short pte35            ; go do exponent
pte50:   inc     cx                     ; allow for digit to left of dp
         mov     al,' '                 ; blank sign 
         mov     si,offset fltint+9     ; point to mantissa
         mov     bl,byte ptr [si]       ; get sign byte
         and     bl,80h                 ; mask sign bit
         je      pte01                  ; go if plus
         mov     al,'-'                 ; else sign is minus
pte01:   call    putchr
         dec     si                     ; point to first significand byte
         xor     bx,bx                  ; nybble flag, 0=left, 1=right
         mov     di,bx                  ; zero digit count
pte02:   mov     al,byte ptr [si]       ; load a nybble, part of a BCD pair
         and     bx,1                   ; test it
         jne     pte03                  ; go if right
         shr     al,1                   ; shift 1 digit to right
         shr     al,1                   ;  using 8086 shifts (1 bit only)
         shr     al,1
         shr     al,1
pte03:   and     al,0Fh                 ; mask the desired digit
         add     al,'0'                 ; make it ascii
         or      di,di                  ; first time thru?
         jne     pte33                  ; go if not
         cmp     al,'0'                 ; first digit zero?
         jne     pte33                  ; go if not
         inc     bx                     ; next nybble
         dec     dx                     ; bump exponent
         jmp     short pte02            ; loop back for more digits   
pte33:   cmp     di,1                   ; first digit put?
         jne     pte34                  ; go if not
         push    ax                     ; if so, put dec pt
         mov     al,'.'
         call    putchr
         pop     ax
pte34:   call    putchr                 ; put it in buffer
         sub     si,bx                  ; decrement pointer if left nybble
         inc     bx                     ; bump nybble pointer
         inc     di                     ; bump digit count
         loop    pte02                  ; loop till all digits stored
pte35:   mov     al,'E'                 ; print 'E'
         call    putchr
         mov     al,'+'                 ; assume exp plus
         add     dx,17                  ; adjust exponent
         jns     pte04                  ; go if plus
         mov     al,'-'                 ; exp minus
         neg     dx                     ; make it plus
pte04:   call    putchr                 ; put sign
         mov     ax,dx                  ; get adjusted exponent
         mov     bx,10                  ; divisor
         mov     cl,expdig              ; # exponent digits
         xor     ch,ch
pte05:   xor     dx,dx                  ; prepare to divide
         div     bx                     ; exp/10
         add     dl,'0'                 ; make remainder ascii
         push    dx
         loop    pte05                  ; do all 3 digits
         mov     cl,expdig              ; # exponent digits
pte06:   pop     ax
         call    putchr
         loop    pte06     
         mov     al,NULL                ; store null terminator
         call    putchr
         pop     di                     ; restore regs
         pop     si
         pop     dx
         pop     cx
         pop     bx
         pop     ax
         ret                            ; done iwth putEXp       

;----------------------------------------------------------------------------- 
; print a floating point number in exponential form
; entry: fp number in ST(0)
; uses:  flt2dec, putExp, exp10
; enter at exp2pad to append float exponential string to pad
;-----------------------------------------------------------------------------

prnExp:  mov     pflag,0                 ; set print to screen flag
         jmp     short prnen             ; go print float
exp2pad: mov     pflag,1                 ; set put to pad flag

prnen:   
; comment out from here if non-rounding on float print is wanted
         FTST                            ; test fp
         FSTSW   temp2
         FWAIT         
         test    byte ptr temp2+1,40h    ; is fp number 0?
         jne     prne04                  ; don't round if so
         FLD     ST(0)                   ; dupe 
         call    flt2dec                 ; convert to dec
         FBSTP   fltint                  ; store bcd significand
         mov     al,byte ptr fltint+8    ; get lsb
         and     al,0F0h                 ; mask lsd
         jne     prne25                  ; go if not zero
         dec     dx                      ; bump exponent if zero
prne25:  mov     al,expdec               ; get # decimals
         xor     ah,ah                   ; make it a word
         sub     ax,dx                   ; adjust for exponent, fd-dx
         sub     ax,17                   ; fd-dx-17
         neg     ax                      ; dx+17-fd=p
         mov     temp,ax                 ; put it on 8087 stack
         FILD    temp                    ; -p,x
         call    exp10                   ; 10^-p,x
         FLD     fp5                     ; 0.5,10^-p,x
         FMUL                            ; .5E-p,x
         and     byte ptr temp2+1,01h    ; is fp number minus?
         jne     prne03                  ; go if so
         FADD                            ; x rounded if +
         jmp     short prne04
prne03:  FSUB                            ; x rounded if -       
; to here
prne04:  call    flt2dec                 ; convert to dec
         call    putExp                  ; put float in float i/o area         
         mov     di,offset fltio         ; point to float i/o area
         test    pflag,1                 ; test for append to pad
         jne     prne01                  ; go if so
         call    prnStr                  ; print it
         ret                             ; done with prnExp
         
prne01:  mov     al,lpad                 ; get length of pad
         xor     ah,ah                   ; make it a word
         push    ax                      ; save it
         mov     ah,expdec               ; get # mantissa decimals
         add     ah,expdig               ; add # exponent digits
         add     ah,5                    ; add 5 for sign,1st s.d.,dec pt,E,E sign
         add     al,ah                   ; adjust pad length
         cmp     al,80                   ; pad exceeded?
         ja      prne02                  ; go if so
         mov     lpad,al                 ; update pad length
         pop     si                      ; get old pad length
         add     si,offset pad           ; make it an address
         call    movStr                  ; move exponenetial float to pad
         clc                             ; clear pad exceeded flag
         ret                             ; done 
prne02:  stc                             ; set carry if pad exceeded
         ret                             ; done with append to pad

;-----------------------------------------------------------------------------
; flt2dec
; convert floating point to integer and exponent
; entry: fp in ST(0)
; exit:  integer significand in ST(0)
;        exponent in dx (negated)
; uses:  exp10
;-----------------------------------------------------------------------------
flt2dec: FLD     ST(0)                  ; R,R
         FXTRACT                        ; R,exp(R),R
         FSTP    ST(0)                  ; exp(R),R
         FISUBR  precbit                ; exp(R)-prec=scale,R
         FLDL2T                         ; log2(10),scale,R
         FDIV                           ; scale/log2(10)=N,R
         FRNDINT                        ; n(N),R
         FIST    fltexp                 ; store base ten scale
         FWAIT
         call    exp10                  ; 10^N,R
         FMUL                           ; R*10^N=integer             
         mov     dx,fltexp              ; dx=exp. of 10, ST(0)=int mantissa
         neg     dx
         ret                            ; done with flt2dec

;-----------------------------------------------------------------------------
; dec2flt
; Convert integer mantissa and exponent to floating point real number
; entry:  integer mantissa in ST(0)
;         exponent in dx
; exit:   floating point number in ST(0)
; uses:   exp10
;-----------------------------------------------------------------------------

dec2flt: mov     fltexp,dx              ; put exp. in memory
         or      dx,dx                  ; test sign of exp.
         jg      d2f01                  ; go if plus
         neg     fltexp                 ; make plus
d2f01:   FILD    fltexp                 ; integer exp in ST(0)
         call    exp10                  ; 10^N
         or      dx,dx                  ; test sign again
         jl      d2f03                  ; go if minus
         FMUL                           ; R*10^N
         ret                            ; float in ST(0) - done
d2f03:   FDIV                           ; R/10^N
d2f04:   ret                            ; float in ST(0) - done with dec2flt

;-----------------------------------------------------------------------------
; asc2dec 
; get a floating point number 
; entry:  asc2dec - convert string in token
;         asc2dca - convert string pointed to by si 
; exit:   integer mantissa in ST(0), exponent in dx
;         di set to -1 if NAN, 0 if good fpn
; uses:   getchr         
;-----------------------------------------------------------------------------

asc2dec:  mov    si,offset token        ; point to token string

; call here if user sets si to point to token or string

asc2dca:  push   ax                     ; save regs
          push   bx
          push   cx
          push   si
          mov    fpexp,0                ; zero E field exponent
          FLDZ
          FSTP   fltint
          mov    di,offset fltint+9     ; point di to end of fltint                    
          xor    bx,bx                  ; nyble tracker
          mov    cx,bx                  ; zero exponent count. & dp flag
          mov    nDigs,bl               ; clear count of actual digits
gf02:     call   getchr                 ; get a character from string
          cmp    al,' '                 ; ignore blanks
          je     gf02              
          cmp    al,'+'                 ; ignore plus sign
          je     gf04
          cmp    al,'-'                 ; test for minus sign
          je     gf03                   ; go if minus
          dec    di                     ; back up pointer to msb fltint
          jmp    short gf05             ; go process char
gf03:     mov    al,80h                 ; here if minus, set sign bit
          mov    [di],al                ;  and store in sign byte
gf04:     dec    di                     ; back up pointer to msb
gf45:     call   getchr                 ; get a char
gf05:     cmp    al,'.'                 ; dec. pt.?
          je     gf07                   ; go if found
          sub    al,'0'                 ; make it a number
          js     gf10                   ; less than 0 so no good
          cmp    al,9                   ; more than 9...
          ja     gf10                   ;  ... no good
          and    al,0Fh                 ; mask it
          inc    nDigs                  ; keeps track of number of digits
          and    bl,1                   ; test for shifting
          jne    gf06                   ; go if right nyble
          shl    al,1                   ; make left nyble
          shl    al,1                   ;  using 8086 shifts
          shl    al,1
          shl    al,1
gf06:     or     [di],al                ; put it in fltint
          sub    di,bx                  ; back up fltint pointer if left nyble
          inc    bx                     ; bump nyble tracker
          inc    cl                     ; count this digit for exponent
          jmp    short gf45             ; go back to process another digit
gf11:     mov    dl,cl                  ; set exponent - cl counts digits to '.' 
          mov    dh,0                   ; make it a word
          sub    dx,18                  ; adjust for proper exponent
          ret                           ; done with set exponent
gf07:     or     ch,ch                  ; here when '.' found, first one?
          jne    gf12                   ; go if not
          call   gf11                   ; go set exponent
          inc    ch                     ; and set flag that dec. pt. found
          jmp    short gf45             ; go back for more
gf10:     dec    si                     ; get char again 
          mov    al,[si]
          cmp    al,'e'                 ; 'e' or 'E' signals exponent
          je     gfe01
          cmp    al,'E'
          je     gfe01        
          or     al,al                  ; is non-number a null?
          jne    gf12                   ;  bad fpn if not
gf15:     cmp    nDigs,0                ;  must have processed at least 1 digit
          je     gf12                   ;  bad fpn if not
          or     ch,ch                  ; here if out of stuff,ensure exp set
          jne    gf08                   ; go if already set (ei, dec.pt. found)
          call   gf11                   ;  else set the exponent
gf08:     xor    di,di                  ; set flag (di) to good fpn 
gf13:     FBLD   fltint                 ; ST(0)=significand and dx=exponent
          add    dx,fpexp               ; update exponent   
gf14:     pop    si                     ; restore regs
          pop    cx
          pop    bx
          pop    ax
          ret                           ; done with ascii to dec (fp)
gf16:     pop    cx                     ; restore reg          
gf12:     mov    di,-1                  ; set NAN code
          jmp    short gf14             ; and go home

getchr:   mov    al,[si]                ; pull a char from input string
          inc    si                     ; bump pointer
          ret                           ; done with getchr

gfe01:    push   cx                     ; here if exponent
          mov    cl,10                  ; decimal multiplier
          mov    ch,3                   ; # exp digits
          xor    bx,bx                  ; start with bl=exp=0, bh=sign=+
          inc    si                     ; readjust pointer
          call   getchr                 ; next char
          cmp    al,'+'                 ; positive exponent?
          je     gfe02                  ; go if so (ignore)
          cmp    al,'-'                 ; negative exponent?
          jne    gfe03                  ; go if not
          dec    bh                     ; set sign minus
gfe02:    call   getchr                 ; next char
gfe03:    or     al,al                  ; null?
          je     gfe04
          sub    al,'0'                 ; ascii to number
          js     gf16                   ; <0 is non numeric
          cmp    al,9                   ; >9 is non numeric
          ja     gf16
          push   ax                     ; save num
          mov    al,bl                  ; exp
          mul    cl                     ; exp*10
          mov    bl,al                  ; to exp
          pop    ax                     ; num back
          add    bl,al                  ; exp*10+num
          dec    ch                     ; count down # exp digits
          jne    gfe02
gfe04:    or     bh,bh                  ; exp minus?
          je     gfe05                  ; go if not
          neg    bl                     ; negate if so
gfe05:    mov    fpexp,bx               ; save in temp
          or     bx,bx                  ; text exp
          jns    gfe06                  ; go if plus
          neg    bx                     ; make plus
gfe06:    cmp    bx,127                 ; test for too large
          jg     gf16                   ; no good if so
          pop    cx                     ; restore reg
          jmp    short gf15             ; go back

;-------------------------------------------------------------------------------
; FCMP - floating compare
; return z flag set if ST(1) <= ST(0)
;------------------------------------------------------------------------------

FCMP:    FCOMP                          ;compare ST(1):ST(0)
         FSTSW   npustat                ;store status
         FWAIT
         and     byte ptr npustat+1,01h ;mask sign flag
         ret                            ;return with z flag working 
 
;-------------------------------------------------------------------------------
; FREDUC - reduce angle so that -2pi < ang < +2pi 
;-------------------------------------------------------------------------------
          
FREDUC:  FLD     f2pi                   ;2*pi,ang
         FDIV                           ;ang/(2*pi)=x
         FLD     ST(0)                  ;x,x
         FRNDINT                        ;n(x),x
         FSUB                           ;f(x)
         FLD     f2pi                   ;2*pi,x
         FMUL                           ;a = angle reduced
         ret

;------------------------------------------------------------------------------
; FTAN - full tangent routine
; entry: angle in radians in ST(0), -inf < ang < +inf
; exit:  tangent of angle in ST(0), -inf < tan < +inf
;        carry set if infinity
; uses:  FCMP, 8087 coprocessor
;-----------------------------------------------------------------------------

FTAN:    FTST                           ;test ST(0)
         FSTSW   trgsgn                 ;store sign
         FWAIT
         FABS                           ;set plus
         call    FREDUC                 ;reduce angle
         FLD     f3pid2                 ;3*pi/2,a
         call    FCMP                   ;a:3*pi/2
         je      ftn01                  ; go if a<=3*pi/2
         FLD     f2pi                   ; else 2*pi,a
         FSUB                           ;a-2*pi=x where -pi/2 <= x <= +pi/2
         jmp     short ftn03            ;do tangent
ftn01:   FLD     fpid2                  ;pi/2,a
         call    FCMP                   ;a:pi/2
         je      ftn03                  ;go do tangent if a<=pi/2
         FLDPI                          ;pi,a
         FSUB                           ;a-pi=x where -pi/2 <= x <= +pi/2
ftn03:   FLD     fp5                    ;.5,x
         FMUL                           ;x/2 where -pi/2 <= x <= +pi/2
         FPTAN                          ;X,Y
         FDIV                           ;tan(x/2)=z thus -pi/4 < z < +pi/4
         FLD     ST(0)                  ;z,z
         FMUL    ST(0),ST(1)            ;z^2,z
         FLD1                           ;1,z^2,z
         FSUBR                          ;1-z^2,z
         FDIV                           ;z/(1-z^2)
         FLD     ST(0)                  ;z/(1-z^2),z/(1-z^2)
         FADD                           ;2z/(1-z^2)=tan(a)
         and     byte ptr trgsgn+1,01h  ;was a negative?
         je      ftn04                  ; go if not
         FCHS                           ; change sign
ftn04:   clc                            ;clear carry for good exit
         ret                            ;done
ftn05:   stc                            ; set carry if inf error
         ret                            ;done with FTAN

;------------------------------------------------------------------------------
; FATAN - arctan routine  
; entry: value in ST(0)  -inf <= x <= +inf  
; exit:  atan in ST(0)   -pi/2 <= atan(x) <= +pi/2
; uses:  FCMP, 8087 coprocessor
;------------------------------------------------------------------------------

FATAN:   FTST                           ;test for sign
         FSTSW   trgsgn                 ;store sign
         FWAIT
         FABS                           ;make positive, x=|a|
         FLD1                           ;1,x
         call    FCMP                   ;x<=1?
         jne     atn01                  ;go if not
         FLD1                           ;1,x
         FPATAN                         ;atan(x)
         jmp     short atn02            ;go check sign
atn01:   FLD1                           ;1,x
         FDIVR                          ;1/x=z
         FLD1                           ;1,z
         FPATAN                         ;atan(z)
         FLD     fpid2                  ;pi/2,atan(z)
         FSUBR                          ;pi/2-atan(z)=atan(x)
atn02:   and     byte ptr trgsgn+1,01h  ;check sign
         je      atn03                  ;done if positive
         FCHS                           ;else change sign
atn03:   ret                            ;atan(a) done

;------------------------------------------------------------------------------
; PFSIN - partial sine routine
; entry: angle in ST(0) in radians  -pi/2 < angle < pi/2
; exit:  sine of angle in ST(0)     -1.0 <= sin(a) <= +1.0
;------------------------------------------------------------------------------

PFSIN:   FTST                           ;test for sign
         FSTSW   sinsgn                 ;store sign
         FWAIT
         FABS                           ;make positive, x=|a|
         FLD     fp5                    ;.5,x
         FMUL                           ;x/2 (0<x<pi/4)
         FPTAN                          ;tan(x/2)=z
         FDIV
         FLD     ST(0)                  ;z,z
         FMUL    ST(0),ST(1)            ;z^2,z
         FLD1                           ;1,z^2,z
         FADD                           ;1+z^2,z
         FDIV                           ;z/(1+z^2)
         FLD     ST(0)                  ;z/(1+z^2),z/(1+z^2)
         FADD                           ;2z/(1+z^2)=sin(x)
         and     byte ptr sinsgn+1,01h  ;check sign
         je      pfsn01                 ;done if positive
         FCHS                           ;else change sign
pfsn01:  ret                            ;done with partial sine

;------------------------------------------------------------------------------
; FFCOS - full cosine routine
; entry: angle in ST(0) in radians  -inf < angle < +inf
; exit:  cosine of angle in ST(0)   -1.0 <= cos(a) <= +1.0
; uses:  FFSIN
;------------------------------------------------------------------------------

FFCOS:    FLD     fpid2                  ;pi/2,x
          FADD                           ;x+pi/2

;------------------------------------------------------------------------------
; FFSIN - full sine routine
; entry: angle in ST(0) in radians  -inf < angle < +inf
; exit:  sine of angle in ST(0)     -1.0 <= sin(a) <= +1.0
; uses:  PFSIN
;------------------------------------------------------------------------------

FFSIN:   call    FREDUC                 ;reduce angle   
         FTST                           ;test for sign
         FSTSW   trgsgn                 ;store sign
         FWAIT
         FABS                           ;make positive, x=|a|
         FLD     fpid2                  ;pi/2,x
         call    FCMP                   ;x<=pi/2?
         jne     ffsn01                 ;go if not
         call    PFSIN                  ;sin(x)
         jmp     short ffsn03           ;go set sign
ffsn01:  FLD     f3pid2                 ;3pi/2,x
         call    FCMP                   ;x<=3pi/2?
         jne     ffsn02                 ;go if not
         FLDPI                          ;pi,x
         FSUBR                          ;x->pi-x
         call    PFSIN                  ;sin(x)
         jmp     short ffsn03           ;go set sign
ffsn02:  FLD     f2pi                   ;2pi,x
         FSUB                           ;x->x-2pi
         call    PFSIN                  ;sin(x)
ffsn03:  and     byte ptr trgsgn+1,01h  ;check sign
         je      ffsn04                 ;done if positive
         FCHS                           ;else change sign
ffsn04:  ret                            ;sin(a) done     
                 
;-----------------------------------------------------------------------------
; log10
; compute log base 10 (x)
; entry: x in ST(0)
; exit:  log10(x) in ST(0)
;        cy set if x <= 0.0
; uses:  8087 coprocessor
;-----------------------------------------------------------------------------

log10:   FTST
         FSTSW   npustat                ; store stats
         FWAIT
         and     byte ptr npustat+1,41h ; ST(0) minus or zero?
         jne     l10a                   ; go if so - math error
         FLD1                           ; 1,x
         FLDL2T                         ; y=log2(10),1,x
         FDIV                           ; y=1/log2(10),x
         FSTP    ST(2)                  ; x,y                           
         FYL2X                          ; log10(x)
         clc                            ; clear carry
         ret                            ; done
l10a:    stc                            ; set carry if x out of range
         ret                            ; done with log10

;-----------------------------------------------------------------------------
; exp10
; compute 10^x
; entry:  x in ST(0) 
; exit:   10^x in ST(0)
; uses:   exp2, 8087 coprocessor
;----------------------------------------------------------------------------- 

exp10:   FLDL2T                         ; log2(10),x
         FMUL                           ; 10^x=2^(x*log2(10))

;-----------------------------------------------------------------------------
; exp2
; compute 2^x
; entry:  x in ST(0)
; exit:   2^x in ST(0)
; uses:   8087 coprocessor
;-----------------------------------------------------------------------------
  
exp2:    FLD     ST(0)                  ; x,x
         FRNDINT                        ; n(x),x
         FXCH                           ; x,n(x)
         FSUB    ST(0),ST(1)            ; f(x),n(x)
         FTST                           ; set condx codes
         FSTSW   npustat                ; store condx
         FWAIT
         and     byte ptr npustat+1,1   ; test for neg.
         jne     exp01                  ; frac part is minus
         F2XM1                          ; 2^f(x)-1,n(x)
         FLD1                           ; 1,2^f(x)-1,n(x)
         FADD                           ; 2^f(x),n(x)
         FSCALE                         ; 2^x,n(x)
         FSTP    ST(1)                  ; 2^x
         ret                            ; done
exp01:   FABS                           ; +f(x),n(x)
         F2XM1                          ; 2^f(x)-1,n(x)
         FLD1                           ; 1,2^f(x)-1,n(x)
         FADD                           ; 2^f(x),n(x)
         FXCH                           ; n(x),2^f(x)
         FLD1                           ; 1,n(x),2^f(x)
         FSCALE                         ; 2^n(x),1,2^f(x)
         FDIVRP  ST(2),ST(0)            ; 1,2^x
         FSTP    ST(0)                  ; 2^x
         ret                            ; done with exp2

;-----------------------------------------------------------------------------
; print error message if error flag is on else jump to user defined error 
; entry: di points to error description
; code snippet at point of error plus +/- 20 chars is printed in blue
; offending word is pointed to and error description printed in red
; destroys ax
; assumes stack is dirty. Jumps to mloop on exit where stack reset MUST occur
; ----------------------------------------------------------------------------

error:   cmp     errflg,0		; check for error-on
         jne     er05                   ; go if so
         mov     ah,forClr              ; save foreground color
         push    ax                     ; on stack
         push    di                     ; save error message pointer
         call    prnCRLF                ; new line
         mov     di,offset codMsg       ; point di to 'Code: '
         call    prnMsg                 ; print it
er01:    mov     di,nxtAdr              ; save pointer to code
         sub     di,20 
er02:    mov     cx,40                  ; for 40 bytes
; print the code but change cr, lf, and spc to a small dot (char 250)
er22:    mov     al,byte ptr[di]        ; get code char
         or      al,al                  ; test for null
         je      er28                   ; done printing when found
         call    iswht                  ; is it a cr lf or spc?
         jne     er24                   ; replace them if so
         mov     al,250                 ; with a small dot
er24:    call    prnChr                 
         inc     di                     ; bump to next char
         loop    er22                   ; keep going
er28:    call    prnCRLF                ; print new line
         mov     cx,25                  ; print 25 spaces
er03:    call    prnSpc                 ; (to align up arrow)
         loop    er03
         mov     di,nxtAdr
         cmp     byte ptr [di],NULL
         jne     er04
         call    prnSPC
er04:    mov     al,UPARW               ; print up arrow
         call    prnChr
         call    prnSPC
         pop     di                     ; di back - points to error msg
         call    prnMsg                 ; print the error description
         mov     al,'!'                 ;  and an exclamation point
         call    prnChr
         pop     ax                     ; get saved data back
         mov     forClr,ah              ; restore foreground color
         call    prnCRLF                ; print new line
         call    warm                   ; warm start
         jmp     mloop                  ; done with error
er05:    mov     si,offset errjmp       ; move tag to token
         mov     di,offset token
         push    di			; save address of tag
         mov     ah,8
er06:    mov     al,[si]
         mov     [di],al
         inc     si
         inc     di
         dec     ah
         jne     er06
         pop     di			; tag address back

         call    srchLbl		; find the tag
         cmp     si,-1                  ; not found?
         je      er07                   ; go if no tag found
         mov     bp,si                  ; set parse address

;         call    chkWrd			; does word exits?
;         or      dx,dx                  ; check error code
;         js      er07			; go if no to error routine
         ret                            ; done with error 

er07:    mov     errflg,0               ; here if no error word - set error off
         jmp     xq37a			; go print no label msg

;------------------------------------------------------------------------------
; print counted string
; entry: di points to message
;        cl is count of chars to print
;------------------------------------------------------------------------------

prnSN:   mov     al,[di]                ; get char
         cmp     al,NULL                ; end of string?
         je      psn01                  ; done if so
         inc     di                     ; bump pointer
         call    prnChr                 ; print char
         dec     cl                     ; decrement char count
         jne     prnSN                  ; print more if not zero
psn01:   ret                            ; done with print counted string

;-----------------------------------------------------------------------------
; upper case
; entry: di points to string
; string converted to all upper case
; non alphabetic chars ignored
; di is preserved 
;-----------------------------------------------------------------------------

upper:   push    di                     ; save regs 
         push    ax 
up01:    mov     al,[di]                ; get first char
         or      al,al                  ; test for null
         je      updone                 ; done if so
         cmp     al,'a'                 ; test for 'a'
         jae     up03                   ; may be a lower case - go                  
up02:    inc     di                     ; else bump to next char
         jmp     up01                   ; keep scanning 
up03:    cmp     al,'z'                 ; here if maybe l.c., test for 'z'
         ja      up02                   ; not l.c. alphabetic - scan more
         sub     al,32                  ; it's a hit, change to u.c.
         mov     [di],al                ; save it back
         jmp     short up02             ; keep scanning
updone:  pop     ax                     ; restore regs
         pop     di
         ret                            ; done with upper case

;-----------------------------------------------------------------------------
; lower case
; entry: di points to string
; string converted to all lower case
; non alphabetic chars ignored
; di is preserved 
;-----------------------------------------------------------------------------

lower:   push    di                     ; save regs 
         push    ax 
lw01:    mov     al,[di]                ; get first char
         or      al,al                  ; test for null
         je      lwdone                 ; done if so
         cmp     al,'A'                 ; test for 'A'
         jae     lw03                   ; may be a upper case - go                  
lw02:    inc     di                     ; else bump to next char
         jmp     lw01                   ; keep scanning 
lw03:    cmp     al,'Z'                 ; here if maybe l.c., test for 'Z'
         ja      lw02                   ; not u.c. alphabetic - scan more
         add     al,32                  ; it's a hit, change to l.c.
         mov     [di],al                ; save it back
         jmp     short lw02             ; keep scanning
lwdone:  pop     ax                     ; restore regs
         pop     di
         ret                            ; done with upper case


;-----------------------------------------------------------------------------
; get next token 
; scan buffer string for white space delimited token 
; entry: bp points to start of scan
; exit:  cx = number of chars in the token
;        di points to token
;        carry set if null terminator found or out of space
;-----------------------------------------------------------------------------

getToke: mov     di,offset token        ; set di to point to token
         push    di                     ; save it
         xor     cx,cx                  ; zero count of chars in token
gtkn01:  cmp     bp,maxPAdr             ; out of program space?
         ja      gtkn04                 ; go if so
         mov     al,DS:[bp]             ; get next char
         or      al,al                  ; test for null
         je      gtkn04                 ; go set carry and done
         inc     bp                     ; bump scan pointer
         call    isWht                  ; is the char white space?
         je      gtkn01                 ; keep looking for first non wht spc
         inc     cx                     ; else it's a token char, bump count
         jmp     short gtkn03           ; go store token char
gtkn02:  mov     al,DS:[bp]             ; get next char 
         cmp     bp,maxPAdr             ; out of program space?
         ja      gtkn05                 ; go if so 
         or      al,al                  ; test for null
         je      gtkn05                 ; go to normal exit
         inc     bp                     ; bump scan pointer           
         call    isWht                  ; test for white space
         je      gtkn05                 ; token found if so
         inc     cx                     ; else count this char
gtkn03:  mov     [di],al                ; store it in token
         inc     di                     ; bump token pointer 
         jmp     short gtkn02           ; and keep scanning
gtkn04:  stc                            ; here if null or end found - set carry
gtkn05:  mov     byte ptr [di],NULL     ; normal exit - store null term.
         mov     nxtAdr,bp              ; save next scan address
         pop     di                     ; restore token pointer
         ret                            ; done with get token

;-----------------------------------------------------------------------------
; test for white space
; entry: character in al
; on exit, zero flag set if space, CR, LF, or tab
;-----------------------------------------------------------------------------

isWht:   cmp     al,SPC                 ; test for blank
         je      isw01                  ; end of token if so
         cmp     al,CR                  ; test for CR
         je      isw01                  ; consider it a blank
         cmp     al,LF                  ; test for LF
         je      isw01                  ; consider it a blank
         cmp     al,TAB                 ; test for tab
isw01:   ret

;-----------------------------------------------------------------------------  
; parse words in buffer 
; each token is scanned and determined to be a valid number or a FLIP word
; token is converted to uppercase
; if token is valid then it will be executed
; if token is invalid then an error message is printed 
; entry: bp is start of scan
; uses getToke, srchDct, Xeq
; parse stops when null terminator is found (bp will point to it), cy set
;-----------------------------------------------------------------------------

parse:   call    getToke                ; get a token
         jc      par04                  ; done if null (carry set)
         call    upper                  ; convert token to upper case
         cmp     byte ptr [di],TILDE    ; ignore ~ - its a label
         je      parse
         call    asc2int                ; try to convert to integer
         inc     di                     ; test flag
         je      par01                  ; go try a float if failure
         call    pushInt                ; push integer on int stack 
         call    tracer                 ; trace it             
         jmp     short parse            ; keep parsing
par01:   call    asc2dec                ; try to convert to float
         inc     di                     ; test flag
         je      par02                  ; go if not a float                 
         call    dec2flt                ; float to ST(0)
         call    pushFlt                ; put float on flt stack
         call    tracer                 ; trace it
         jmp     short parse            ; keep parsing

; dictionary words are searched before user defined words
; this means that dictionary words cannot not be overridden
; if one wishes, user words can be searched first but the
; token will have to be save and restored first (because chkWrd
; standardizes the token by padding with spaces or paring off chars)

par02:   call    srchDct                ; search dictionary
         cmp     bl,-1                  ; test for failure
         je      par03                  ; go if not in dictionary
         call    xeq                    ; go execute the code
         call    tracer                  ; trace it
         jmp     short parse            ; keep parsing
par03:   mov     di,offset token        ; get token address
         call    toke8                  ; pad or pare to 8 chars
         call    chkWrd                 ; check token for word name
         or      dx,dx                  ; dx is -1 it it's not a word         
         jns     parse                  ; it was so return to parsing
         call    chkFvar                ; check token for fvariable
         or      dx,dx                  ; test flag
         jns     parse                  ; go parse more if it was
         call    chkVar                 ; check token for variable             
         or      dx,dx                  ; test flag
         jns     parse                  ; go parse more if it was
         call    chkfCon                ; check token for fconstant
         or      dx,dx                  ; test flag
         jns     parse                  ; go parse more if it was
         call    chkCon                 ; check token for constant
         or      dx,dx                  ; test flag
         jns     parse                  ; go parse more if it was                

prsErr:  mov     di,offset wrdErr       ; word unmatched - point to err msg
         jmp     error                  ; go print message

par04:   ret

;------------------------------------------------------------------------------
; trace routine
; entry: trflag checked, traces if set, returns if reset
; prints words being executed and stack contents
;------------------------------------------------------------------------------
 
tracer:   test    trflag,1               ; check flag
          je      trc01                  ; done if reset
          call    prnCRLF                ; newline
          mov     ah,forClr              ; save foreground color
          push    ax
          mov     ah,DARKCYAN            ; set foreground color
          mov     forclr,ah
          mov     di,offset trwmMsg      ; print 'word: '
          call    prnStr
          mov     di,offset token        ; print word (token)          
          call    prnStr
          call    prnCRLF
          mov     di,offset trisMsg      ; print 'int stk: '
          call    prnStr
          call    xeq001                 ; print int stack
          mov     di,offset trfsMsg      ; print 'flt stk: '
          call    prnStr
          call    xeq010                 ; print flt stack
          pop     ax
          mov     forClr,ah
          call    inpChr                 ; wait for key
trc01:    ret


;------------------------------------------------------------------------------
; execute the word
; all registers are free to use except bp, which must be preserved
;------------------------------------------------------------------------------

Xeq:     dec     bl                     ; dispatch table starts at 0
         xor     bh,bh                  ; convert token number to a word
         sal     bx,1                   ; times 2
         add     bx,offset dsptch       ; compute index into dispatch table
         jmp     CS:[bx]                ; dispatch

;------------------------------------------------------------------------------
; dispatch table
;------------------------------------------------------------------------------

dsptch:  dw      offset xeq001          ; .s
         dw      offset xeq002          ; .
         dw      offset xeq003          ; +
         dw      offset xeq004          ; -
         dw      offset xeq005          ; *
         dw      offset xeq006          ; /
         dw      offset xeq007          ; d*
         dw      offset xeq008          ; d/
         dw      offset xeq009          ; cr
         dw      offset xeq010          ; .f
         dw      offset xeq011          ; f.
         dw      offset xeq012          ; f+
         dw      offset xeq013          ; f-
         dw      offset xeq014          ; f*
         dw      offset xeq015          ; f/
         dw      offset xeq016          ; dup
         dw      offset xeq017          ; fdup
         dw      offset xeq018          ; drop
         dw      offset xeq019          ; fdrop
         dw      offset xeq020          ; swap
         dw      offset xeq021          ; fswap
         dw      offset xeq022          ; quote
         dw      offset xeq023          ; type
         dw      offset xeq024          ; pad
         dw      offset xeq025          ; cls
         dw      offset xeq026          ; \ 
         dw      offset xeq027          ; (
         dw      offset xeq028          ; fvariable
         dw      offset xeq029          ; variable
         dw      offset xeq030          ; f!
         dw      offset xeq031          ; f@
         dw      offset xeq032          ; !
         dw      offset xeq033          ; @
         dw      offset xeq034          ; f?
         dw      offset xeq035          ; ?
         dw      offset xeq036          ; if
         dw      offset xeq037          ; jump
         dw      offset xeq038          ; 0<
         dw      offset xeq039          ; 0=
         dw      offset xeq040          ; 0>
         dw      offset xeq041          ; f0<
         dw      offset xeq042          ; f0=
         dw      offset xeq043          ; f0>
         dw      offset xeq044          ; <
         dw      offset xeq045          ; =
         dw      offset xeq046          ; >
         dw      offset xeq047          ; f<
         dw      offset xeq048          ; f=
         dw      offset xeq049          ; f>
         dw      offset xeq050          ; f0~
         dw      offset xeq051          ; not
         dw      offset xeq052          ; and
         dw      offset xeq053          ; or
         dw      offset xeq054          ; xor
         dw      offset xeq055          ; rot
         dw      offset xeq056          ; frot
         dw      offset xeq057          ; pick
         dw      offset xeq058          ; fpick
         dw      offset xeq059          ; :
         dw      offset xeq060          ; ;
         dw      offset xeq061          ; test
         dw      offset xeq062          ; skip
         dw      offset xeq063          ; do
         dw      offset xeq064          ; loop
         dw      offset xeq065          ; +loop
         dw      offset xeq066          ; i
         dw      offset xeq067          ; j
         dw      offset xeq068          ; k
         dw      offset xeq069          ; undo
         dw      offset xeq070          ; begin
         dw      offset xeq071          ; again
         dw      offset xeq072          ; unbegin
         dw      offset xeq073          ; accept
         dw      offset xeq074          ; execute
         dw      offset xeq075          ; emit
         dw      offset xeq076          ; space
         dw      offset xeq077          ; spaces
         dw      offset xeq078          ; 1+
         dw      offset xeq079          ; 1-
         dw      offset xeq080          ; inc
         dw      offset xeq081          ; dec
         dw      offset xeq082          ; fint
         dw      offset xeq083          ; f>n
         dw      offset xeq084          ; n>f
         dw      offset xeq085          ; key
         dw      offset xeq086          ; fsqrt
         dw      offset xeq087          ; f10^x
         dw      offset xeq088          ; fe^x
         dw      offset xeq089          ; flog10
         dw      offset xeq090          ; fln
         dw      offset xeq091          ; ftan
         dw      offset xeq092          ; pi
         dw      offset xeq093          ; fd2r
         dw      offset xeq094          ; fr2d
         dw      offset xeq095          ; fatan
         dw      offset xeq096          ; fsin
         dw      offset xeq097          ; fcos
         dw      offset xeq098          ; abs
         dw      offset xeq099          ; fabs
         dw      offset xeq100          ; neg
         dw      offset xeq101          ; fneg
         dw      offset xeq102          ; fempty
         dw      offset xeq103          ; empty 
         dw      offset xeq104          ; clean
         dw      offset xeq105          ; depth
         dw      offset xeq106          ; fdepth 
         dw      offset xeq107          ; places
         dw      offset xeq108          ; fplaces 
         dw      offset xeq109          ; fgcolor
         dw      offset xeq110          ; bgcolor
         dw      offset xeq111          ; incolor
         dw      offset xeq112          ; cold
         dw      offset xeq113          ; list
         dw      offset xeq114          ; load
         dw      offset xeq115          ; +load
         dw      offset xeq116          ; words
         dw      offset xeq117          ; h.
         dw      offset xeq118          ; at-xy
         dw      offset xeq119          ; true
         dw      offset xeq120          ; false
         dw      offset xeq121          ; bye
         dw      offset xeq122          ; mem
         dw      offset xeq123          ; allot
         dw      offset xeq124          ; fallot
         dw      offset xeq125          ; push
         dw      offset xeq126          ; pop 
         dw      offset xeq127          ; fpush
         dw      offset xeq128          ; fpop
         dw      offset xeq129          ; epsilon!
         dw      offset xeq130          ; epsilon@
         dw      offset xeq131          ; fconstant
         dw      offset xeq132          ; constant
         dw      offset xeq133          ; vlist
         dw      offset xeq134          ; comma
         dw      offset xeq135          ; f comma
         dw      offset xeq136          ; tee-on
         dw      offset xeq137          ; tee-off
         dw      offset xeq138          ; graphics
         dw      offset xeq139          ; plot
         dw      offset xeq140          ; text
         dw      offset xeq141          ; forget
         dw      offset xeq142          ; fatan2
         dw      offset xeq143          ; mod
         dw      offset xeq144          ; facos
         dw      offset xeq145          ; fasin
         dw      offset xeq146          ; fpwr
         dw      offset xeq147          ; fcubrt
         dw      offset xeq148          ; fsinh
         dw      offset xeq149          ; fcosh
         dw      offset xeq150          ; ftanh
         dw      offset xeq151          ; fatanh
         dw      offset xeq152          ; line
         dw      offset xeq060          ; exit
         dw      offset xeq154          ; number
         dw      offset xeq155          ; fnumber
         dw      offset xeq156          ; string
         dw      offset xeq157          ; fstring
         dw      offset xeq158          ; circle
         dw      offset xeq159          ; pad store
         dw      offset xeq160          ; pad fecth
         dw      offset xeq161          ; free
         dw      offset xeq162          ; pad
         dw      offset xeq163          ; c!
         dw      offset xeq164          ; c@
         dw      offset xeq165          ; '
         dw      offset xeq166          ; drop2
         dw      offset xeq167          ; fdrop2
         dw      offset xeq168          ; +spaces
         dw      offset xeq169          ; slice
         dw      offset xeq170          ; concat
         dw      offset xeq171          ; upper
         dw      offset xeq172          ; lower
         dw      offset xeq173          ; black
         dw      offset xeq174          ; darkblue
         dw      offset xeq175          ; darkgreen
         dw      offset xeq176          ; darkcyan
         dw      offset xeq177          ; darkred
         dw      offset xeq178          ; darkmagenta
         dw      offset xeq179          ; brown
         dw      offset xeq180          ; grey
         dw      offset xeq181          ; darkgrey
         dw      offset xeq182          ; blue
         dw      offset xeq183          ; green
         dw      offset xeq184          ; cyan
         dw      offset xeq185          ; red
         dw      offset xeq186          ; magenta
         dw      offset xeq187          ; yellow
         dw      offset xeq188          ; white
         dw      offset xeq189          ; date
         dw      offset xeq190          ; time
         dw      offset xeq191          ; random
         dw      offset xeq192          ; aspect
         dw      offset xeq193          ; =>
         dw      offset xeq194          ; <=
         dw      offset xeq195          ; f=>
         dw      offset xeq196          ; f<=
         dw      offset xeq197          ; !=
         dw      offset xeq198          ; f!=
         dw      offset xeq199          ; pad=
         dw      offset xeq200          ; search
         dw      offset xeq201          ; create-file
         dw      offset xeq202          ; open-file
         dw      offset xeq203          ; close-file
         dw      offset xeq204          ; write-file
         dw      offset xeq205          ; read-file
         dw      offset xeq206          ; tab
         dw      offset xeq207          ; locate
         dw      offset xeq208          ; read
         dw      offset xeq209          ; write
         dw      offset xeq210          ; append
         dw      offset xeq211          ; trace-on
         dw      offset xeq212          ; trace-off
         dw      offset xeq213          ; e.
         dw      offset xeq214          ; eplaces
         dw      offset xeq215          ; estring
         dw      offset xeq216          ; .p
         dw      offset xeq217          ; g.
         dw      offset xeq218          ; l.
         dw      offset xeq219          ; font
         dw      offset xeq220          ; xorgr-on
         dw      offset xeq221          ; xorgr-off
         dw      offset xeq222          ; system
         dw      offset xeq223          ; word fetch
         dw      offset xeq224          ; word store
         dw      offset xeq225          ; reparse
         dw      offset xeq226          ; shift right
         dw      offset xeq227          ; shift left
         dw      offset xeq228          ; error-on
         dw      offset xeq229          ; error-off
         dw      offset xeq230          ; clear-mem
         dw      offset xeq231          ; free
                  
;-----------------------------------------------------------------------------
; execution routines
; land here thru the word dispatch table
;-----------------------------------------------------------------------------

xeq001:  mov     bl,iStkX               ; .S - get int stack index
         or      bl,bl                  ; test for empty
         je      xq1b                   ; go if so
         xor     bl,bl                  ; set count to 0
xq1a:    push    bx                     ; save count
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; times 2
         mov     ax,intStack[bx]        ; get integer from stack
         call    prnInt                 ; print it
         call    prnSPC                 ; print space
         pop     bx                     ; get count back
         inc     bl                     ; increment it
         cmp     bl,iStkX               ; test for all elements printed
         js      xq1a                   ; go back to do more
         jmp     short xq1c             ; go issue a new line, done
xq1b:    mov     di,offset imtMsg       ; point to message
         call    prnStr                 ; print it
xq1c:    jmp     prnCRLF                ; go do a new line 

xeq035:  call    xeq033                 ; ? - fetch integer and fall thru
      
xeq002:  call    popInt                 ; dot - pop TOIS
         jmp     prnInt                 ; go print integer

xeq003:  call    popInt                 ; + get first value from int stack
         mov     bx,ax                  ; hold it in bx
         call    popInt                 ; get second value
         add     ax,bx                  ; add them
         jmp     pushInt                ; put value back on TOIS

xeq004:  call    popInt                 ; - get first value from int stack
         mov     bx,ax                  ; hold it in bx
         call    popInt                 ; get second value
         sub     ax,bx                  ; subtract them
         jmp     pushInt                ; put value back on TOIS

xeq005:  call    popInt                 ; * get first value from int stack
         mov     bx,ax                  ; hold it in bx
         call    popInt                 ; get second value
         imul    bx                     ; multiply them
         jmp     pushInt                ; put value back on TOIS

xeq006:  call    popInt                 ; / get first value from int stack
         mov     bx,ax                  ; hold it in bx
         or      bx,bx                  ; test divisor
         je      xq06a                  ; divide by zero trap
         call    popInt                 ; get second value
         xor     dx,dx                  ; clear dx
         or      ax,ax                  ; work ax flags
         jns     xq06b                  ; skip if ax plus
         dec     dx                     ; else make dx -1
xq06b:   idiv    bx                     ; divide them
         jmp     pushInt                ; put value back on TOIS
xq06a:   jmp     xq86a                  ; go print math error

xeq007:  call    popInt                 ; d* get first value from int stack
         mov     bx,ax                  ; hold it in bx
         call    popInt                 ; get second value
         mul     bx                     ; multiply them
         mov     mulDX,dx               ; save dx
         jmp     pushInt                ; put value back on TOIS

xeq008:  call    popInt                 ; d/ get first value from int stack
         mov     bx,ax                  ; hold it in bx
         or      bx,bx                  ; test divisor
         je      xq06a                  ; divide by zero trap
         call    popInt                 ; get second value
         mov     dx,mulDX               ; set dx
         div     bx                     ; divide them
         jmp     pushInt                ; put value back on TOIS

xeq009:  jmp     prnCRLF                ; cr - go print CR/LF

xeq010:  mov     bl,fStkX               ; .f - get flt stack index
         or      bl,bl                  ; test for empty
         je      xq10b                  ; go if so
         xor     bl,bl                  ; set count to 0
xq10a:   push    bx                     ; save count
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fltStack [bx]          ; get float from stack
         call    prnFlt                 ; print it
         pop     bx                     ; get count back
         mov     al,bl                  ; every four numbers...
         sub     al,3
         xor     ah,ah
         mov     bh,4
         div     bh
         or      ah,ah                  ;  (test for no rem after div by 4)
         jne     xq10d
         call    prnCRLF                ; ... print a new line
xq10d:   inc     bl                     ; increment count
         cmp     bl,fStkX               ; test for all elements printed
         js      xq10a                  ; go back for more
         jmp     short xq10c            ; go issue a new line, done
xq10b:   mov     di,offset fmtMsg       ; float stack empty message
         call    prnStr                 ; print it
xq10c:   jmp     prnCRLF                ; go do a new line 

xeq034:  call    xeq031                 ; f? - fetch float and fall thru
      
xeq011:  call    popFlt                 ; f. - pop TOFS
         jmp     prnFlt                 ; go print float

xeq012:  call    popFlt                 ; f+ get first value of float stack
         call    popFlt                 ; get second value
         FADD                           ; add floats
         jmp     pushFlt                ; put answer on TOFS, done

xeq013:  call    popFlt                 ; f- get first value of float stack
         call    popFlt                 ; get second value
         FSUBR                          ; subtract floats
         jmp     pushFlt                ; put answer on TOFS, done

xeq014:  call    popFlt                 ; f* get first value of float stack
         call    popFlt                 ; get second value
         FMUL                           ; multiply floats
         jmp     pushFlt                ; put answer on TOFS, done

xeq015:  call    popFlt                 ; f/ get first value of float stack
         FTST                           ; test ST(0), divisor
         FSTSW   npustat                ; store status
         FWAIT
         and     byte ptr npustat+1,40h ; test zero flag
         jne     xq15a                  ; divide by zero trap
         call    popFlt                 ; get second value
         FDIVR                          ; divide floats
         jmp     pushFlt                ; put answer on TOFS, done
xq15a:   jmp     xq86a                  ; go print math error

xeq016:  call    popInt                 ; dup - pop TOIS
         call    pushInt                ; push back
         jmp     pushInt                ; push again, done

xeq017:  call    popFlt                 ; fdup - pop TOFS
         FST     ST(1)                  ; dup 8087 stack
         call    pushFlT                ; push back
         jmp     pushFlt                ; push again, done

xeq018:  jmp     popInt                 ; drop - pop TOIS, done

xeq019:  jmp     popFlt                 ; fdrop - pop TOFS, done

xeq020:  cmp     iStkX,2                ; swap - need at least 2 entries
         js      xq20a                  ; go if not
         mov     bl,iStkX               ; get stack top               
         xor     bh,bh                  ; make it a word
         dec     bx                     ; starts at zero
         sal     bx,1                   ; times 2
         add     bx,offset intStack     ; point bx to stack element
         mov     ax,[bx]                ; get TOIS
         mov     cx,[bx-2]              ; get next int entry
         mov     [bx],cx                ; done with swap
         mov     [bx-2],ax
         ret                            ; done
xq20a:   jmp     iStkEmt                ; here if stack err - go prt msg

xeq021:  cmp     fStkX,2                ; fswap - need at least 2 entries
         js      xq21a                  ; go if not
         mov     bl,fStkX               ; get stack top
         xor     bh,bh                  ; make it a word
         dec     bx                     ; starts at zero
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fltStack[bx]           ; get TOFS into ST(0)
         FLD     fltStack[bx-10]        ; get next flt entry
         FSTP    fltStack[bx]           ; swap
         FSTP    fltStack[bx-10]
         ret                            ; done with fswap
xq21a:   jmp     fStkEmt                ; here if stack err - gp prt msg 

xeq022:  mov     si,bp                  ; quote - get scan parm
         mov     cx,100                 ; length of prepad+pad
         mov     di,offset prepad+3     ; point to prepad
xq22f:   mov     byte ptr [di],SPC      ; fill with spaces
         inc     di                     ; next pad space
         loop    xq22f                  ; loop till done
         mov     al,QUOTE               ; search char quote
         call    fndChr                 ; go find it
         jc      xq22c                  ; cy means not found so go
         cmp     dx,80                  ; exceeds pad?
         jns     xq22d                  ; error if so
         mov     cx,dx                  ; save count to next token
         mov     lpad,dl                ; else set lpad to length
         mov     di,offset pad          ; move string - point di to pad
         or      dl,dl                  ; is length 0?
         je      xq22e                  ; go store terminator if so
xq22a:   mov     al,[si]                ; move loop
         mov     [di],al
         inc     si                     ; bump up pointers
         inc     di
         dec     dl                     ; bump down count
         jne     xq22a                  ; and continue move till zero
xq22e:   mov     byte ptr [di],NULL     ; store null terminator
xq22b:   add     bp,cx                  ; update scan pointer
         inc     bp                     ; bump to just past quote
         mov     si,offset pad          ; set next number address
         mov     nxtNum,si
         ret                            ; done with quote
xq22c:   mov     di,offset qotErr       ; no end quote found - error
         jmp     error
xq22d:   mov     di,offset qtpErr       ; pad size exceeded error
         jmp     error

xeq023:  call    popInt                 ; type - get TOIS
         mov     di,offset pad          ; point di to pad
         mov     cx,ax                  ; copy parm to cx
         or      cx,cx                  ; is it zero?
         je      xq23d                  ; go if so, nothing to do
         jns     xq23a                  ; go if positive
         neg     cx                     ; it's minus, negate parm copy
         cmp     cl,lpad
         jns     xq23b
         add     al,lpad                ; al is now length(pad - parm)
         xor     ah,ah                  ; make it a word
         add     di,ax                  ; di now ignores (length-parm) chars
         jmp     short xq23c            ; go print string
xq23a:   cmp     cl,lpad                ; here if parm +, compare to lpad
         js      xq23c
xq23b:   jmp     prnStr                 ; else print all of pad 
xq23c:   call    PrnSN                  ; go print counted string
xq23d:   ret

xeq024:  mov     al,lpad                ; len - get length
         xor     ah,ah                  ; make it a word
         jmp     pushInt                ; go push on TOIS

xeq025:  test    gflag,1                ; cls - in graphics mode?
         je      xq25a                  ; go if not
         call    gcls                   ; clear graphics screen
         jmp     ghome                  ; home graphics cursor
xq25a:   call    cls                    ; clear text screen
         jmp     homCur                 ; go home cursor - done with cls

xeq026:  mov     si,bp                  ; \ - comment - get scan parm
         mov     al,CR                  ; look for eol
         call    fndChr
         jc      xq26a
         inc     bp                     ; not eol, bump just past srch char
xq26a:   inc     bp                     ; bump one char otherwise             
         add     bp,dx                  ; set scan parm 
         ret                            ; done with \ comment

xeq027:  mov     si,bp                  ; ( - comment - get scan parms
         mov     al,')'                 ; look for closing paren
         call    fndChr                 ; go look
         jc      xq27a                  ; error if not found
         add     bp,dx                  ; adjust scan parm
         inc     bp                     ;   to char after )
         ret
xq27a:   mov     di,offset cmtErr       ; here if no ) found
         jmp     error                  ; go print error msg

xeq028:  call    getToke                ; fvariable - get name
         call    toke8                  ; pad or pare to 8 chars
         call    is1alf                 ; make UC and check validity
         je      xq28d                  ; if null, no name - error
         jc      xq28d                  ; if carry, name invalid
         mov     di,offset token        ; set pointer to token
         call    fndfCon                ; name used as an fconstant?
         jnc     xq28f                  ; if so, then it's a duplicate
         call    fndCon                 ; name used as a constant?
         jnc     xq28f                  ; if so, then it's a duplicate
         call    fndVarn                ; name used as a variable name?
         jnc     xq28f                  ; if so then it's a duplicate
         call    fndWrd                 ; find word name
         jnc     xq28f                  ; if so then it's a duplicate
         call    fndFvrn                ; find this name
         jnc     xq28c                  ; if found, fvar exists, do nothing
         cmp     fvarX,varMax           ; can we create another one?
         je      xq28e                  ; go if no, fvar space full
xq28a:   inc     fvarX                  ; bump up index
         mov     bl,fvarX               ; put it in bl
         mov     fvadr,bl               ; save it as current fvar address    
         xor     bh,bh                  ; make it a word
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 8
         sal     bx,1
         sal     bx,1
         mov     si,offset fvnams       ; set pointers
         mov     di,offset token

; call here to move a name into a name table
; moves cl words from [di] to [si]
; bx is offset into table

nmMov:   mov     cl,4                   ; move name 4 x word size = 8 elements
xq28b:   mov     ax,[di]                ; move token name into array
         mov     [bx+si],ax
         add     si,2                   ; bump pointers by 1 word
         add     di,2
         dec     cl                     ; dec count
         jne     xq28b                  ; and keep moving if more
         ret                            ; done 

xq28c:   inc     dx                     ; here if fvar exists
         mov     fvadr,dl               ; store current address
         ret                            ; done with fvariable

xq28d:   mov     di,offset vrnErr       ; here if fvar name error
         jmp     error                  ;   print error message              
xq28e:   mov     di,offset ofvsErr      ; here if out of fvar space
         jmp     error                  ;   print error message 
xq28f:   mov     di,offset vdupErr      ; here if duplicate name   
         jmp     error                  ;   print error message

xeq029:  call    getToke                ; variable - get name
         call    toke8                  ; pad or pare to 8 chars
         call    is1alf                 ; make UC and check validity
         je      xq28d                  ; if null, no name - error
         jc      xq28d                  ; if carry, name invalid
         mov     di,offset token        ; set pointer to token
         call    fndfCon                ; name used as an fconstant?
         jnc     xq28f                  ; if so, then it's a duplicate
         call    fndCon                 ; name used as a constant?
         jnc     xq28f                  ; if so, then it's a duplicate
         call    fndFvrn                ; name used as fvariable?
         jnc     xq28f                  ; if so then it's a duplicate
         call    fndWrd                 ; find word name
         jnc     xq28f                  ; if so then it's a duplicate
         call    fndVarn                ; find this name
         jnc     xq29c                  ; if found, var exists, do nothing
         cmp     varX,varMax            ; can we create another one?
         je      xq28e                  ; go if no, var space full
xq29a:   inc     varX                   ; bump up index
         mov     bl,varX                ; put it in bl
         mov     vadr,bl                ; save as current variable address
         xor     bh,bh                  ; make it a word
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 8
         sal     bx,1
         sal     bx,1
         mov     si,offset vnams        ; set pointers
         mov     di,offset token
         jmp     short nmMov            ; go move name                     

xq29c:   inc     dx                     ; here if variable found
         mov     vadr,dl                ; store current variable address
         ret                            ; done with fvariable

xeq030:  call    popInt                 ; f! - get flt array index
         mov     bx,ax                  ; put in bx     
         call    popFlt                 ; get flt value   
         cmp     bx,1                   ; check range of index
         jb      xq30a                  ; too low, error
         cmp     bx,fvarMax+20          ; fvarMax is the biggest
         ja      xq30a                  ; too high, error
         dec     bx                     ; zero base it
         sal     bx,1                   ; fast multiply by ten
         mov     cx,bx
         sal     bx,1
         sal     bx,1
         add     bx,cx
         FSTP    fvars[bx]              ; put flt value in the array
xq30a:   ret                            ; done with f!

xeq031:  call    popInt                 ; f@ - get flt array index
         mov     bx,ax                  ; put in bx
         cmp     bx,1                   ; check range of index
         jb      xq30a                  ; too low, error
         cmp     bx,fvarMax+20          ; fvarMax+20 is the biggest
         ja      xq30a                  ; too high, error
         dec     bx                     ; zero base it
         sal     bx,1                   ; fast multiply by ten
         mov     cx,bx
         sal     bx,1
         sal     bx,1
         add     bx,cx         
         FLD     fvars[bx]              ; get value from the flt array
         jmp     pushFlt                ; put it on the float stack - done

xeq032:  call    popInt                 ; ! - get int array index
         mov     bx,ax                  ; put in bx     
         call    popInt                 ; get int value   
         cmp     bx,1                   ; check range of index
         jb      xq32a                  ; too low, error
         cmp     bx,varMax+20           ; varMax is the biggest
         ja      xq32a                  ; too high, error
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 2
         mov     vars[bx],ax            ; put int value in the array
xq32a:   ret                            ; done with !

xeq033:  call    popInt                 ; @ - get int array index
         mov     bx,ax                  ; put in bx
         cmp     bx,1                   ; check range of index
         jb      xq32a                  ; too low, error
         cmp     bx,varMax+20           ; varMax+20 is the biggest
         ja      xq32a                  ; too high, error
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 2        
         mov     ax,vars[bx]            ; get value from the int array
         jmp     pushInt                ; put it on the float stack - done

xeq036:  mov     si,bp                  ; if - get parsing parms
         call    popInt                 ; get TOIS
         cmp     ax,FALSE               ; test for false 
         jne     xq32a                  ; if true then do nothing
         mov     al,CR                  ; if false - search for line end
         call    fndChr
;        jc      xq36a                  ; for now, do nothing if no eol
xq36a:   add     bp,dx                  ; update scan parm
         inc     bp                     ; point to Cr or null terminator
         ret

xeq037:  call    getToke                ; jump - get label
         cmp     byte ptr [di],NULL     ; no name?
         je      xq37a                  ; error if so
         call    toke8                  ; pad or pare to 8 chars
         call    upper                  ; make UC
         call    srchLbl                ; find a label match
         cmp     si,-1                  ; check for not found
         je      xq37b                  ; error if so
         mov     bp,si                  ; update scan parm
         ret                            ; done with jump
xq37a:   mov     di,offset lbnErr       ; here if label name error
         jmp     error                  ; go print error message
xq37b:   mov     di,offset lnfErr       ; here if label not found error
         jmp     error                  ; go print error eror message

xeq038:  call    popInt                 ; 0< - get TOIS
         or      ax,ax                  ; test it
         jge     xq38a                  ; go if g.e.
xq38b:   mov     ax,TRUE                ; force true
         jmp     pushInt                ; put it on the stack
xq38a:   mov     ax,FALSE               ; force false
         jmp     pushInt                ; put it on the stack

xeq039:  call    popInt                 ; 0= - get TOIS
         or      ax,ax                  ; test it
         je      xq38b                  ; go if zero - force true
         jmp     short xq38a            ; go if not zero - force false

xeq040:  call    popInt                 ; 0> - get TOIS
         or      ax,ax                  ; test it
         jg      xq38b                  ; go if g.t. - force true
         jmp     short xq38a            ; go if not pos - force false

xeq041:  call    popFlt                 ; f0< - get TOFS
         FTST                           ; test ST(0)
         FSTSW   npustat                ; store stats
         FWAIT
         FSTP    ST(0)                  ; pop ST(0) off 8087 stack
         and     byte ptr npustat+1,01h ; ST(0) minus?
         jne     xq38b                  ; go if so - force true
         jmp     short xq38a            ; else force false 

xeq042:  call    popFlt                 ; f0= - get TOFS
         FTST                           ; test ST(0)
         FSTSW   npustat                ; store stats
         FWAIT
         FSTP    ST(0)                  ; pop ST(0) off 8087 stack
         and     byte ptr npustat+1,40h ; ST(0) zero?
         jne     xq38b                  ; go if so - force true
         jmp     short xq38a            ; else force false   

xeq043:  call    popFlt                 ; f0> - get TOFS
         FTST                           ; test ST(0)
         FSTSW   npustat                ; store stats
         FWAIT
         FSTP    ST(0)                  ; pop ST(0) off 8087 stack
         and     byte ptr npustat+1,41h ; ST(0) positive?
         je      xq38b                  ; go if so - force true
         jmp     short xq38a            ; else force false

xeq044:  call    popInt                 ; < - get TOIS
         mov     bx,ax                  ; put it in bx
         call    popInt                 ; get next int stack element
         cmp     ax,bx                  ; compare
         jge     xq44a                  ; go if stack next > stack top
xq44b:   mov     ax,TRUE                ; force true
         jmp     pushInt                ; put it on TOIS
xq44a:   mov     ax,FALSE               ; force false
         jmp     pushInt                ; put it on TOIS

xeq045:  call    popInt                 ; = - get TOIS
         mov     bx,ax                  ; put it in bx
         call    popInt                 ; get next int stack element
         cmp     ax,bx                  ; compare stack next = stack top
         je      xq44b                  ; go if equal force true
         jmp     short xq44a            ; else force false

xeq046:  call    popInt                 ; > - get TOIS
         mov     bx,ax                  ; put it in bx
         call    popInt                 ; get next int stack element
         cmp     ax,bx                  ; compare stack next > stack top
         jle     xq44a                  ; go if equal force true
         jmp     short xq44b            ; else force false

xeq047:  call    popFlt                 ; f< - get TOFS
         call    popFlt                 ; get next flt stack element
         FCOMPP                         ; compare next - top
         FSTSW   npustat                ; store stats
         FWAIT
         and     byte ptr npustat+1,01h ; ST(0) minus?
         je      xq44a                  ; go if so - force true
         jmp     short xq44b            ; else force false

xeq048:  call    popFlt                 ; f= - get TOFS
         call    popFlt                 ; get next flt stack element
         FCOMPP                         ; compare next - top
         FSTSW   npustat                ; store stats
         FWAIT
         and     byte ptr npustat+1,40h ; ST(0) zero?
         je      xq44a                  ; go if so - force true
         jmp     short xq44b            ; else force false

xeq049:  call    popFlt                 ; f> - get TOFS
         call    popFlt                 ; get next flt stack element
         FCOMPP                         ; compare next - top
         FSTSW   npustat                ; store stats
         FWAIT
         and     byte ptr npustat+1,41h ; ST(0) pos?
         jne     xq44a                  ; go if so - force true
         jmp     short xq44b            ; else force false

xeq050:  call    popFlt                 ; f0~ - get TOFS
         FABS                           ; |ST(0)|
         FLD     epsilon                ; |ST(0)|,epsilon
         FSUB                           ; |ST(0)|-epsilon
         FTST                           ; test ST(0)       
         FSTSW   npustat                ; store stats
         FWAIT
         FSTP    ST(0)                  ; pop 8087 stack
         and     byte ptr npustat+1,01h ; ST(0) minus?
         jne     xq50b                  ; go if so - force true
         jmp     xq44a                  ; else force false 
xq50b:   jmp     xq44b                  ; force true

xeq051:  call    popInt                 ; not - get TOIS
         xor     ax,0FFFFh              ; compliment it
         jmp     pushInt                ; put it back - done

xeq052:  call    popInt                 ; and - get TOIS
         mov     bx,ax                  ; put it in bx
         call    popInt                 ; get next
         and     ax,bx                  ; AND them
         jmp     pushInt                ; put result back on TOIS - done

xeq053:  call    popInt                 ; or - get TOIS
         mov     bx,ax                  ; put it in bx
         call    popInt                 ; get next
         or      ax,bx                  ; OR them
         jmp     pushInt                ; put result back on TOIS - done

xeq054:  call    popInt                 ; xor - get TOIS
         mov     bx,ax                  ; put it in bx
         call    popInt                 ; get next
         xor     ax,bx                  ; XOR them
         jmp     pushInt                ; put result back on TOIS - done

xeq055:  cmp     iStkX,2                ; rot - enough stack?
         ja      xq55a                  ; yes, go rotate
         je      xq55b                  ; only 2 elements, go swap
         ret                            ; 1 or less then do nothing
xq55a:   mov     bl,iStkX               ; get int stack index
         xor     bh,bh                  ; now in bx
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 2
         mov     ax,intStack[bx-4]      ; get 3rd
         xchg    ax,intStack[bx]        ; swap top
         xchg    ax,intStack[bx-2]      ; swap 2nd      
         mov     intStack[bx-4],ax      ; store 3rd
         ret                            ; done with rotate
xq55b:   jmp     xeq020                 ; swap

xeq056:  cmp     fStkX,2                ; frot - enough stack?
         ja      xq56a                  ; yes, go rotate
         je      xq56b                  ; only 2 elements, go swap
         ret                            ; 1 or less then do nothing
xq56a:   mov     bl,fStkX               ; get flt stack index
         xor     bh,bh                  ; now in bx
         dec     bx                     ; zero base it
         sal     bx,1                   ; fast x 10
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fltStack[bx-20]        ; get 3rd
         FSTP    fptemp                 ; save 3rd 
         FWAIT          
         FLD     fltStack[bx-10]        ; get 2nd
         FSTP    fltStack[bx-20]        ; store 3rd
         FWAIT
         FLD     fltStack[bx]           ; get top
         FSTP    fltStack[bx-10]        ; store 2nd
         FWAIT
         FLD     fptemp                 ; get 3rd
         FSTP    fltStack[bx]           ; store top
         FWAIT   
         ret                            ; done with float rotate
xq56b:   jmp     xeq021                 ; fswap

xeq057:  mov     bl,iStkX               ; pick - get stack size
         xor     bh,bh                  ; make it a word
         dec     bx                     ; ignore pick entry
         call    popInt                 ; get TOIS
         sub     bx,ax                  ; subtract pick amount 
         cmp     bl,0                   ; check stack size
         jb      xq57a                  ; error if not enough
         cmp     bl,iStkX               ; or too much
         jae     xq57a
         sal     bx,1                   ; x 2
         mov     ax,intStack[bx]        ; get stack element
         jmp     pushInt                ; put on top of int stack, done
xq57a:   jmp     iStkEmt                ; relay to int stack empty error

xeq058:  mov     bl,fStkX               ; fpick - get stack size
         xor     bh,bh                  ; make it a word
         call    popInt                 ; get pick value of int stack
         sub     bx,ax                  ; subtract it            
         cmp     bl,0                   ; check stack size
         jb      xq58a                  ; error if not enough
         cmp     bl,fStkX               ; or too much
         jae     xq58a
         sal     bx,1                   ; fast x 10
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fltStack[bx]           ; load the stack element
         jmp     pushFlt                ; push on TOFS
xq58a:   jmp     fStkEmt                ; relay to flt stack empty error

xq59f:   mov     di,offset wdupErr      ; here if word name in use 
         jmp     error                  ;  print error message
xq59d:   mov     di,offset wdnErr       ; here if word name error
         jmp     error                  ;  print error message

xeq059:  mov     ax,nxtAdr              ; : - get next pos. pointer
         dec     ax                     ; back up to :
         dec     ax
         mov     strAdr,ax              ; store ptr to colon at start adr
         call    getToke                ; get word name
         mov     ax,nxtAdr              ; get next pos. pointer
         sub     ax,strAdr              ; compute offset
         mov     offVal,ax              ; save it as offVal
         call    toke8                  ; pad or pare to 8 chars

         call    upper
         cmp     byte ptr [di],NULL
         je      xq59d

;         call    is1alf                 ; make UC and check validity
;         je      xq59d                  ; if null, no name - error
;         jc      xq59d                  ; if carry, name invalid
         mov     di,offset token        ; set pointer to token
         call    fndfCon                ; name used as an fconstant?
         jnc     xq59f                  ; if so, then it's a duplicate
         call    fndCon                 ; name used as a constant?
         jnc     xq59f                  ; if so, then it's a duplicate
         call    fndVarn                ; name used as a variable name?
         jnc     xq59f                  ; if so then it's a duplicate
         call    fndFvrn                ; name used as fvar name?
         jnc     xq59f                  ; if so then it's a duplicate
         call    fndWrd                 ; find word name
         jc      xq59k                  ; go if not found
         jmp     xq59j                  ; warning if found
xq59k:   cmp     wrdX,wrdMax            ; can we create another one?
         jb      xq59k1                 ; skip over if ok
         jmp     xq59e                  ; go to error if not        
xq59k1:  inc     wrdX                   ; bump up index
         mov     bl,wrdX                ; put it in bl
         xor     bh,bh                  ; make it a word
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 8
         mov     temp,bx                ; (save x 2)
         sal     bx,1 
         sal     bx,1                  
         mov     si,offset wrdnams      ; set pointers
         mov     di,offset token
         call    nmMov                  ; go move name into word space
         mov     ax,prgEnd		; save pointer to end of program space
         mov     holdEnd,ax

; move word definition to program space

         cmp     bp,offset program	; test scan address
         js      xq59m                  ; if prompt mode, skip and do 
         jmp     xq59h                  ; go if program mode, don't move                 
xq59m:   call    fndSmc                 ; find semi-colon
         jc      xq59n

         call    movpl			; move into program space
         mov     byte ptr [di],NULL     ; store null terminator
xq59b:   mov     bx,temp                ; get pointer into wrdAdr
         mov     ax,offVal              ; pointer to execute this word
         add     ax,holdEnd             ; reference to old prgEnd
         mov     wrdAdr[bx],ax          ; store it for word execute 
         mov     bp,endAdr              ; get address next xeq
xq59l:   ret                            ; done with define word

wrd2big: dec     wrdX                   ; no room, forget word
         jmp     fil2big                ; relay to out of program space

xq59n:   call    movpl			; move this line into program spc
         mov     al,'+'                 ; show special prompt '+>'
         call    prnChr
         call    getPLn			; get next line
         mov     strAdr,offset buffer   ; set start address
         mov     nxtAdr,offset buffer   ;  and next adr - same as start now
         jmp     short xq59m		; go look for semi-colon in new line

; move colon def program lines

movpl:   mov     cx,endAdr		; check for room
         sub     cx,strAdr
         add     cx,3                   ; allow for cr/lf, null
         cmp     cx,maxPAdr
         ja      wrd2big                ; too big, go
         sub     cx,3
         mov     si,strAdr              ; move def to program area 
         mov     di,prgEnd              ; located at progam end 
xq59a:   mov     al,[si]                ; move loop
         mov     [di],al
         inc     si                     ; bump up pointers
         inc     di
         loop    xq59a                  ; loop till done
         mov     word ptr [di],LF*256+CR ; store CR/LF 
         inc     di                     ; allow for cr/lf 
         inc     di
         mov     prgEnd,di              ; update program end address 
	 ret				; done with move program lines
         
xq59e:   mov     di,offset owdsErr      ; here if out of fword space
         jmp     error                  ;   print error message 
xq59g:   dec     wrdX                   ; here if ; not found - forget this def
xq59g1:  mov     di,offset nscErr       ; point to error message
         jmp     error                  ;   print it

xq59j:   mov     cl,lbuf		; set up default end address
         xor     ch,ch			;  in case no ";" exists
         add     cx,strAdr
         inc     cx
         mov     endAdr,cx
         call    fndSmc                 ; here if word already exists - find ;
         cmp     bp,offset program      ; test bp
         jns     xq59j1                 ; print warning in prompt mode only
         mov     ah,forClr              ; save foreground color
         push    ax
         call    prnCRLF                ; print warning that word exists
         mov     di,offset token
         call    prnStr
         call    prnSPC
         mov     di,offset wddWrn
         call    prnMsg
         pop     ax                     ; restore color
         mov     forClr,ah
         call    prnCRLF                ; new line, done with warning
xq59j1:  mov     bp,endAdr              ; adjust pointer for new scan
         ret                            ; done with word exists warning

  
xq59h:   call    fndsmc                 ; here if program mode - find ;
         jc      xq59g                  ; error if not found
         mov     bx,temp                ; get pointer into wrdAdr
         mov     ax,nxtAdr              ; pointer to execute this word
         mov     wrdAdr[bx],ax          ; store it for word execute  
         mov     bp,endAdr              ; adjust pointer for new scan
         ret                            ; done with program mode

; find semicolon
fndSmc:  mov     si,nxtAdr              ; point to char after word name
         mov     al,';'                 ; search char ";"
fdsc02:  call    fndChr                 ; go find it
         pushf                          ; save cy flag
         inc     dx                     ; include ";"
         add     si,dx                  ; compute end address
         mov     endAdr,si              ; set end address      
         popf                           ; restore cy flag
         mov     bx,si                  ; PATCH
         jc      fdsc01                 ;   must have a spc, null or CR
         cmp     byte ptr [bx],SPC      ;   after ; 
         je      fdsc03
         cmp     byte ptr [bx],NULL
         je      fdsc03
         cmp     byte ptr [bx],CR
         je      fdsc03
         jmp     short fdsc02
fdsc03:  cmp     byte ptr [bx]-2,SPC    ;   and spc or LF before ;
         je      fdsc01 
         cmp     byte ptr [bx]-2,LF
         je      fdsc01 
         jmp     short fdsc02           ; END PATCH
fdsc01:  ret				; done with find semi-colon 

xeq060:  mov     bl,rStkX               ; ; - get run stack index
         or      bl,bl                  ; test it
         je      xq60a                  ; if zero, stack is empty - error
         dec     bl                     ; zero base it 
         dec     rstkX                  ; drop stack index
         xor     bh,bh                  ; it's OK, so make it a word                  
         sal     bx,1                   ; x 2
         mov     bp,rStack[bx]          ; get scan pos. off run stack
         ret                            ; go to main loop when all is done   
xq60a:   mov     di,offset RStkEmt      ; get run stack empty message
         jmp     error                  ;   print error meassage

xeq061:  jmp     mloop                  ; stop - go to prompt mode

xeq062:  mov     si,bp                  ; skip - set si to scan pos
         dec     si                     ; make sure not on CR
         dec     si
         call    popInt                 ; get # lines to skip
         or      ax,ax                  ; test it
         js      xq62c                  ; go if skip back
xq62a:   inc     si                     ; next char
         cmp     byte ptr [si],NULL     ; is it a terminator?
         je      xq62b                  ; go if so - error
         cmp     byte ptr [si],CR       ; is it eol?
         jne     xq62a                  ; keep looking if not
         dec     ax                     ; drop count
         jns     xq62a                  ; keep looking if postive
         jmp     short xq62d
xq62b:   mov     di,offset lnxErr       ; point to line not exist error
         jmp     error                  ;   go print it
xq62c:   dec     si                     ; previous char
         cmp     byte ptr [si],NULL     ; is it a terminator?
         je      xq62b                  ; go if so - error
         cmp     byte ptr [si],LF       ; is it bol?
         jne     xq62c                  ; keep looking if not
         inc     ax                     ; bump up count
         jne     xq62c                  ; keep looking if negative
xq62d:   cmp     si,prgEnd              ; more checks on range
         jns     xq62b
         cmp     si,offset subr
         js      xq62b
         mov     bp,si                  ; update scan
         ret                            ; done with skip

xeq063:  call    popInt                 ; do - get start value
         mov     cx,ax                  ; hold it 
         call    popInt                 ; get limit value
         mov     dx,ax                  ; hold it
         mov     bl,doStkX              ; get do stack index
         cmp     bl,doStkMax            ; check for overflow
         jae     doStkOv                ; go if stack overflow
         xor     bh,bh                  ; make index a word
         sal     bx,1                   ; times 2
         mov     ax,nxtAdr              ; get do address
         mov     doStack[bx],ax         ; put it on the stack
         mov     doIJKStk[bx],cx        ; put start value on stack
         mov     doLimStk[bx],dx        ; put limit value on satck
         inc     doStkX                 ; bump stack index
         ret                            ; done with do
doStkOv: mov     di,offset doStkOEr     ; here if do stk overflow - pt to msg
         jmp     error                  ; go print the error

xeq064:  mov     cx,1                   ; default step is 1
xq64a:   mov     bl,doStkX              ; loop - get do stack index
         or      bl,bl                  ; test it
         je      xq64b                  ; if zero, stack is empty - error
         dec     bl                     ; zero base it 
         xor     bh,bh                  ; make it a word                  
         sal     bx,1                   ; x 2
         mov     ax,doIJKStk[bx]        ; get do index var
         add     ax,cx                  ; add step
         cmp     ax,doLimStk[bx]        ; compare to limit - stop yet
         jg      xq64c                  ; go if index var > limit
         mov     doIJKStk[bx],ax        ; update do index var
         mov     bp,doStack[bx]         ; get loop back 
         ret                            ; done 
xq64b:   mov     di,offset nodoErr      ; here if no do
         jmp     error                  ;   print error message
xq64c:   dec     doStkX                 ; drop stack index
         ret                            ; done with loop

xeq065:  call    popInt                 ; +loop - get step value
         mov     cx,ax                  ; hold it in cx
         jmp     short xq64a            ; go do loop

xeq066:  mov     bx,offset doIJKstk     ; i - get stack value start addr
         mov     al,doStkX              ; get stack index
         xor     ah,ah                  ; make it a word
         dec     ax                     ; base it on zero
         sal     ax,1                   ; align with word
         add     bx,ax                  ; adjust pointer
         cmp     bx,offset doIJKstk     ; less than stack value start?
         js      xq66a                  ; means index doesn't exist
         mov     ax,[bx]                ; get i value
         jmp     pushInt                ; put on TOIS - done with i
xq66a:   mov     di,offset idxaErr      ; here if index var not avail.
         jmp     error                  ;   print error message

xeq067:  mov     bx,offset doIJKstk     ; j - get stack value start addr
         mov     al,doStkX              ; get stack index
         xor     ah,ah                  ; make it a word
         dec     ax                     ; base it on zero
         sal     ax,1                   ; align with word
         add     bx,ax                  ; adjust pointer
         dec     bx
         dec     bx
         cmp     bx,offset doIJKstk     ; less than stack value start?
         js      xq66a                  ; means index doesn't exist
         mov     ax,[bx]                ; get j value
         jmp     pushInt                ; put on TOIS - done with j

xeq068:  mov     bx,offset doIJKstk     ; k - get stack value start addr
         mov     al,doStkX              ; get stack index
         xor     ah,ah                  ; make it a word
         dec     ax                     ; base it on zero
         sal     ax,1                   ; align with word
         add     bx,ax                  ; adjust pointer
         sub     bx,4
         cmp     bx,offset doIJKstk     ; less than stack value start?
         js      xq66a                  ; means index doesn't exist
         mov     ax,[bx]                ; get k value
         jmp     pushInt                ; put on TOIS - done with k

xeq069:  mov     bl,doStkX              ; undo - get do stack index
         cmp     bl,1                   ; must have at least 1 do
         js      xq64b                  ; go if not
         dec     doStkX                 ; forget last do
         ret                            ; done with undo

xeq070:  mov     bl,baStkX              ; begin - get begin/again stk index
         cmp     bl,doStkMax            ; check for overflow
         jae     baStkOv                ; go if stack overflow
         xor     bh,bh                  ; make index a word
         sal     bx,1                   ; times 2
         inc     baStkX                 ; bump up index
         mov     ax,nxtAdr              ; get ba address
         mov     baStack[bx],ax         ; save on ba stack
         ret                            ; done with begin
baStkOv: mov     di,offset baStkOEr     ; here if do stk overflow - pt to msg
         jmp     error                  ; go print the error

xeq071:  mov     bl,baStkX              ; again - get ba stack index
         or      bl,bl                  ; test it
         je      xq71b                  ; if zero, stack is empty - error
         dec     bl                     ; zero base it 
         xor     bh,bh                  ; make it a word                  
         sal     bx,1                   ; x 2
         mov     bp,baStack[bx]         ; get loop back 
         ret                            ; done 
xq71b:   mov     di,offset nobaErr      ; here if no do
         jmp     error                  ;   print error message

xeq072:  mov     bl,baStkX              ; unbegin - get do stack index
         cmp     bl,1                   ; must have at least 1 dbegin/again
         js      xq71b                  ; go if not
         dec     baStkX                 ; forget last do
         ret                            ; done with undo

xeq073:  mov     di,offset pad          ; accept - point to pad
         mov     nxtNum,di
         mov     dl,80                  ; 80 chars max
         call    inpStr                 ; fill pad from keyboard
         mov     lpad,dl                ; set pad size
         ret                            ; done with accept

xeq074:  push    bp                     ; execute - save scan position
         mov     bp,offset pad          ; point parser to pad
         call    parse                  ; parse it
         pop     bp                     ; restore scan position
         ret                            ; done with execute

xeq075:  call    popInt                 ; emit - get TOIS
         jmp     prnChr                 ; print char equivalent of al

xeq076:  jmp     prnSPC                 ; space - go print a space

xeq077:  call    popInt                 ; spaces - get TOIS
         mov     cx,ax                  ; transfer to cx for count
xq77a:   call    prnSpc                 ; print a space
         loop    xq77a                  ; again till done
         ret                            ; done with spaces

xeq078:  call    popInt                 ; 1+ - get TOIS
         inc     ax                     ; increment
         jmp     pushInt                ; done with 1+ - put back TOIS

xeq079:  call    popInt                 ; 1- - get TOIS
         dec     ax                     ; decrement
         jmp     pushInt                ; done with 1- - put back TOIS

xeq080:  jmp     short xeq078           ; inc - same as 1+

xeq081:  jmp     short xeq079           ; dec - same as 1-

xeq082:  call    popFlt                 ; fint - get TOFS
         FSTCW   npucntl                ; store control word
         FWAIT
         or      npucntl,rndzbit        ; set for round to 0
         FLDCW   npucntl
         FRNDINT                        ; make an integer
         and     npucntl,rndzmsk        ; return to normal rounding
         FLDCW   npucntl
         FWAIT
         jmp     pushFlt                ; done with fint - put back TOFS

xeq083:  call    popFlt                 ; f>n - get TOFS
         FLD     fnhi                   ; compare to 32768.
         FCOMP
         FSTSW   npustat
         FWAIT
         and     npustat+1,1
         jne     xq83c                  ; go if f>=32768
         FLD     fnlo                   ; compare to -32769
         FCOMP
         FSTSW   npustat
         FWAIT
         and     npustat+1,1
         je      xq83c                  ; go if f<=32769
         FSTCW   npucntl                ; store control word
         FWAIT
         or      npucntl,rndzbit        ; set for round to 0
         FLDCW   npucntl
         FISTP   temp                   ; convert to int and store in temp
         FWAIT
         and     npucntl,rndzmsk        ; return to normal rounding
         FLDCW   npucntl
         FWAIT
         mov     ax,temp                ; get integer
         jmp     pushInt                ; done with f>n - put on TOIS
xq83c:   FSTP    ST(0)
         jmp     short xq86a

xeq084:  call    popInt                 ; n>f - get TOIS
         mov     temp,ax                ; save in temp
         FILD    temp                   ; convert to float and put in ST(0)
         jmp     pushFlt                ; done with n>f - put on TOFS

xeq085:  call    inpChr                 ; key - get char from KB
         xor     ah,ah                  ; make it an integer
         jmp     pushInt                ; done with key - put on TOIS

xeq086:  call    popFlt                 ; fsqrt - get TOFS
         FTST                           ; test ST(0)
         FSTSW   npustat                ; store stats
         FWAIT
         and     byte ptr npustat+1,01h ; ST(0) minus?
         jne     xq86a                  ; go if so - math error
         FSQRT                          ; take square root
         jmp     pushFlt                ; done with fsqrt - put sqrt on TOFS
xq86a:   mov     di,offset mathErr      ; point to math error message
         jmp     error                  ;   go print it

xeq087:  call    popFlt                 ; f10^x - get TOFS
         call    exp10                  ; do 10^x
         jmp     pushFlt                ; done with f10^x - put back TOFS

xeq088:  call    popFLT                 ; fe^x - get TOFS
         FLD     log10e                 ; log10e,x
         FMUL                           ; x*log10e
         call    exp10                  ; e^x=10^(x*log10e)
         jmp     pushFlt                ; done with fe^x - put back TOFS

xeq089:  call    popFlt                 ; flog10 - get TOFS
         FTST                           ; test ST(0)
         FSTSW   npustat                ; save status
         FWAIT
         and     byte ptr npustat+1,40h ; trap negative or zero
         jne     xq86a                  ; go print math error
         call    log10                  ; do log 10
         jc      xq86a                  ; go for math error
         jmp     pushFlt                ; done with Flog10 - put back TOFS
           
xeq090:  call    popFlt                 ; fln - get TOFS
         FTST                           ; test ST(0)
         FSTSW   npustat                ; save status
         FWAIT
         and     byte ptr npustat+1,41h ; trap negative or zero
         jne     xq86a                  ; go print math error
         call    log10                  ; do log10
         jc      xq86a                  ; go for math error
         FLD     loge10                 ; loge10,X
         FMUL                           ; loge(x) = loge(10) * log10(x)
         jmp     pushFlt                ; put back TOFS

xeq091:  call    popFlt                 ; ftan - get TOFS
         call    FTAN                   ; do full tangent
         jc      xq91a                  ; error trap
         jmp     pushFlt                ; put back on TOFS
xq91a:   jmp     xq86a                  ; go print math error

xeq092:  FLDPI                          ; pi - make ST(0) pi
         jmp     pushFlt                ; put on TOFS - done with pi

xeq093:  call    popFlt                 ; fd2r - get TOFS
         FLD     fd2r                   ; load conversion factor
         FMUL                           ; change degrees to radians
         jmp     pushFlt                ; put on TOFS - done with fd2r

xeq094:  call    popFlt                 ; fr2d - get TOFS
         FLD     fr2d                   ; load conversion factor
         FMUL                           ; change radians to degrees
         jmp     pushFlt                ; put on TOFS - done with fr2d

xeq095:  call    popFlt                 ; fatan - get TOFS
         call    FATAN                  ; do arc tan
         jmp     pushFlt                ; put on TOFS - done with fatan           

xeq096:  call    popFlt                 ; fsin - get TOFS
         call    FFSIN                  ; do sin
         jmp     pushFlt                ; put on TOFS - done with fsin

xeq097:  call    popFlt                 ; fcos - get TOFS
         call    FFCOS                  ; do cos
         jmp     pushFlt                ; put on TOFS - done with fcos

xeq098:  call    popInt                 ; abs - get TOIS
         or      ax,ax                  ; test integer
         jns     xq98a                  ; go if postive
         neg     ax                     ; make plus
xq98a:   jmp     pushInt                ; put back on TOIS - done with abs

xeq099:  call    popFlt                 ; fabs - get TOFS
         FABS                           ; take absolute value
         jmp     pushFlt                ; put back on TOFS - done with fabs

xeq100:  call    popInt                 ; neg - get TOIS
         neg     ax                     ; negate
         jmp     pushInt                ; put back on TOIS - done with neg

xeq101:  call    popFlt                 ; fneg - get TOFS
         FCHS                           ; negate
         jmp     pushFlt                ; put back of TOFS - done with fneg 

xeq102:  mov     fStkX,0                ; fempty - set flt stack index to 0
         ret                            ; done with fempty

xeq103:  mov     iStkX,0                ; empty - set int stack index to 0
         ret                            ; done with empty 

xeq104:  mov     al,0                   ; clean - clear both stack indices
         mov     fStkX,al
         mov     iStkX,al
         mov     pStkX,al               ; clear push stacks
         mov     fpStkX,al
         ret                            ; done with clean

xeq105:  mov     al,iStkX               ; depth - get int stack size
xq105a:  xor     ah,ah                  ; make it a word
         jmp     pushInt                ; put on TOIS - done with depth

xeq106:  mov     al,fStkX               ; fdepth - get flt stack size
         jmp     short xq105a           ; go put it on TOIS - done fdepth 

xeq107:  call    popInt                 ; places - get TOIS
         or      ax,ax                  ; take abs
         jns     xq107a
         neg     ax                         
xq107a:  mov     bx,10                  ; divisor = 10
         xor     dx,dx
         div     bx                     ; |x|/10
         mov     iFmt,dl                ; remainder is integer format
         ret                            ; done with places

xeq108:  call    popInt                 ; fplaces - get TOIS
         or      ax,ax                  ; take abs
         jns     xq108a
         neg     ax
xq108a:  mov     bx,10                  ; |x|/10
         xor     dx,dx
         div     bx
         mov     cx,dx                  ; save decimals
         xor     dx,dx
         div     bx                     ; |x|/100
         add     dx,cx                  ; add two formats
         mov     fmtWid,dl              ; save as width
         mov     fmtDec,cl              ; save decimals            
         ret                            ; done with fplaces

xeq109:  call    popInt                 ; fgcolor - get TOIS
         mov     forClr,al              ; set forground color
         ret                            ; done with fgcolor

xeq110:  call    popInt                 ; bgcolor - get TOIS
         mov     bakClr,al              ; set background color
         ret                            ; done with background color

xeq111:  call    popInt                 ; incolor - get TOIS
         mov     inpClr,al              ; set input routine text color
         ret                            ; done with incolor

xeq112:  jmp     cold                   ; cold

xeq113:  xor     bx,bx                  ; list - zero line count
         mov     cx,maxPSz              ; get max program size
         mov     di,offset program      ; set di to start of program space
xq113a:  mov     al,[di]                ; get a char
         or      al,al                  ; test it
         je      xq113c                 ; quit when terminator found
         call    prnChr                 ; print it
         test    gflag,1                ; text mode?
         jne     xq113b                 ; go if not
         call    getCur                 ; get cursor
         cmp     dl,0                   ; newlined?
         jne     xq113b                 ; go if not
         cmp     al,LF                  ; line feed?
         je      xq113b                 ; don't count them
         inc     bx                     ; count CRs and overflows
         cmp     bx,24                  ; at bottom yet?                 
         jl      xq113b                 ; go if not
         xor     bx,bx                  ; zero line count
         call    inpChr                 ; wait for key
xq113b:  inc     di                     ; next char
         loop    xq113a                 ; keep listing
xq113c:  ret                            ; done with list                          ; done with list

xeq114:  call    getToke                ; load - get filename
         cmp     byte ptr [di],NULL     ; no name?
         je      xq114e                 ; go use default if so  
         cmp     cx,12                  ; check number of chars
         ja      xq114a                 ; error if too many
         jmp     short xq114d           ; go move fn and open
xq114e:  mov     di,offset defFn        ; get default file name
xq114d:  mov     si,offset fname        ; move token name to filename
         call    movStr
         call    closTee		; stop teed output
         mov     prnTee,0
xq114x:  mov     di, offset fname       ; PATCH if no ext, add '.4'
xq114y:  cmp     byte ptr [di],'.'      ; search for dot
         je      xq114z                 ; go if found
         cmp     byte ptr [di],NULL     ; if null found before dot, no ext
         je      xq114w                 ; go set default ext
         inc     di                     ; else keep looking
         jmp     short xq114y
xq114w:  mov     word ptr [di],'4.'     ; store extension
         inc     di 
         inc     di
         mov     byte ptr [di],NULL     ; new terminator END PATCH
xq114z:  call    openFil                ; open the file
         call    fileSiz                ; get its size
         call    cold2                  ; init
xq114c:  mov     dx,prgEnd              ; load at PrgEnd
         call    readFil                ; read in the file to program space
         call    closFil                ; close the file
         mov     di,PrgEnd              ; update PrgEnd pointer
         push    di                     ; save as start for scan
         add     di,fsize
         mov     byte ptr [di],Null     ; store the terminator
         mov     PrgEnd,di         
         mov     di,offset fname        ; print filename: message
         call    prnStr
         mov     al,':'
         call    prnChr
         mov     ax,fsize               ; print file size
         call    prnInt
         mov     di,offset ldfMsg       ; print bytes loaded
         call    prnStr         
         pop     bp                     ; set scan pointer
         ret                            ; done with load
xq114a:  jmp     openFal                ; go do open failure
xq114b:  mov     di,offset flnErr       ; here if file name bad
         jmp     error                  ;  go print error message

xeq115:  call    getToke                ; +load - get filename
         cmp     byte ptr [di],NULL     ; no name? 
         je      xq114b                 ; error if so 
         cmp     cx,11                  ; check number of chars
         ja      xq114a                 ; error if too many
         mov     si,offset fname        ; move token name to filename
         call    movStr
         call    closTee                ; stop teed output
         mov     prnTee,0
         call    openFil                ; open the file
         call    fileSiz                ; get its size
         push    bp                     ; save scan position
         call    xq114c                 ; load file
         call    parse                  ; parse it
         pop     bp                     ; restore scan position
         ret                            ; done with +load

prnMsgc: mov     ah,forClr              ; save color parm
	 mov     tmpClr,ah              ; print message and restore color
         call    prnMsg
         mov     ah,tmpClr              ; restore color
         mov     forClr,ah
         ret                            ; done with print message ic color

xeq116:  mov     di,offset rutwMsg      ; words - root words first
         call    prnMsgc                ; print message in color
         mov     cl,nDct                ; get number of FLIP words
         xor     ch,ch                  ; make it a 16 bit word
         xor     bl,bl                  ; zero line count
         mov     di,offset Dict         ; point to root dictionary
xq116a:  call    prnStr                 ; print a word & space             
         call    prnSpc
         test    gflag,1                ; in graphics mode?
         je      xq116g                 ; go if not
         cmp     gcurx,540              ; check for last char col
         jle     xq116b                 ; go if not at eol
         call    prnCRLF                ; newline if so
         inc     bl                     ; inc line count
         cmp     bl,34                  ; at bol yest?
         jle     xq116b                 ; go if not
         jmp     short xq116k           ; go wait for key
xq116g:  call    getCur                 ; get text cursor position
         cmp     dl,70                  ; check for past col 70
         jb      xq116b                 ; not at eol yet
         call    prnCRLF                ; newline otherwise
         inc     bl                     ; inc line count    
         cmp     bl,23                  ; 23 lines yet?
         jb      xq116b                 ; go if not
xq116k:  call    inpChr                 ; wait for key 
         xor     bl,bl                  ; zero line count        
xq116b:  loop    xq116a                 ; loop for more words
         call    prnCRLF                ; all words printed, newline 
         inc     bl                     ; inc line count
         mov     di,offset usrwMsg      ; next print user words
         call    prnMsgc
         inc     bl                     ; inc line count
         mov     di,offset wrdnams      ; point to user word name table
         mov     cl,wrdX                ; get number of user words
         or      cl,cl                  ; test it
         je      xq116f
xq116c:  mov     ch,8                   ; print 8 chars of word name
xq116d:  mov     al,[di]
         call    prnChr
         inc     di
         dec     ch
         jne     xq116d
         call    prnSPC                 ; print a space after word 
         test    gflag,1                ; in graphics mode?
         je      xq116l                 ; go if not
         cmp     gcurx,540              ; check for last char col
         jle     xq116e                 ; go if not at eol
         call    prnCRLF                ; newline if so
         inc     bl                     ; inc line count
         cmp     bl,34                  ; at bol yet?
         jle     xq116e                 ; go if not
         jmp     short xq116n           ; go wait for key
xq116l:  call    getCur                 ; get text cursor position
         cmp     dl,70                  ; check for past col 70
         jb      xq116e                 ; not at eol yet
xq116m:  call    prnCRLF                ; newline otherwise
         inc     bl                     ; inc line count
         cmp     bl,23                  ; 23 lines yet?
         jb      xq116e                 ; go if not
xq116n:  call    inpChr                 ; wait for key
         xor     bl,bl                  ; zero line count
xq116e:  dec     cl
         jne     xq116c
         call    prnCRLF                ; final newline
         ret                            ; done 
xq116f:  mov     di,offset nunMsg       ; print none
         jmp     prnStr                 ; done with words

xeq117:  call    popInt                 ; h. - get TOIS
         jmp     prnHex                 ; go print hex value

xeq118:  call    popInt                 ; at-xy - get TOIS
         mov     bx,ax                  ; save it
         call    popInt                 ; get next stack value
         test    gflag,1                ; test for graphics mode
         jne     xq118a                 ; go if graphics mode set
         mov     dh,bl                  ; text row
         mov     dl,al                  ; text col
         jmp     setCur                 ; go set cursor location - done
xq118a:  mov     gcury,bx               ; here if graphics mode - set y pos.
         mov     gcurx,ax               ; set x position              
         ret                            ; done with at-xy

xeq119:  mov     ax,TRUE                ; true - load true value
         jmp     pushINT                ; put it on TOIS - done with true

xeq120:  mov     ax,FALSE               ; false - load false value
         jmp     pushInt                ; put it on TOIS - done with false

xeq121:  call    xeq140                 ; bye - set to text mode
         jmp     exit                   ; exit to dos

xeq122:  mov     ax,offset endPM        ; mem - compute memory space available
         sub     ax,PrgEnd
         inc     ax
         jmp     pushInt                ; put on TOIS - done with mem

xeq123:  call    popInt                 ; allot - get TOIS
         mov     bl,varX                ; get variable space index
         xor     bh,bh                  ; make it a word
         or      ax,ax                  ; test request
         js      xq123a                 ; negative request is an error
         add     bx,ax                  ; add request
         or      bh,bh                  ; test for > 255
         jne     xq123a                 ; it is, so error
         cmp     bl,varMax              ; check for space
         ja      xq123a                 ; too much - error
         mov     varX,bl                ; update variable space index
         ret                            ; done with allot
xq123a:  mov     di,offset ovsErr       ; here if variable space error
         jmp     error                  ;   go print out of variable space

xeq124:  call    popInt                 ; allot - get TOIS
         mov     bl,fvarX               ; get float variable space index
         xor     bh,bh                  ; make it a word
         or      ax,ax                  ; test request
         js      xq124a                 ; negative request is an error
         add     bx,ax                  ; add request
         or      bh,bh                  ; test for > 255
         jne     xq124a                 ; it is, so error
         cmp     bl,fvarMax             ; check for space
         ja      xq124a                 ; too much - error
         mov     fvarX,bl               ; update variable space index
         ret                            ; done with allot
xq124a:  mov     di,offset ofvsErr      ; here if variable space error
         jmp     error                  ;   go print out of variable space

xeq125:  call    popInt                 ; push - get TOIS
xq125a:  mov     bl,pStkX               ; get push stack index (0 - 2)
         cmp     bl,3                   ; check for overflow, index >= 3
         jae     pStkOvr                ; go if stack overflow
         xor     bh,bh                  ; make index a word
         sal     bx,1                   ; times 2
         mov     pStack[bx],ax          ; put the integer on the stack
         inc     pStkX                  ; bump stack index
         ret                            ; done with push 
pStkOvr: mov     di,offset pStkOEr      ; here if int stk overflow - pt to msg
         jmp     error                  ;   go print the error 

xeq126:  mov     bl,pStkX               ; pop - get push stack index (0 - 2)
         or      bl,bl                  ; test for 0
         je      pStkEmt                ; stack empty if so
         dec     bl                     ; drop index
         mov     pStkX,bl               ; save it back
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; times 2
         mov     ax,pStack[bx]          ; get stack element into ax reg
         jmp     pushInt                ; pop integer - done
pStkEmt: mov     di,offset pStkEEr      ; here if p stk empty - pt to err msg
         jmp     error                  ;   go print the error

xeq127:  call    popFlt                 ; fpush - get TOFS
xq127a:  mov     bl,fpStkX              ; get float stack index (0 - 2)
         cmp     bl,3                   ; check for overflow, index >= 3
         jae     fpStkOv                ; go if stack overflow
         xor     bh,bh                  ; make index a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FSTP    fpStack[bx]            ; put the integer on the stack
         FWAIT
         inc     fpStkX                 ; bump stack index
         ret                            ; done with push integer
fpStkOv: mov     di,offset fpStkOEr     ; here if int stk overflow - pt to msg
         jmp     error                  ;   go print the error description

xeq128:  mov     bl,fpStkX              ; fpop - get fp stack index (0 - 2)
         or      bl,bl                  ; test for 0
         je      fpStkEm                ; stack empty if so
         dec     bl                     ; drop index
         mov     fpStkX,bl              ; save it back
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fpStack[bx]            ; get stack element into ST(0) reg
         jmp     pushFlt                ; go put on TOFS
fpStkEm: mov     di,offset fpStkEEr     ; here if int stk empty - pt to err msg
         jmp     error                  ; go print the error description

xeq129:  call    popFlt                 ; epsilon! - get TOFS
         FSTP    epsilon                ; store it 
         FWAIT
         ret                            ; done with epsilon!

xeq130:  FLD     epsilon                ; epsilon@ - get it
         jmp     pushFlt                ; put on TOFS - done with epsilon@

xeq131:  call    getToke                ; fconstant - get name
         call    toke8                  ; pad or pare to 8 chars
         call    is1alf                 ; make UC and check validity
         je      xq131d                 ; if null, no name - error
         jc      xq131d                 ; if carry, name invalid
         mov     di,offset token        ; move token to temp
         mov     si,offset tname
         call    movStr
         call    fndFvrn                ; name used as an fvariable name?
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndVarn                ; name used as a variable name?
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndWrd                 ; find word name
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndCon                 ; is it a constant name
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndFcon                ; find this name
         jnc     xq131c                 ; if found, fcon exists, do nothing
         cmp     fConX,conMax           ; can we create another one?
         je      xq131e                 ; go if no, fcon space full
         call    getToke                ; get the value
         cmp     byte ptr [di],NULL     ; no token if null
         je      xq131g                 ; error if so
         call    asc2dec                ; try to convert to float
         inc     di                     ; test flag
         je      xq131g                 ; failed to find float
         call    dec2Flt                ; float in ST(0)         
         inc     fConX                  ; bump up index
         mov     bl,fConX               ; put it in bl
         xor     bh,bh                  ; make it a word
         dec     bx                     ; zero base it
         sal     bx,1
         mov     ax,bx
         sal     bx,1
         sal     bx,1                   ; x8
         mov     di,bx
         add     di,ax                  ; x 10
         FSTP    Fcons[di]              ; put float on fconstant stack
         mov     si,offset fconnams     ; set pointers
         mov     di,offset tname
         jmp     nmMov                  ; go move name - done with fconstant
xq131d:  jmp     xq28d                  ; relay for name error
xq131f:  mov     di,offset cdupErr      ; here if con or fcon name used
         jmp     error                  ;   print error message
xq131e:  mov     di,offset ofcnsEr      ; here if out of fcon space
         jmp     error                  ;   print error message 
xq131g:  mov     di,offset fcnmErr      ; here if invalid constant
         jmp     error                  ;   print error message
xq131c:  jmp     getToke                ; here if fcon redef, ignore token

xeq132:  call    getToke                ; constant - get name
         call    toke8                  ; pad or pare to 8 chars
         call    is1alf                 ; make UC and check validity
         je      xq131d                 ; if null, no name - error
         jc      xq131d                 ; if carry, name invalid
         mov     di,offset token        ; move token to temp
         mov     si,offset tname
         call    movStr          
         call    fndFvrn                ; name used as an fvariable name?
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndVarn                ; name used as a variable name?
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndWrd                 ; find word name
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndFCon                ; is it an fconstant name
         jnc     xq131f                 ; if so then it's a duplicate
         call    fndCon                 ; find this name
         jnc     xq131c                 ; if found, con exists, do nothing
         cmp     ConX,conMax            ; can we create another one?
         je      xq132e                 ; no - error
         call    getToke                ; get the value
         cmp     byte ptr [di],NULL     ; no token if null
         je      xq132g                 ; error if so
         call    asc2int                ; try to convert to integer
         inc     di                     ; test flag
         je      xq132g                 ; failed to find integer
         inc     conX                   ; bump up index
         mov     bl,conX                ; put it in bl
         xor     bh,bh                  ; make it a word
         dec     bx                     ; zero base it
         sal     bx,1                   ; x 2
         mov     cons[bx],ax
         sal     bx,1
         sal     bx,1                   ; x 8
         mov     si,offset connams      ; set pointers
         mov     di,offset tname
         jmp     nmMov                  ; go move name
xq132e:  mov     di,offset ocnsErr      ; here if out of fcon space
         jmp     error                  ;   print error message 
xq132g:  mov     di,offset cnmErr       ; here if invalid constant
         jmp     error                  ;   print error message
         ret                            ; done with constant

xeq133:  mov     ah,forClr              ; vlist - save fg color
         mov     tmpClr,ah
         mov     temp,0                 ; zero line count
         mov     si,offset vrlMsg       ; print 'variables:'
         mov     ch,VarX                ; get count of variables
         mov     di,offset vnams        ; point to variable names
         call    xq133e                 ; print table
         mov     si,offset fvrlMsg      ; print 'fvariables:
         mov     ch,fVarX               ; get count of fvariables
         mov     di,offset fvnams       ; point to fvar names
         call    xq133e                 ; print table
         mov     si,offset cnlMsg       ; print 'constants'
         mov     ch,ConX                ; get counr of constants
         mov     di,offset connams      ; point to con names
         call    xq133e                 ; print table
         mov     si,offset fcnlMsg      ; print 'fconstants'
         mov     ch,fConX               ; get counr of fconstants
         mov     di,offset fconnams     ; point to fcon names
         call    xq133e                 ; print table
         mov     ah,tmpClr              ; restore color
         mov     forClr,ah
         ret                            ; done with vlist

xq133e:  push    di                     ; print table
         mov     di,si                  ; save di
         call    prnMsgc                ; print msg in color
         inc     temp                   ; inc line count
         pop     di                     ; restore di
         or      ch,ch                  ; test var count
         jne     xq133b                 ; skip print if not zero
         mov     di,offset nunMsg       ; point to --none-- msg
         jmp     prnStr                 ; print it
xq133b:  xor     cl,cl                  ; zero column counter
         xor     bx,bx                  ; zero offset into name table
         mov     al,[di]                ; get tag char
         cmp     al,SPC                 ; space?
         jne     xq133c                 ; skip if not
         mov     al,250                 ; ascii small dot
         call    prnChr                 ; print it
         inc     cl                     ; count it
         jmp     short xq133d           ; go check for more tags
xq133c:  mov     al,[bx+di]             ; it's not a space, get char again
         inc     bx                     ; bump offset
         cmp     al,SPC                 ; this time space terms the tag
         je      xq133d                 ; go if it is
         call    prnChr                 ; else print good char
         inc     cl                     ; count it
         jmp     short xq133c           ; get rest of tag
xq133d:  call    prnSPC                 ; here when tag printed
         inc     cl                     ; count it
         test    gflag,1                ; text mode?
         je      xq133d1                ; go if so
         cmp     cl,56                  ; in grafix mode col<=56
         jle     xq133f
         jmp     short xq133d2
xq133d1: cmp     cl,70                  ; text cursor col <=70?
         jle     xq133f                 ; skip new line if so
xq133d2: call    prnCRLF                ; else issue newline
         inc     temp                   ; inc line count
         xor     cl,cl                  ; clear column count
xq133f:  add     di,8                   ; point to next tag
         dec     ch                     ; drop tag count
         jne     xq133b                 ; go if more 
         call    prnCRLF                ; newline at end of list         
         inc     temp                   ; inc line count
         test    gflag,1                ; text mode?
         je      xq133f1                ; go if so
         cmp     temp,34                ; 34 lines yet?
         jb      xq133g                 ; no, done
         jmp     short xq133f2          ; else go wait for key         
xq133f1: cmp     temp,23                ; 23 lines yet?
         jb      xq133g                 ; no, done
xq133f2: call    inpChr                 ; else wait for key
         mov     temp,0                 ; zero line count
xq133g:  ret                            ; done with print table       

xeq134:  call    popInt                 ; , - get TOIS
         mov     bl,vadr                ; get current var index
         or      bl,bl                  ; is it zero?
         je      xq134a                 ; error if so
         dec     bl                     ; zero base it
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; x 2
         mov     vars[bx],ax            ; store int
         inc     vadr                   ; bump up current var index
         ret                            ; done with ,
xq134a:  mov     di,offset cmaErr       ; point to comma error 
         jmp     error                  ;   print message      
        
xeq135:  call    popFlt                 ; f, - get TOFS
         mov     bl,fvadr               ; get current fvar index
         or      bl,bl                  ; is it zero?
         je      xq134a                 ; error if so
         dec     bl                     ; zero base it
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast x 10
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FSTP    fvars[bx]              ; store float
         inc     fvadr                  ; bump up current fvar index
         ret                            ; done with f,

xeq136:  and     prnTee,1               ; tee-on - test flag
         jne     xq136d                 ; do nothing if set
         call    openTee                ; else open output.asc file
         mov     prnTee,1               ; set flag on
xq136d:  ret                            ; done with tee-on

xeq137:  and     prnTee,1               ; tee-off - test flag
         je      xq136d                 ; do nothing if off
         call    closTee                ; else close ouput.asc file
         mov     prnTee,0               ; set flag off
         ret                            ; done with tee-off

xeq138:  mov     ah,0Bh                 ; graphics, set bg to black
         mov     bx,0
         int     BIOS
         mov     ax,0010h               ; set video mode 640X350x16
         mov     gflag,1                ; set graphics flag
         int     BIOS
         call    ghome                  ; set cursor home 
         ret                            ; done with graphics

xeq139:  test    gflag,1                ; plot - graphics mode?
         je      xq139a                 ; go if not
         call    popint                 ; set y pixel
         mov     dx,ax
         mov     Y1,ax                  ; set last plot y pixel
         call    popint                 ; set x pixel
         mov     cx,ax
         mov     X1,ax                  ; set last plot x pixel
         mov     ah,0Ch                 ; write pixel call
         mov     al,forClr              ; use foreground color
         or      al,xorGrfx             ; set xor grafix bit
         mov     bh,0                   ; set video page 
         int     BIOS                   ; set pixel
         ret                            ; done with plot
xq139a:  mov     di,offset pltErr       ; here if not in graphics mode
         jmp     error                  ; go print message

xeq140:  mov     ax,0003h               ; text - set mode 3 80x25 color text
         mov     gflag,0                ; clear gpaphics flag
         int     BIOS
         call    cls                    ; clear screen
         ret                            ; done with text
 
xeq141:  mov     bl,wrdX		; forget - get word index
         or      bl,bl                  ; test it
         je      xq141d                 ; no words to forget
         xor     bh,bh                  ; make index a word
         dec     bx                     ; zero base it
         sal     bx,1                   ; x2
         mov     bx,wrdAdr[bx]          ; get word address
xq141a:  mov     al,byte ptr [bx]       ; get char at address
         cmp     al,':'			; looking for :
         je      xq141b                 ; go if found
         or      al,al                  ; null?
         je      xq141b                 ; should never happen
         dec     bx                     ; look at previous char
         jmp     xq141a                 ; loop back
xq141b:  mov     byte ptr [bx],NULL     ; when : found, store null
         mov     prgEnd,bx              ; set program end address
         dec     wrdX                   ; then forget word
         ret                            ; done with forget

xq141d:  mov     di,offset frgtEr       ; here if no words
         jmp     error                  ; go print message 

xeq142:  push    bp                     ; fatan2 - save scan address
         mov     bp,offset fatan2       ; set scan to fatan2 sub
         call    parse                  ; execute it
         pop     bp                     ; restore scan address
         ret                            ; done with fatan2
   
xeq143:  call    popInt                 ; mod - get TOIS
         push    ax                     ; save divisor
         call    popInt                 ; get dividend
         or      ax,ax                  ; test it
         js      xq143b                 ; go if negative
         xor     dx,dx                  ; set dx to 0
xq143a:  pop     bx                     ; bx is divisor
         or      bx,bx                  ; tets divisor
         je      xq143c                 ; math error if zero
         idiv    bx                     ; divide
         mov     ax,dx                  ; get remainder
         jmp     pushInt                ; put on int stack - done with mod
xq143b:  mov     dx,-1                  ; here if div. minus - extend
         jmp     short xq143a           ; and go divide 
xq143c:  jmp     xq86a                  ; go print math error 

; test for ST(0) in neighboorhood of 1
; return with cy set if it is    
nh1tst:  FLD     tst1lo                 ; 1-eps,x
         FCOMP                          ; x:1-eps
         FSTSW   npustat                ; save stats
         FWAIT   
         and     npustat+1,1            ; x<1-eps?
         je      nh1ta                  ; go if so, not in n.h.
         FLD     tst1hi                 ; 1+eps,x
         FCOMP                          ; x:1+eps
         FSTSW   npustat                ; save stats
         FWAIT
         and     npustat+1,1            ; x<1+eps?
         jne     nh1ta                  ; go if not, not in n.h.
         stc                            ; in n.h., set cy
nh1ta:   ret                            ; done with neighboor of 1 test

xeq144:  call    popFlt                 ; facos - get TOFS
         FLD     ST(0)                  ; x,x
         FABS                           ; |x|,x
         call    nh1tst                 ; X in n.h. of 1?
         jc      xq144d                 ; go if so
         FSTP    ST(0)                  ; x
         FLD     ST(0)                  ; x,x
         FMUL    ST(0),ST(1)            ; x^2,x
         FLD1                           ; 1,x^2,x
         FSUBR                          ; 1-x^2,x
         FTST                           ; test 1-x^2
         FSTSW   npustat                ; save stats
         FWAIT
         and     npustat+1,1            ; neg?
         jne     xq144c                 ; go if so - math error
         FSQRT                          ; sqrt(1-x^2),x
         FDIV                           ; x/sqrt(1-x^2)
         call    FATAN                  ; atn(x/sqrt(1-x^2))
         FLD     fpid2                  ; pi/2,atn(x/sqrt(1-x^2))
         FSUBR                          ; acos(x)
         jmp     pushFlt                ; put on TOFS - done
xq144d:  FSTP    ST(0)
         FTST                           ; test x
         FSTSW   npustat                ; save stats
         FWAIT
         and     npustat+1,1            ; x -1?
         je      xq144b                 ; go if not
xq144a:  FSTP    ST(0)                  ; here if x=-1, clear 8087 stack
         FLDPI                          ; acos(-1)=pi
         jmp     pushFlt                ; put on TOFS - done        
xq144b:  FSTP    ST(0)                  ; here if x=1, clear 8087 stack
         FLDZ                           ; acos(1)=0
         jmp     pushFLT                ; put on TOFS - done with facos
xq144c:  jmp     xq86a                  ; jump to math error                                
                     
xeq145:  call    popFlt                 ; fasin - get TOFS
         FLD     ST(0)                  ; x,x
         FABS                           ; |x|,x
         call    nh1tst                 ; X in n.h. of 1?
         jc      xq145d                 ; go if so
         FSTP    ST(0)                  ; x
         FLD     ST(0)                  ; x,x
         FMUL    ST(0),ST(1)            ; x^2,x
         FLD1                           ; 1,x^2,x
         FSUBR                          ; 1-x^2,x
         FTST                           ; test 1-x^2
         FSTSW   npustat                ; save stats
         FWAIT
         and     npustat+1,1            ; neg?
         jne     xq145c                 ; go if so - math error
         FSQRT                          ; sqrt(1-x^2),x
         FDIV                           ; x/sqrt(1-x^2)
         call    FATAN                  ; atn(x/sqrt(1-x^2))
         jmp     pushFlt                ; put on TOFS - done
xq145d:  FSTP    ST(0)
         FTST                           ; test x
         FSTSW   npustat                ; save stats
         FWAIT
         and     npustat+1,1            ; x -1?
         je      xq145b                 ; go if not
xq145a:  FSTP    ST(0)                  ; here if x=-1, clear 8087 stack
         FLDZ                           ; asin(-1)=-pi/2
         FLD     fpid2
         FSUB
         jmp     pushFlt                ; put on TOFS - done        
xq145b:  FSTP    ST(0)                  ; here if x=1, clear 8087 stack
         FLD     fpid2                  ; asin(1)=pi/2
         jmp     pushFLT                ; put on TOFS - done with fasin
xq145c:  jmp     xq86a                  ; jump to math error    

xeq146:  call    popFlt                 ; fpwr - ST(0) = y
         FSTP    fptemp                 ; store y
         FWAIT
         call    popFlt                 ; ST(0) = x
	 FTST                           ; test x
         FSTSW   npustat                ; save status
         FWAIT
         test    byte ptr npustat+1,1   ; x neg?
         jne     xq146b                 ; go print math error if so
         test    byte ptr npustat+1,40h ; x zero?
         jne     xq146a                 ; go set x^y to 0 if so
         call    log10                  ; do log10
         jc      xq146c                 ; go if log10 error
         FLD     fptemp                 ; y,log10(x)                
         FMUL                           ; y*log10(x)
         call    exp10                  ; x^y=10^(ylog(x)) on TOFS
         jmp     pushFlt                ; x^y to TOFS
xq146a:  FSTP    ST(0)                  ; here if x zero, drop 8087 stack
         FLDZ                           ; set result to 0
         jmp     pushFlt                ; push it on TOFS
xq146b:  FSTP    ST(0)                  ; here if x neg - drop stack
xq146c:  jmp     xq86a			; go print math error - done with fpwr
 
xeq147:  call    popFlt                 ; fcubrt - ST(0) = x
	 FTST                           ; test x
         FSTSW   trgsgn                 ; save status
         FWAIT
         test    byte ptr trgsgn+1,40h  ; x zero?
         jne     xq146a                 ; go set cubrt to zero if so    
         test    byte ptr trgsgn+1,1    ; x neg?
         je      xq147a                 ; skip if plus
         FABS                           ; |x|
xq147a:  call    log10                  ; log10(|x|)
         jc      xq146c                 ; go if log10 error
         FLD     f1d3                   ; 1/3,log10(|x|)
         FMUL                           ; 1/3*log10(|x|)
         call    exp10                  ; cubrt(|x|)
         test    byte ptr trgsgn+1,1    ; check again for neg
         je      xq147b                 ; go if plus
         FCHS                           ; cubrt(x)
xq147b:  jmp     pushFlt                ; push on TOFS - done fcubrt
         
xeq148:  push    bp                     ; fsinh - save scan address
         mov     bp,offset fsinh        ; set scan to fsinh sub
         call    parse                  ; execute it
         pop     bp                     ; restore scan address
         ret                            ; done with fsinh

xeq149:  push    bp                     ; fcosh - save scan address
         mov     bp,offset fcosh        ; set scan to fcosh sub
         call    parse                  ; execute it
         pop     bp                     ; restore scan address
         ret                            ; done with fcosh

xeq150:  push    bp                     ; ftanh - save scan address
         mov     bp,offset ftanh        ; set scan to ftanh sub
         call    parse                  ; execute it
         pop     bp                     ; restore scan address
         ret                            ; done with ftanh

xeq151:  push    bp                     ; fatanh - save scan address
         mov     bp,offset fatanh       ; set scan to fatanh sub
         call    parse                  ; execute it
         pop     bp                     ; restore scan address
         ret                            ; done with fatanh

xeq152:  push    bp                     ; line - save scan address
         mov     bp,offset line         ; set scan to draw line routine
         call    parse                  ; do it
         pop     bp                     ; restore scan address
         ret                            ; done with line

xeq154:  push    bp                     ; number - save scan address
         mov     bp,nxtNum              ; point to next number in pad
         call    getToke                ; get a token from pad
         mov     si,offset token        ; point to token
         call    asc2ina                ; convert token to an int number
         mov     bx,nxtAdr              ; set next number address
         mov     nxtNum,bx
         pop     bp                     ; restore scan address
         or      di,di                  ; test bad number flag
         js      xq154a                 ; go if not an integer
         jmp     pushInt                ; put it on stack - done with number
xq154a:  mov     di,offset intWrn       ; warning - not an integer
         jmp     error                  ; go print message
         
xeq155:  push    bp			; fnumber - save scan address
         mov     bp,nxtNum              ; point to next number in pad
         call    getToke                ; get a token from pad
         mov     si,offset token        ; point to token
         call    asc2dec                ; convert token to a flt number
         mov     bx,nxtAdr              ; set next number address
         mov     nxtNum,bx
         pop     bp                     ; restore scan addr
         or      di,di                  ; text bad number flag
         js      xq155a                 ; go if not a float
         call    dec2flt                ; put float on 8087 stack
         jmp     pushFlt                ; put it on flt stack - done w/ fnumber
xq155a:  mov     di,offset fltWrn       ; warning - not a float
         jmp     error                  ; go print message

xeq156:  call    popInt                 ; string - get TOIS
         call    int2pad                ; call integer to pad
         jc      xq156a                 ; go if pad length exceeded
         ret                            ; done with string
xq156a:  jmp     xq22d                  ; go print pad size exceeded error

xeq157:  call    popFlt                 ; fstring - get TOFS
         call    flt2pad                ; call float to pad
         jc      xq156a                 ; go if pad length exceeded
         ret                            ; done with fstring

xeq158:  push    bp                     ; circle - save scan address
         mov     bp,offset circle       ; point to draw circle routine
         call    parse                  ; do it
         pop     bp                     ; get scan address back
         ret                            ; done with circle

xeq159:  call    popInt                 ; pad store - get pad stk addr
         cmp     ax,1                   ; test range
         jl      xq159a                 ; not ok if <0
         cmp     ax,3
         jg      xq159a                 ; not ok if >3
         dec     ax                     ; zero base it
         mov     bx,80
         mul     bx                     ; start addr
         mov     si,ax
         add     si,offset padStack     ; point to destination
         mov     di,offset pad          ; point to source
         jmp     movStr                 ; store pad
xq159a:  mov     di,offset bpdsErr      ; here if pad stack range error
         jmp     error                  ; go print message
                          
xeq160:  call    popInt                 ; pad fetch - get pad stk addr
         cmp     ax,1                   ; test range
         jl      xq159a                 ; not ok if <0
         cmp     ax,3
         jg      xq159a                 ; not ok if >3
         dec     ax                     ; zero base it
         mov     bx,80
         mul     bx                     ; start addr
         mov     di,ax
         add     di,offset padStack     ; point to source 
         mov     si,offset pad          ; point to destination
         call    movStr
         sub     si,di                  ; compute pad length
         mov     ax,si
         dec     ax                     ; don't count null
         mov     lpad,al                ; store pad length
         ret                            ; done with pad fetch

xeq161:  mov     ax,PrgEnd              ; free - get program end address
         inc     ax                     ; bump past null
         jmp     pushInt                ; put on TOIS - done with free

xeq162:  mov     ax,offset pad          ; pad - get pad address
         jmp     pushInt                ; put on TOIS - done with pad
 
xeq163:  call    popInt                 ; c! - get TOIS = addr
         push    ax                     ; save it
         call    popInt                 ; get char
         pop     bx                     ; addr to bx
         mov     byte ptr [bx],al       ; store char
         ret                            ; done with c!

xeq164:  call    popInt                 ; c@ - get TOIS - addr
         mov     bx,ax                  ; put in bx
         mov     al,byte ptr [bx]       ; get char
         xor     ah,ah                  ; make it a word
         jmp     pushInt                ; put on TOIS - done with c@
        
xeq165:  call    getToke                ; tic - get next word
         jc      xq165a                 ; go of no char
         mov     al,byte ptr [di]       ; get char
         xor     ah,ah                  ; make it a word
         jmp     pushInt                ; put on TOIS - done with tic
xq165a:  mov     di,offset ticErr       ; here if no char - point to error msg
         jmp     error                  ; go process error

xeq166:  call    xeq018                 ; drop2 - drop one
         jmp     xeq018                 ; and drop again - done with drop2

xeq167:  call    xeq019                 ; fdrop2 - fdrop one
         jmp     xeq019                 ; and fdrop again - done with fdrop2

xeq168:  call    popInt                 ; +spaces - get count from TOIS
         mov     bx,ax                  ; put in bx
         add     bl,lpad                ; add current length of pad
         push    bx                     ; save it
         cmp     bl,80                  ; check for pad exceeded
         ja      xq168b                 ; go if so
         mov     cx,ax                  ; put count in cx
         mov     bl,lpad                ; get pad length
         xor     bh,bh                  ; make it a word
         add     bx,offset pad
xq168a:  mov     byte ptr [bx],SPC      ; store spaces in pad
         inc     bx
         loop    xq168a
         mov     byte ptr[bx],NULL      ; store null terminator
         pop     bx                     ; new length
         mov     lpad,bl                ; store it
         ret                            ; done with +spaces
xq168b:  mov     di,offset qtpErr       ; here if pad exceeded
         jmp     error                  ; go process error

xeq169:  call    popInt                 ; slice - get 2nd loc
         mov     bx,ax                  ; save 2nd in bx
         call    popInt                 ; get 1st loc
         mov     dl,lpad                ; set dx to current pad len
         xor     dh,dh
         cmp     ax,1                   ; range checks
         jl      xq169a                 ; bad if 1st<1
         cmp     ax,dx
         jg      xq169b                 ; bad if 1st>len
xq169e:  cmp     bx,1
         jl      xq169c                 ; bad if 2nd<1
         cmp     bx,dx
         jg      xq169d                 ; bad if 2nd>len
xq169f:  mov     cx,bx                  ; make count
         sub     cx,ax
         inc     cx
         add     ax,offset pad          ; create source address
         mov     si,ax
         dec     si                     ; zero base it
         mov     di,offset pad          ; get dest address       
         mov     bx,cx                  ; save count
xq169g:  mov     al,byte ptr [si]       ; move slice
         mov     byte ptr [di],al
         inc     si
         inc     di
         loop    xq169g
         mov     byte ptr [di],NULL     ; store null terminator
         mov     lpad,bl                ; store new len
         ret                            ; done          
xq169a:  mov     ax,1                   ; set 1st to 1 if <1
         jmp     short xq169e
xq169b:  mov     ax,dx                  ; set 1st to len if >len
         jmp     short xq169e
xq169c:  mov     bx,1                   ; set 2nd to 1 if <1
         jmp     short xq169f
xq169d:  mov     bx,dx                  ; set 2nd to len if >len
         jmp     short xq169f

xeq170:  call    popInt                 ; concat - get TOIS
         cmp     ax,1                   ; check range
         jl      xq170b                 ; bad if <1
         cmp     ax,3
         jg      xq170b                 ; bad if >3
         dec     ax                     ; zero base it
         mov     bx,80                  ; x 80 chars size
         mul     bx                     ; start addr
         mov     si,ax
         add     si,offset padStack     ; point to source 
         mov     al,lpad                ; get length of pad
         xor     ah,ah                  ; make it a word
         mov     di,ax                  ; end addr
         add     di,offset pad          ; point to dest
xq170a:  cmp     ax,80                  ; at 80 chars yet?
         je      xq170c                 ; go if so
         mov     bl,byte ptr [si]       ; move string
         mov     byte ptr [di],bl
         or      bl,bl
         je      xq170c                 ; done if null
         inc     si
         inc     di
         inc     ax                     ; count
         jmp     short xq170a
xq170c:  mov     lpad,al                ; set new pad length
         ret                            ; done
xq170b:  jmp     xq159a                 ; here if range error - go prn msg

xeq171:  mov     di,offset pad          ; upper - point to pad
         jmp     upper                  ; go convert to uc

xeq172:  mov     di,offset pad          ; lower - point to pad
         jmp     lower                  ; go covert to lc

xeq173:  mov     ax,BLACK               ; black - 0
         jmp     pushInt                ; done with black

xeq174:  mov     ax,DARKBLUE            ; darkblue - 1
         jmp     pushInt                ; done with darkblue

xeq175:  mov     ax,DARKGREN            ; darkgreen - 2
         jmp     pushInt                ; done with darkgreen

xeq176:  mov     ax,DARKCYAN            ; darkcyan - 3
         jmp     pushInt                ; done with darkcyan

xeq177:  mov     ax,DARKRED             ; darkred - 4
         jmp     pushInt                ; done with darkred

xeq178:  mov     ax,DARKMAGE            ; darkmagenta - 5
         jmp     pushInt                ; done with darkmagenta

xeq179:  mov     ax,BROWN               ; brown - 6
         jmp     pushInt                ; done with brown

xeq180:  mov     ax,GREY                ; grey - 7
         jmp     pushInt                ; done with grey

xeq181:  mov     ax,DARKGREY            ; darkgrey - 8
         jmp     pushInt                ; done with darkgrey

xeq182:  mov     ax,BLUE                ; blue - 9
         jmp     pushInt                ; done with blue

xeq183:  mov     ax,GREEN               ; green - 10
         jmp     pushInt                ; done with green

xeq184:  mov     ax,CYAN                ; cyan - 11
         jmp     pushInt                ; done with cyan

xeq185:  mov     ax,RED                 ; red - 12
         jmp     pushInt                ; done with red

xeq186:  mov     ax,MAGENTA             ; magenta - 13
         jmp     pushInt                ; done with magenta

xeq187:  mov     ax,YELLOW              ; yellow - 14
         jmp     pushInt                ; done with yellow

xeq188:  mov     ax,WHITE               ; white - 15
         jmp     pushInt                ; done with white

xeq189:  mov     ah,2Ah                 ; date - call DOS
         int     DOS
         mov     ax,cx                  ; get year
         call    pushInt                ; put on TOIS
         mov     ax,dx                  ; get month,day
         xor     ah,ah                  ; day
         call    pushInt                ; put on TOIS
         mov     ax,dx
         mov     al,ah                  ; month
         xor     ah,ah
         jmp     pushInt                ; put on TOIS - done with date

xeq190:  mov     ah,2Ch                 ; time - call DOS
         int     DOS
         mov     ax,dx                  ; get secs + hundreds
         mov     al,ah                  ; seconds
         xor     ah,ah
         call    pushInt                ; put on TOIS
         mov     ax,cx                  ; get hours + mins
         xor     ah,ah                  ; minutes
         call    pushInt                ; put on TOIS
         mov     ax,cx
         mov     al,ah                  ; seconds
         xor     ah,ah
         jmp     pushInt                ; put on TOIS - done with time

; this is a very simple random function
xeq191:  mov     ax,ranNum              ; random - get seed
         mov     bx,13                  ; multiply by 13
         mul     bx
         xchg    ah,al                  ; swap bytes
         mov     ranNum,ax              ; put seed back
         and     ax,7FFFh               ; make it plus
         jmp     pushInt

seed:    mov     ah,2Ch                 ; seed - call DOS time function
         int     DOS
         mov     ranNum,dx              ; seed = hundreds secs
         ret                            ; done with seed

xeq192:  call    popFlt                 ; aspect - get TOFS
         FABS                           ; make it plus
         FLD1                           ; make a 3
         FLD     ST(0)
         FLD     ST(0)
         FADD
         FADD
         call    FCMP                   ; test
         jne     xq192a                 ; go if aspect >= 3        
         FSTP    aspect                 ; store it
         ret                            ; done 
xq192a:  FSTP   ST(0)                   ; here if aspect >= 3, ignore
         ret                            ; done with aspect 

xeq193:  mov     bl,iStkX               ; => - get int stack index (0 - 9)
         or      bl,bl                  ; test for 0
         je      xq193a                 ; stack empty if so
         dec     bl                     ; drop index
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; times 2
         mov     ax,intStack[bx]        ; get stack element into ax reg
         jmp     xq125a                 ; go push - done with =>
xq193a:  jmp     iStkEmt                ; here if int stk empty - go print error 
 
xeq194:  mov     bl,pStkX               ; <= - get push stack index (0 - 2)
         or      bl,bl                  ; test for 0
         je      xq194a                 ; stack empty if so
         dec     bl                     ; drop index
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; times 2
         mov     ax,pStack[bx]          ; get stack element into ax reg
         jmp     pushInt                ; go push - done with <=
xq194a:  jmp     pstkEmt                ; here if pstk empty - go print error

xeq195:  mov     bl,fStkX               ; f=> - get int stack index (0 - 9)
         or      bl,bl                  ; test for 0
         je      xq195a                 ; stack empty if so
         dec     bl                     ; drop index
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fltStack[bx]           ; get stack element into ST(0) reg
         jmp     xq127a                 ; go fpush - done with f=>
xq195a:  jmp     fStkEmt                ; here is flt stk empty - go print error

xeq196:  mov     bl,fpStkX              ; f<= - get fp stack index (0 - 2)
         or      bl,bl                  ; test for 0
         je      xq196a                 ; stack empty if so
         dec     bl                     ; drop index
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fpStack[bx]            ; get stack element into ST(0) reg
         jmp     pushFlt                ; go push - done with f<=
xq196a:  jmp     fpStkEm                ; here if fpstk empty - go print error

xeq197:  call    popInt                 ; != - get TOIS
         mov     bx,ax                  ; save it
         call    popInt                 ; get next stack element
         cmp     bx,ax                  ; compare them
         je      xq197a                 ; go set FALSE
         jmp     xq44b                  ; go set TRUE
xq197a:  jmp     xq44a                  ; done with !=

xeq198:  call    popFlt                 ; f!= - get TOFS
         call    popFlt                 ; get next flt stack element
         FCOMPP                         ; compare next - top
         FSTSW   npustat                ; store stats
         FWAIT
         and     byte ptr npustat+1,40h ; ST(0) zero?
         jne     xq197a                 ; go if so - force false
         jmp     xq44b                  ; else force true

xq199a:  jmp     xq159a                 ; go print pad range error   
xeq199:  call    popInt                 ; pad= - get TOIS
         cmp     ax,1                   ; test range
         jl      xq199a                 ; not ok if <0
         cmp     ax,3
         jg      xq199a                 ; not ok if >3
         dec     ax                     ; zero base it
         mov     bx,80
         mul     bx                     ; start addr
         mov     si,ax
         add     si,offset padStack     ; point to destination
         mov     di,offset pad          ; point to source
xq199b:  mov     al,byte ptr [di]       ; get char from pad
         or      al,al                  ; test it
         je      xq199c                 ; go if null
         cmp     al,byte ptr [si]       ; else compare to pad stack char
         jne     xq199d                 ; strings not equal if compare fails
         inc     si                     ; bump pointers
         inc     di
         jmp     short xq199b           ; keep comparing
xq199c:  mov     al, byte ptr [si]      ; get char in pad stack
         or      al,al                  ; test it
         jne     xq199d                 ; if pad stk char not null, strings ne          
         mov     ax,TRUE                ; force true
         jmp     pushInt                ; put on TOIS
xq199d:  mov     ax,FALSE               ; force false
         jmp     pushInt                ; put on TOIS - done with pad=

xeq200:  call    popInt                 ; search - get search char
         or      al,al                  ; test char
         je      xq200c                 ; do nothing if null
         mov     ah,1                   ; counter init
         mov     si,offset pad          ; point to pad
xq200a:  mov     bl,byte ptr [si]       ; get pad char
         or      bl,bl                  ; test it
         je      xq200c                 ; go if null
         cmp     al,bl                  ; search char = pad char?
         je      xq200b                 ; go if so
         inc     ah                     ; bump counter
         inc     si                     ; bump pointer
         jmp     short xq200a           ; keep searching
xq200c:  xor     ah,ah
xq200b:  mov     al,ah                  ; make counter a word
         xor     ah,ah
         jmp     pushInt                ; put on TOIS - done with search
                
xq201a:  mov     di,offset creErr       ; here if file create error
         jmp     error                  ; go print error message 
xeq201:  call    getToke                ; create-file - get filename
         cmp     byte ptr [di],NULL     ; no name?
         je      xq201a                 ; go if so
         cmp     cx,12                  ; check number of chars
         ja      xq201a                 ; go if more than 12
         mov     si,offset iofname      ; point to io file name 
         call    movStr                 ; move token name to iofname
         mov     dx,offset iofname      ; set up for file create 
         mov     cx,0                   ; normal file
         mov     ax,3C00h               ; go create
         int     DOS
         jc      xq201a                 ; go if file create error
         jmp     pushInt                ; put handle on TOIS- done with create
         
xq202a:  jmp     openFal                ; here if open error 
xeq202:  call    getToke                ; open-file - get filename
         cmp     byte ptr [di],NULL     ; no name?
         je      xq202a                 ; go if so
         cmp     cx,12                  ; check number of chars
         ja      xq202a                 ; go if more than 12
         mov     si,offset iofname      ; point to io file name 
         call    movStr                 ; move token name to iofname
         call    popInt                 ; get access code 0=r,1=w,2=a
         cmp     ax,0                   ; check range
         jl      xq202b
         cmp     ax,2
         ja      xq202b
         mov     dx,offset iofname      ; set up for file create 
         mov     ah,3Dh                 ; go open it
         int     DOS
         jc      xq202a                 ; go if file create error
         jmp     pushInt                ; put handle on TOIS - done with open
xq202b:  mov     di,offset invmdEr      ; invalid mode error
         jmp     error                  ; print error message

xeq203:  call    popInt                 ; close-file - get file handle
         mov     bx,ax                  ; set up close call
         mov     ax,3E00h               ; close
         int     DOS
         jc      xq203a                 ; go if error
         ret                            ; done with close
xq203a:  mov     di,offset closErr      ; get close error message pointer
         jmp     error                  ; go print it

xq204b:  mov     di,offset wrtErr       ; here if write error
         jmp     error                  ; go print error message
xeq204:  call    popInt                 ; write-file - get file handle
         mov     iohand,ax              ; put io file handle in iohand
         mov     dx,offset pad          ; point dx to pad
         mov     di,dx                  ; copy pointer to di
xq204a:  mov     al, byte ptr [di]      ; get char
         or      al,al                  ; null terminator?
         je      xq204c                 ; go if so
         push    di                     ; save pointer
         call    writ1                  ; write char
         pop     di                     ; pointer back
         inc     di                     ; bump it
         mov     dx,di                  ; put in dx, needed for dos call
         jmp     xq204a                 ; loop back
xq204c:  mov     dx,offset prepad+1     ; point to CR,LF
         call    writ1                  ; write CR
         inc     dx                     
writ1:   mov     cx,1                   ; write 1 byte
         mov     bx,iohand              ; set bx to io file handle
         mov     ax,4000h               ; write it
         int     DOS
         cmp     ax,1                   ; confirm write
         jne     xq204b                 ; go if error
         ret
         
xeq205:  call    popInt                 ; read-file - get file handle          
         mov     iohand,ax              ; put io file handle in iohand
         mov     dx,offset pad          ; point dx to pad
xq205a:  push    dx                     ; save pointer
         mov     bx,iohand              ; set up dos call
         mov     cx,1                   ; read 1 char
         mov     ax,3F00h               ; go read it
         int     DOS
         jc      xq205c                 ; go if read error
         pop     di                     ; get pointer back to di
         cmp     byte ptr [di],CR       ; is it a CR?
         je      xq205b                 ; go if so
         inc     di                     ; bump pointer
         mov     dx,di                  ; put pointer in dx for dos call
         cmp     di,offset pad+80       ; at end of pad?
         jne     xq205a                 ; go if not so
xq205b:  mov     byte ptr [di],NULL     ; store null terminator
         sub     di,offset pad          ; compute pad length
         mov     ax,di
         mov     lpad,al                ; set pad length parm
         mov     dx,offset temp         ; point to a dummy area
         mov     bx,iohand              ; set up dos call
         mov     cx,1                   ; to read the LF                   
         mov     ax,3F00h               ; read it
         int     DOS
         jc      xq205c
         ret                            ; done with read-file
xq205c:  mov     di,offset rdfErr       ; here if read error 
         jmp     error                  ; go print error message    
                                 
xeq206:  call    popInt                 ; tab - get TOIS
         and     al,1Fh                 ; limit to 0 - 15
         mov     tabWid,al              ; set tab width
         ret 
        
xeq207:  test    gflag,1                ; locate - check for graphics mode
         jne     xq207a                 ; go if graphics mode
         call    getCur                 ; text mode - get cursor coords
         mov     al,dl                  ; set row
         xor     ah,ah
         call    pushInt                ; push on int stack
         mov     al,dh                  ; set column
         xor     ah,ah
         jmp     pushInt                ; push on int stack - done  
xq207a:  mov     ax,X1                  ; graphics mode - get x pixel location
         call    pushInt                ; put on int stack
         mov     ax,Y1                  ; get y
         jmp     pushInt                ; push on int stack - done with locate

xeq208:  mov     ax,0                   ; read = 0
         jmp     pushInt

xeq209:  mov     ax,1                   ; write = 1
         jmp     pushInt

xeq210:  mov     ax,2                   ; append = 2
         jmp     pushInt

xeq211:  mov     trflag,1               ; trace-on - set flag
         ret                            ; done with trace-on

xeq212:  mov     trflag,0               ; trace-off - reset flag
         ret                            ; done with trace-off
 
xeq213:  call    popFlt                 ; e. - get float   
         jmp     prnExp                 ; print float in exponential form
                                        ; done with e. 
xeq214:  call    popInt                 ; eplaces - get TOIS
         or      ax,ax                  ; take abs
         jns     xq214a
         neg     ax
xq214a:  mov     bx,10                  ; divisor = 10
         xor     dx,dx
         div     bx                     ; |fmt|/10 = exponent digs
         cmp     dl,1                   ; must have 1 <= digs <= 4
         jae     xq214b
         mov     dl,1
         jmp     short xq214c
xq214b:  cmp     dl,4
         jle     xq214c
         mov     dl,4
xq214c:  mov     expdig,dl              ; store in # exponent digits
         xor     dx,dx
         mov     bx,100                 ; divisor = 100
         div     bx                     ; |fmt|/1000 = mantissa digs
         cmp     dl,1                   ; must have 1 <= digs <= 17
         jae     xq214d
         mov     dl,1
         jmp     short xq214e
xq214d:  cmp     dl,17
         jle     xq214e
         mov     dl,17
xq214e:  mov     expdec,dl              ; store in #  mantissa decimnals
         ret                            ; done with eplaces       

xeq215:  call    popFlt                 ; estring - get TOFS
         call    exp2pad                ; call float expo form to pad
         jc      xq215a                 ; go if pad length exceeded
         ret                            ; done with e.
xq215a:  jmp     xq22d                  ; go print pad exceeded error

xeq216:  mov     bl,pStkX               ; .p - get push stack index
         or      bl,bl                  ; test for empty
         je      xq216b                 ; go if so
         xor     bl,bl                  ; set count to 0
xq216a:  push    bx                     ; save count
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; times 2
         mov     ax,pStack[bx]          ; get integer from stack
         call    prnInt                 ; print it
         call    prnSPC                 ; print space
         pop     bx                     ; count back
         inc     bl                     ; increment it
         cmp     bl,pStkX               ; test for all elements printed
         js      xq216a                 ; go if not
         mov     bl,fpStkX              ; get fpush stack index
         or      bl,bl                  ; test for empty
         je      xq216c                 ; go if so
xq216e:  xor     bl,bl                  ; set count to 0
xq216d:  push    bx                     ; save count
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fpStack[bx]            ; get float from stack
         call    prnFlt                 ; print it
         call    prnSPC                 ; print sapce
         pop     bx                     ; count back
         inc     bl                     ; increment it
         cmp     bl,fpStkX              ; test for all elements printed
         js      xq216d                 ; go if not
xq216c:  jmp     prnCRLF                ; print newline - done
xq216b:  mov     bl,fpStkX              ; get fpush stack index
         or      bl,bl                  ; test for empty
         jne     xq216e                 ; go display fpstack if not
         mov     di,offset pstMsg       ; point to pstack empty message
         call    prnStr                 ; print it
         jmp     prnCRLF                ; newline - done with .p

xq217a:  jmp     xq139a                 ; go print graphics mode error 
xeq217:  call    popInt                 ; g. - get char
         test    gflag,1                ; test for graphics mode
         je      xq217a                 ; go do graphics error if not
         jmp     gprnChr                ; else go print graphics char

xeq218:  call    popInt                 ; l. - get top of integer stack
         or      ax,ax                  ; test it
         je      xq218a                 ; go if false
         mov     di,offset truMsg       ; get 'True' message pointer
         jmp     prnStr                 ; go print it - done
xq218a:  mov     di,offset flsMsg       ; here if false, get pointer
         jmp     prnStr                 ; go print it - done with l.
         

xeq219:  mov     ax,offset gfont        ; font - get font table address
         jmp     pushInt                ; put on TOIS - done with font

xeq220:  mov     xorGrfx,80h            ; xorgr-on - set xor graphics bit on
         ret                            ; done with xorgr-on

xeq221:  mov     xorGrfx,0              ; xorgr-off - set xor graphics bit off
         ret

xeq222:  mov     ax, offset lbuf        ; system - get address
         jmp     pushInt                ; put on TOIS
 
xeq223:  call    popInt                 ; w@ - get TOIS - addr
         mov     bx,ax                  ; put in bx
         mov     ax,word ptr [bx]       ; get word
         jmp     pushInt                ; put on TOIS - done with w@

xeq224:  call    popInt                 ; w! - get TOIS = addr
         push    ax                     ; save it
         call    popInt                 ; get word
         pop     bx                     ; addr to bx
         mov     word ptr [bx],ax       ; store word
         ret                            ; done with w!

xeq225:  call    cold3                  ; reparse - cold start
         mov     bp,offset program      ; set parse address
         ret				; done with reparse

xeq226:  call    popInt                 ; >> - get shift count from TOIS
         mov     cl,al                  ; save count
         call    popInt                 ; get integer
         shr     ax,cl                  ; shift right logical
         jmp     pushInt                ; put integer back - done with shift right

xeq227:  call    popInt                 ; << - get shift count from TOIS
         mov     cl,al                  ; save count
         call    popInt                 ; get integer
         shl     ax,cl                  ; shift left logical
         jmp     pushInt                ; put integer back - done with shift left

xeq228:  call    getToke                ; error-on - get next token: error tag
         cmp     byte ptr [di],NULL     ; no tag?
         je      xq228a                 ; error if so
         call    toke8                  ; compact to 8 chars
         call    upper                  ; make it uppercase
         mov     si,offset token        ; move token to error jump tag
         mov     di,offset errjmp
         mov     ah,8
xq228b:  mov     al,[si]
         mov     [di],al
         inc     si
         inc     di
         dec     ah
         jne     xq228b
         dec     ah                     ; set error-on flag
         mov     errflg,ah
         ret                            ; done with error-on
xq228a:  jmp     xq37a                  ; bad label error msg

xeq229:  mov     errflg,0               ; error-off - clear flag
         ret                            ; done with error-off      

xeq230:  mov     di,offset program      ; clear-mem (undocumented)
         mov     prgEnd,di
         mov     byte ptr [di],NULL
         xor     al,al
         mov     lblX,al
         mov     wrdX,al
         jmp     cold4

xeq231:  mov     ax,prgEND	        ; free
         inc     ax
         jmp     pushINT

;-----------------------------------------------------------------------------
; gprnChr
; graphics print char
; entry: al = ascii char
;        gcurx holds x position of upper left corner char block
;        gcury holds y position of upper left corner char block
; exit:  char plotted
;        updates cursor x by 10
;        handles BS, CR, LF, and pause at bottom of screen + screen clear
; uses:  pltChr, BIOS plot dot
;-----------------------------------------------------------------------------

gprnChr: or   al,al                    ; ignore NULL
         je   gprchex
         cmp  al,LF                    ; line feed?
         jne  gprch40                  ; go if not
         mov  gcurx,0                  ; set to pixel 0
         ret                           ; done
gprch40: push ax                       ; save reg
         cmp  al,CR                    ; carriage return?
         je   gprch01                  ; go set to next line
         mov  cx,gcurx                 ; get x position
         mov  dx,gcury                 ; get y position 
         cmp  al,BS                    ; back space?
         jne  gprch50                  ; go if not
         cmp  cx,10                    ; don't back up if < 10 pixels
         jl   gprch55
         sub  cx,10                    ; back up to previous char
         call clrChr                   ; blank it
         sub  gcurx,10                 ; update cursor 
         jmp  gprch55         
gprch50: call pltChr                   ; plot char
         add  gcurx,10                 ; update x position
         cmp  gcurx,630                ; newline if at eol
         jle  gprch55                           ; done 
gprch01: mov  gcurx,0                  ; set x position to 0
         mov  ax,gcury                 ; get y
         add  ax,10                    ; bump down screen ten pixels
         cmp  ax,340                   ; at bottom?
         jle  gprch02                  ; go if not
         call inpChr
         call gcls                     ; back to top if so
         xor  ax,ax
gprch02: mov  gcury,ax                 ; update y position
gprch55: pop  ax                       ; restore reg
gprchex: ret                           ; done with gprnChr

;-----------------------------------------------------------------------------
; clear char
; entry: cx,dx = x,y coordinates of upper left of char to clear
;-----------------------------------------------------------------------------

clrChr:  push    ax                     ; save regs
         push    bx
         push    di
         mov     di,8                   ; 8 pixels high
clrch01: push    di                     ; save it
         mov     di,8                   ; 8 pixels wide
clrch02: mov     al,0                   ; no color
         mov     bh,al                   ; page 0
         mov     ah,0Ch                 ; write dot call
         INT     BIOS                   ; clear pixel
         inc     cx                     ; next x pixel
         dec     di                     ; decrement and loop
         jne     clrch02
         inc     dx                     ; next y 
         sub     cx,8                   ; set x back to start
         pop     di                     ; bump y and loop
         dec     di
         jne     clrch01
         pop     di                     ; restore regs
         pop     bx
         pop     ax
         ret                            ; done with clear char
         
;-----------------------------------------------------------------------------
; pltChr
; plot graphics char
; entry: cx,dx = x,y coordinates of upper left of text char
;        al = ascii char
;        forClr is pixel color
; uses:  font table gfont
;-----------------------------------------------------------------------------

pltChr:  cmp     al,255                 ; handle char 255 as a special case
         je      clrChr                 ; it's a clear graphics char
plch04:  push    ax                     ; save regs
         push    bx
         push    di
         push    si
         xor     bh,bh                  ; page 0
         xor     ah,ah                  ; make char byte a word
         sal     ax,1                   ; x 8
         sal     ax,1
         sal     ax,1
         mov     si,offset gfont        ; get font table address
         add     si,ax                  ; offset into table 
         mov     di,8                   ; 8 bytes per character
plch01:  push    di                     ; save bytes/pixel count
         mov     al,[si]                ; get row of pixels
         mov     di,8                   ; 8 bits per byte
plch02:  sal     al,1                   ; get left bit into cy
         jnc     plch03                 ; go if no plot
         push    ax                     ; save the graphics byte
         mov     al,forClr              ; plot in current fg color
         or      al,xorGrfx             ; set xor grafix bit
         mov     ah,0Ch                 ; write dot call parm
         int     BIOS                   ; plot the graphics pixel 
         pop     ax                     ; graphics byte back
plch03:  inc     cx                     ; next x position
         dec     di                     ; bump down pixel counter
         jne     plch02                 ; go process next pixel this row
         sub     cx,8                   ; back up x
         inc     dx                     ; next y (row)
         inc     si                     ; next font byte
         pop     di                     ; row counter back
         dec     di                     ; bump it down
         jne     plch01                 ; loop back for next row of pixels
plch05:  pop     si                     ; restore regs
         pop     di
         pop     bx
         pop     ax
         ret                            ; done with plot char       
                          
;-----------------------------------------------------------------------------
; search dictionary
; compares token word to each word of root dictionary
; exit: returns bl as an index on match or -1 if no match
; dictionary is simply a string of FLIP words separated by nulls
; destroys bh, si, di
;-----------------------------------------------------------------------------

srchDct: mov     bl,1                   ; default to match on first dct word
         mov     di,offset Dict         ; point to dictionary
srd01:   mov     si,offset token        ; point to token
         call    cmpStr                 ; match?
         je      srd05                  ; done if so
srd02:   cmp     byte ptr [di],NULL     ; if not, find end of this dct word
         je      srd03                  ; go when null found
         inc     di                     ; bump dct pointer  
         jmp     short srd02            ; keep looking
srd03:   inc     di                     ; found end, bump to start of next
         cmp     bl,nDct                ; test for end of dictionary
         je      srd04                  ; go if so
         inc     bl                     ; else bump up counter
         jmp     short srd01            ; go try another match
srd04:   mov     bl,-1                  ; here if no match, set bl false
srd05:   ret                            ; else return with bl = index

;------------------------------------------------------------------------------
; search for label
; entry: di points to match token null terminated
; exit: si = address of word just following label if found
;            or -1 if not found
; searches pad, buffer & program area
;------------------------------------------------------------------------------

srchLbl: mov     si,offset subr         ; point si to presub
srlb15:  cmp     byte ptr [si],TILDE    ; is it a label
         je      srlb01                 ; go if yes
         cmp     byte ptr [si],NULL     ; terminator?
         jne     srlb17                 ; no match if so
         cmp     si,offset program      ; search program area yet?
         jns     srlb03                 ; go if so - not found
srlb17:  inc     si                     ; bump pointer
         jmp     short srlb15           ; keep looking
srlb01:  inc     si                     ; found ~, bump to label char
         mov     al,[si]                ; get label char
         cmp     byte ptr [si],NULL     ; last char?
         je      srlb05                 ; go if so
         call    isWht                  ; or end by white space?
         je      srlb05                 ; go if so
         cmp     al,'a'                 ; test against 'a'
         jl      srlb02                 ; not l.c.                  
         cmp     al,'z'                 ; test against 'z'
         ja      srlb02                 ; not l.c.
         sub     al,32                  ; it's l.c., change to u.c.             
srlb02:  push    ax                     ; save label char
         mov     al,[di]                ; get token char
         cmp     al,NULL                ; last char?
         je      srlb04                 ; go if so
         call    isWht                  ; or end by white space?
         je      srlb04                 ; go if so
         pop     bx                     ; label char in bl                  
         cmp     bl,al                  ; compare to token char
         jne     srlb07                 ; go if mismatch
         inc     di                     ; bump token pointer
         jmp     short srlb01           ; keep checking  
srlb04:  pop     bx       
srlb03:  mov     si,-1
         ret
srlb05:  mov     al,[di]
         cmp     al,NULL
         je      srlb06
         call    isWht
         jne     srlb02
srlb06:  ret     
srlb07:  inc     si                     ; here if this ~ doesn't match, bump si
         mov     di,offset token        ; reset pointer to token
         jmp     short srlb15           ; go look for next ~  
             
;------------------------------------------------------------------------------
; compare strings
; entry: di and si point to strings
; exit:  zero flag set if strings equal, clear otherwise
; uses bh
;------------------------------------------------------------------------------

cmpStr:  mov     bh,[si]                ; get str1 char 
         or      bh,bh                  ; end of str1?
         je      cps01                  ; go if so
         cmp     bh,[di]                ; match str2 char?
         jne     cps02                  ; go if not, no word match
         cmp     byte ptr[di],NULL      ; end of str2?
         je      cps03                  ; go if so
         inc     si                     ; match so far, point to next chars
         inc     di
         jmp     short cmpStr           ;  and keep trying to match
cps01:   cmp     byte ptr [di],NULL     ; exact match?
cps02:   ret                            ; done, if z flag set = match
cps03:   or      bh,bh                  ; clear zero flag
         ret                            ; done with compare strings 

;-----------------------------------------------------------------------------
; move string
; entry: di points to start of string to be moved
;        si points to start of move to location
; string must be null terminated
; caller needs to make sure there is enough room for the string
; exit:  di points to new string start 
;-----------------------------------------------------------------------------

movStr: push     si                     ; save destination
mvs01:  mov      al,[di]                ; get char of string
        mov      [si],al                ; move it
        inc      di                     ; bump pointers
        inc      si
        or       al,al                  ; test char
        jne      mvs01                  ; loop for more if not terminator
        pop      di                     ; get start address relocated string
        ret                             ; done with move string
        
;-----------------------------------------------------------------------------
; search string for char
; entry:
;   al = char to search for
;   si points to string start, null terminated
; exit:
;   dx = count of chars not including search char if found (or terminator)
;   si still points to string start
;   cy flag set if null terminator found before char is found 
;-----------------------------------------------------------------------------

fndChr:  push    si                     ; save regs
         push    cx              
         xor     dx,dx                  ; initialize count
         dec     dx                     ; don't include search char
fnc01:   mov     cl,[si]                ; get next char
         or      cl,cl                  ; test it
         je      fnc03                  ; go if null (terminator)
         inc     si                     ; bump up char position
         inc     dx                     ; bump up count
         cmp     cl,al                  ; is it the search char?
         jne     short fnc01            ; keep looking if not
fnc02:   pop     cx                     ; restore regs (cy clear)
         pop     si
         ret                            ; done  
fnc03:   stc                            ; set cy if term found
         jmp     short fnc02            ; done with find character
  
;-----------------------------------------------------------------------------
; pad token with trailing spaces, if needed, to make it 8 chars in length
; pare token to 8 chars if longer than 8 chars
; add a new null terminator
; entry: cx is count of characters
;        di points to token
; exit:  di still points to token
; destroys cx
;-----------------------------------------------------------------------------

toke8:   push    di                    ; save pointer
         push    cx                    ; save count
         cmp     cx,8                  ; compare count of chars to 8                  
         je      tok8c                 ; done if exactly 8
         ja      tok8b                 ; go if more than 8
         add     di,cx                 ; if less than 8, point to last char
         sub     cx,8                  ; find difference
         neg     cx
tok8a:   mov     byte ptr [di],' '     ; store a blank...
         inc     di                    ; point to next char
         dec     cx                    ;   ...difference times
         jne     tok8a                 ;   ...until 8-cx exhausted
         mov     byte ptr [di],NULL    ; store new null terminator
         jmp     short tok8c           ; done
tok8b:   mov     byte ptr [di+8],NULL  ; here if more than 8, store new null
tok8c:   pop     cx                    ; count back
         pop     di                    ; pointer back
         ret                           ; done with pad token

;-----------------------------------------------------------------------------
; Are characters of a string alpha-numeric? 
; String is first converted to all upper case
; first char must be A-Z only
; entry: di points to string
; exit:  carry set if no chars or characters not A - Z, 0 - 9, $ @ _
;        bx is count of chars
;        di points to string
; destroys al,si
;-----------------------------------------------------------------------------

is1alf:  call   upper                   ; convert to upper case
         cmp    byte ptr [di],NULL      ; null string?
         je     isa03                   ; exit with error
         xor    bx,bx                   ; initialize count
isa01:   mov    al,[di+bx]              ; get character
         inc    bx                      ; bump count     
         or     al,al                   ; is it null terminator
         je     isa04                   ; return with z and c flags set
         cmp    al,SPC                  ; is it a space
         je     isa04                   ;  ends string             
         cmp    bx,1                    ; first time thru only alpha
         je     isa02
         cmp    al,'$'                  ; $ ok
         je     isa01
         cmp    al,'@'                  ; @ ok
         je     isa01
         cmp    al,'_'                  ; _ ok
         je     isa01
         cmp    al,'.'                  ; dot ok
         je     isa01
         cmp    al,'0'                  ; 0 - 9 ok
         jb     isa03
         cmp    al,'9'
         jbe    isa01
isa02:   cmp    al,'A'                  ; A - Z ok
         jb     isa03
         cmp    al,'Z'
         jbe    isa01
isa03:   stc                            ; set carry, bad char found
         mov    ax,bx
         ret                            ; done                
isa04:   cmp    al,-1                   ; clear zero flag
         clc                            ; clear carry, all good chars
         ret                            ; done with is1alf


;------------------------------------------------------------------------------
; find word name
; entry: di points to token string with label standardized to 8 chars
; exit:  dx holds index into array if match found (0 based)
;        carry set if no match
;------------------------------------------------------------------------------

fndWrd:  mov     si,offset wrdnams      ; set token to name array
         mov     ch,wrdX                ; current last entry index in ch
         jmp     short fndTag           ; go to find tag


;------------------------------------------------------------------------------
; find fvariable name
; entry: di points to token string with label standardized to 8 chars
; exit:  dx holds index into array if match found (0 based)
;        carry set if no match
;------------------------------------------------------------------------------

fndVarn: mov     si,offset vnams        ; set token to name array
         mov     ch,varX                ; current last entry index in ch
         jmp     short fndTag           ; go to find tag

;------------------------------------------------------------------------------
; find variable name
; entry: di points to token string with label standardized to 8 chars
; exit:  dx holds index into array if match found (0 based)
;        carry set if no match
;------------------------------------------------------------------------------

fndfVrn: mov     si,offset fvnams       ; set token to name array
         mov     ch,fvarX               ; current last entry index in ch

;------------------------------------------------------------------------------         
; find tag
; entry: di points to token
;        si points to name array
;        ch is last array entry index 
; exit:  dx holds index into array if match found (0 based)
;        carry set if no match
;------------------------------------------------------------------------------
 
fndTag:  mov     temp,di                ; save token pointer
         mov     temp2,si               ; save tag array pointer
         or      ch,ch                  ; check for no tags
         je      ftg04                  ; go to failure if so
         xor     dx,dx                  ; set index into array to zero
ftg01:   mov     bx,dx                  ; put index into base reg
         sal     bx,1                   ; fast x 8
         sal     bx,1
         sal     bx,1
         mov     di,temp                ; restore pointers
         mov     si,temp2
         mov     cl,4                   ; compare 4 words each pass
ftg02:   mov     ax,[di]                ; get a token pair
         cmp     ax,[bx+si]             ; does token pair match tag element?
         jne     ftg03                  ; no for this index value
         add     si,2                   ; bump pointers 1 word
         add     di,2
         dec     cl                     ; dec pass count
         jne     ftg02                  ; loop back for more compares
         clc                            ; found a match, clear carry
         jmp     short ftg05            ; done                           
ftg03:   inc     dl                     ; no match this index, bump to next
         cmp     dl,ch                  ; searched all tags?
         jne     ftg01                  ; no, keep trying 
ftg04:   stc                            ; here if no match found - set cy
ftg05:   mov     di,temp                ; restore token pointer
         ret  

;------------------------------------------------------------------------------
; find fconstant name
; entry: di points to token string with label standardized to 8 chars
; exit:  dx holds index into array if match found (0 based)
;        carry set if no match
;------------------------------------------------------------------------------

fndfCon: mov     si,offset fconnams     ; set token to name array
         mov     ch,fConX               ; current last entry index in ch
         jmp     short fndTag           ; go to find tag

;------------------------------------------------------------------------------
; find constant name
; entry: di points to token string with label standardized to 8 chars
; exit:  dx holds index into array if match found (0 based)
;        carry set if no match
;------------------------------------------------------------------------------

fndCon:  mov     si,offset connams      ; set token to name array
         mov     ch,ConX                ; current last entry index in ch
         jmp     short fndTag           ; go to find tag

;-----------------------------------------------------------------------------
; is token a word name?
; entry: di points to token processed by toke8
; exit:  dx is array element if found
;        dx is -1 if not found (not a word name)
; if token is a word, push scan pos. (bp) on run stack then execute the word
; It is the function of ";" to pop the scan position  back off the stack and 
; resume execution
;-----------------------------------------------------------------------------

chkWrd:  mov     si,offset wrdnams      ; scan words
         mov     ch,wrdX                ; get word adr stack index
         call    fndTag                 ; go find the tag
         jc      chv01                  ; doesn't exist if cy - go error
         mov     bx,dx                  ; index value to bx reg
         sal     bx,1                   ; x 2
         mov     cx,wrdAdr[bx]          ; get the word execution address
         mov     bl,rStkX               ; current run stack index
         cmp     bl,rStkMax             ; check for overflow
         jae     cwd01                  ; go if overflow
         inc     rStkX                  ; bump up run stack index
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; x 2
         mov     rStack[bx],bp          ; store return point
         mov     bp,cx                  ; go execute
         ret                            ; done with check word
cwd01:   mov     di,offset rStkOEr      ; get run stack over flow message
         jmp     error                  ;  go print it

;-----------------------------------------------------------------------------
; is token an fvariable name?
; entry: di points to token processed by toke8
; exit:  dx is array element if found
;        dx is -1 if not found (not an fvariable name)
; if token is a fvariable name, put index value on integer stack
;-----------------------------------------------------------------------------

chkfVar: mov     si,offset fvnams       ; check float variable name table
         mov     ch,fvarX               ; get flt var name stack index
         jmp     short chvc             ; go process the tag

;-----------------------------------------------------------------------------
; is token a variable name?
; entry: di points to token processed by toke8
; exit:  dx is array element if found
;        dx is -1 if not found (not a variable name)
; if token is a variable name, put index value on integer stack
;-----------------------------------------------------------------------------
             
chkVar:  mov     si,offset vnams        ; check variable name table
         mov     ch,varX                ; get var name stack index
chvc:    call    fndTag                 ; go find the tag
         jnc     chv02                  ; tag exists if no carry - go 
chv01:   mov     dx,-1                  ; else set dx to failure code
         ret                            ; done
chv02:   inc     dx                     ; 1 base the index
         mov     ax,dx                  ; and push on TOIS
         jmp     pushInt                ; done with check variable

;-----------------------------------------------------------------------------
; is token a constant name?
; entry: di points to token processed by toke8
; exit:  dx is -1 if not found (not a constant name)
; if token is a constant name, put the value on the integer stack
;-----------------------------------------------------------------------------

chkCon:  mov     si,offset connams      ; check constant name table
         mov     ch,conX                ; get constant name stack index
         call    fndTag                 ; go find the tag
         jnc     chC02                  ; tag exists if no carry - go 
         mov     dx,-1                  ; else set dx to failure code
         ret                            ; done
chC02:   mov     bx,dx                  ; get the constant value
         sal     bx,1
         mov     ax,cons[bx]
         jmp     pushInt                ; put on TOIS - done with chkCon

;-----------------------------------------------------------------------------
; is token an fconstant name?
; entry: di points to token processed by toke8
; exit:  dx is -1 if not found (not a fconstant name)
; if token is an fconstant name, put the value on the float stack
;-----------------------------------------------------------------------------

chkFcon: mov     si,offset fconnams     ; check fconstant name table
         mov     ch,fConX               ; get fconstant name stack index
         call    fndTag                 ; go find the tag
         jnc     chfC02                 ; tag exists if no carry - go 
         mov     dx,-1                  ; else set dx to failure code
         ret                            ; done
chfC02:  mov     bx,dx                  ; get the constant value
         sal     bx,1                   ; fast x 10
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fcons[bx]
         jmp     pushFlt                ; put on TOIS - done with chkfCon

;-----------------------------------------------------------------------------
; push integer 
; places integer in ax on 10 level psuedo stack
; iStkX is index into stack
; error message generated on stack overflow
;-----------------------------------------------------------------------------

pushInt: push    bx                     ; save reg
         mov     bl,iStkX               ; get integer stack index (0 - 9)
         cmp     bl,10                  ; check for overflow, index >= 10
         jae     iStkOvr                ; go if stack overflow
         xor     bh,bh                  ; make index a word
         sal     bx,1                   ; times 2
         mov     intStack[bx],ax        ; put the integer on the stack
         inc     iStkX                  ; bump stack index
         pop     bx                     ; restore reg
         ret                            ; done with push integer
iStkOvr: mov     di,offset iStkOEr      ; here if int stk overflow - pt to msg
         jmp     error                  ; go print the error 

;------------------------------------------------------------------------------
; pop integer 
; pops value off 10 level integer stack and places it in the ax register
; iStkX is index into stack
; error message generated if stack empty
;------------------------------------------------------------------------------

popInt:  push    bx                     ; save reg
         mov     bl,iStkX               ; get int stack index (0 - 9)
         or      bl,bl                  ; test for 0
         je      iStkEmt                ; stack empty if so
         dec     bl                     ; drop index
         mov     iStkX,bl               ; save it back
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; times 2
         mov     ax,intStack[bx]        ; get stack element into ax reg
         pop     bx                     ; restore reg
         ret                            ; done with pop integer
iStkEmt: mov     di,offset iStkEEr      ; here if int stk empty - pt to err msg
         jmp     error                  ; go print the error 

;-----------------------------------------------------------------------------
; push float
; places float in ST(0) on 10 level psuedo stack
; fStkX is index into stack
; error message generated on stack overflow
;-----------------------------------------------------------------------------

pushFlt: push    bx                     ; save reg
         mov     bl,fStkX               ; get float stack index (0 - 9)
         cmp     bl,10                  ; check for overflow, index >= 10
         jae     fStkOvr                ; go if stack overflow
         xor     bh,bh                  ; make index a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FSTP    fltStack[bx]           ; put the integer on the stack
         FWAIT
         inc     fStkX                  ; bump stack index
         pop     bx                     ; restore reg
         ret                            ; done with push integer
fStkOvr: mov     di,offset fStkOEr      ; here if int stk overflow - pt to msg
         jmp     error                  ; go print the error description

;-----------------------------------------------------------------------------
; pop float 
; pulls value off 10 level float stack and places it in the ST(0) register
; fStkX is index into stack
; error message generated if stack empty
;-----------------------------------------------------------------------------

popFlt:  push    bx                     ; save reg
         mov     bl,fStkX               ; get int stack index (0 - 9)
         or      bl,bl                  ; test for 0
         je      fStkEmt                ; stack empty if so
         dec     bl                     ; drop index
         mov     fStkX,bl               ; save it back
         xor     bh,bh                  ; make it a word
         sal     bx,1                   ; fast multiply by ten
         mov     ax,bx
         sal     bx,1
         sal     bx,1
         add     bx,ax
         FLD     fltStack[bx]           ; get stack element into ST(0) reg
         pop     bx                     ; restore reg
         ret                            ; done with pop integer
fStkEmt: mov     di,offset fStkEEr      ; here if int stk empty - pt to err msg
         jmp     error                  ; go print the error description

;------------------------------------------------------------------------------
; open output.asc
; teed output file
; all prints go to this file when teed
; entry: none
; exit:  out file handle stored in prnHand
;------------------------------------------------------------------------------

openTee: push    dx                     ; save reg
         mov     dx,offset outFN        ; get output filename
         mov     ax,3D01h               ; open for write
         int     DOS
         jc      openFal                ; go if open error 
         mov     prnhand,ax             ; save output file handle
         pop     dx                     ; restore reg
         ret                            ; done with open tee

;------------------------------------------------------------------------------
; close output.asc
; closes teed output
; prints no longer go to output.asc
; entry: out file handle in prnHand
; exit:  none
;------------------------------------------------------------------------------

closTee: push    bx                     ; save reg
         mov     bx,prnhand             ; get out file handle
         jmp     closFl1                ; go close
 
       
;------------------------------------------------------------------------------
; open file
; entry: file name in fname
; exit:  prints error message and goes to main loop if an open error occurs
;        else file handle is put in fhand
; destroys ax
;------------------------------------------------------------------------------

openFil: push    dx                     ; save reg
	 mov     dx,offset fname        ; get filename
	 mov     ax,3D00h               ; open for read
	 int     DOS
	 jc      openFal                ; go if open error
	 mov     fhand,ax               ; save file handle
	 pop     dx                     ; restore reg
	 ret                            ; done with open file
openFal: mov     di,offset opnErr       ; here for open error
	 jmp     error

;------------------------------------------------------------------------------
; get file size
; entry: file handle in fhand
; exit:  file size in ax
;        error message printed if file exceeds 32K
;------------------------------------------------------------------------------

fileSiz: push    dx                     ; save regs
         push    cx
         push    bx
         xor     cx,cx                  ; file offset 0
         mov     dx,cx                  
         mov     bx,fhand               ; get file handle
         mov     ax,4202h               ; move file pointer to file end
         int     DOS                  
         or      dx,dx                  ; does it exceed 64K?
         jne     fil2big                ; error if so
         cmp     ax,maxPsz
         ja      fil2big                ; error if so
         mov     fSize,ax               ; else, save file size
         mov     ax,4200h               ; set file pointer back
         int     DOS
         pop     bx                     ; restore regs
         pop     cx
         pop     dx
         ret                            ; done with file size
fil2big: mov     di,offset opmErr       ; here if file too big
         jmp     error                  ;   go print error message
	 
;-----------------------------------------------------------------------------
; close file
; entry: file handle in fhand
; destroys ax
;-----------------------------------------------------------------------------

closFil: push    bx                     ; save reg
         mov     bx,fhand               ; get file handle
closFl1: mov     ax,3E00h               ; close file
         int     DOS
         pop     bx                     ; restore reg
         ret                            ; done with close file

;-----------------------------------------------------------------------------
; read file
; entry: file handle in fhand
;        file size in fsize
;        dx points to start address to load file
; reads file into program space
; destroys ax
;-----------------------------------------------------------------------------

readFil: push    bx                     ; save regs
         push    cx
         mov     bx,fhand               ; get file handle
         mov     cx,fsize               ; get file size
         mov     ax,3F00h               ; read file
         int     DOS
         jc      readFal                ; read error if carry set
         pop     cx                     ; restore regs
         pop     bx
         ret                            ; done with read file
readFal: mov     di,offset rdfErr       ; here if read file error
         jmp     error                  ; go print error

;-----------------------------------------------------------------------------
;write tee
;charcter to be printed is also wriiten to file output.asc
;entry: al holds character
;exit:  al sent to file sequentially
;-----------------------------------------------------------------------------
writTee: push    ax                     ; save regs
         push    bx
         push    cx
         push    dx
         mov     ChrTmp,al              ; put char in temp location
         mov     bx,prnhand             ; get output file handle
         mov     cx,1                   ; output just one character
         mov     dx,offset ChrTmp       ; point to character
         mov     ax,4000h               ; write file (character)
         int     DOS
         cmp     ax,1                   ; error if count other than 1
         jne     wrtFErr
         pop     dx                     ; restore regs
         pop     cx
         pop     bx
         pop     ax
         ret                            ; done with write tee
wrtFErr: mov     di,offset wrtErr       ; here if write file error
         jmp     error                  ; go print error

;-----------------------------------------------------------------------------
; hex2asc
; entry:  ax contains intger
;         di points to string location
;-----------------------------------------------------------------------------

hex2asc: push    cx                     ; save register
         mov     cx,4                   ; loop for a count of 4
h2a01:   rol     ax,1                   ; rotate left 1 digit
         rol     ax,1
         rol     ax,1
         rol     ax,1
         push    ax                     ; save it
         and     al,0Fh                 ; mask digit
         daa                            ; add 6 if A-F
         add     al,0F0h                ; bump a carry if A-F
         adc     al,040h                ; now ascii
         mov     byte ptr [di],al       ; store in string 
         inc     di                     ; bump string pointer
         pop     ax                     ; number back
         loop    h2a01                  ; do next digit
         pop     cx                     ; restore register
	 ret                            ; done with hex to ascii

;------------------------------------------------------------------------------
; prnHex
; entry: ax contains integer
;------------------------------------------------------------------------------

prnHex:  mov     di,offset intIO        ; point di to integer i/o string
         push    di                     ; save pointer
         call    hex2asc                ; convert to hex
         mov     byte ptr [di],NULL     ; store string terminator
         pop     di                     ; point to string
         jmp     prnStr                 ; go print it    
      
;------------------------------------------------------------------------------
; cold start routines
; called on program start and after a load
;------------------------------------------------------------------------------

cold:    mov     forClr,YELLOW          ; default colors
         mov     bakClr,DARKBLUE
         mov     ax,0003h               ; text - set mode 3 80x25 color text
         mov     gflag,0                ; clear gpaphics flag
         int     BIOS
         call    cls                    ; clear screen 
         mov     al,SPC
         call    prnChr
         call    homCur                 ; home cursor 
         mov     di,offset msgLogo      ; print logo message
         call    prnStr
         call    prnCRLF
         mov     fsize,0
cold2:   mov     di,offset program      ; init program space
         mov     prgEnd,di
         mov     byte ptr [di],NULL     ; put terminator
cold3:   xor     ax,ax                  ; zero out all stack indices
         mov     iStkX,al
         mov     fStkX,al
         mov     VarX,al
         mov     fVarX,al
         mov     fConX,al
         mov     ConX,al
         mov     lblX,al
         mov     wrdX,al
         mov     ifmt,6                 ; set numeric formats
         mov     fmtWid,13
         mov     fmtDec,7
         mov     expdig,2
         mov     expdec,9
         mov     tabWid,3               ; set tab width
; warm start entry
warm:    xor     al,al
         mov     pStkX,al
         mov     fpStkX,al
         mov     doStkX,al
         mov     baStkX,al
cold4:   and     prnTee,1               ; test tee
         je      cold1                  ; no tee if zero
         call    closTee                ; else close down tee
         mov     prnTee,0               ; turn tee off
cold1:   mov     errflg,0               ; error-off
         ret


;-----------------------------------------------------------------------------
; M A I N
;-----------------------------------------------------------------------------

start:   mov     ax,_DATA               ; initialize
         mov     ds,ax
         mov     ax,STACK
         mov     ss,ax
         cld

         call    cold

         mov     dx,offset outFN        ; create outfile.asc
         xor     cx,cx
         mov     ax,3C00h
         int     DOS
         mov     prnhand,ax
         call    seed

         mov     si,80h                 ; get command line address
         mov     cl,es:[si]             ; get count of chars
         or      cl,cl                  ; test count
         je      mloop                  ; go if none
         dec     cl                     ; first char a space
         je      mloop                  ; go if only a space
         xor     ch,ch                  ; make count a word
         inc     si                     ; bump pointer to filename
         inc     si
         mov     di,offset fname        ; point to fname storage
str01:   mov     al,byte ptr es:[si]    ; move filename to fname
         mov     byte ptr [di],al
         inc     si
         inc     di
         loop    str01
         mov     byte ptr [di],NULL     ; store null terminator
         mov     di,offset fname+2      ; point error parser to fname
         mov     nxtAdr,di
         call    xq114x                 ; go load and xeq
         call    parse
         
;------------------------------------------------------------------------------
; M A I N  L O O P
;------------------------------------------------------------------------------

mloop:   mov     ax,STACK               ; reset 8087 stack
         mov     ss,ax
         xor     ax,ax
         mov     rStkX,al               ; reset run, do and begin stacks
         mov     doStkX,al
         mov     baStkX,al
         mov     trflag,al              ; no trace
         mov     errflg,0               ; turn error jump off
         inc     ranNum                 ; juggle random number


         FINIT                          ; init coprocessor

m1:      call    getPLn                 ; get prompt line
         call    parse                  ; parse it
         jmp     short mloop            ; and do main loop all over

exit:    mov     ax,4C00h               ; DOS return
         int     DOS



.stack

end      start
